/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* -------------------------------------- */
/* The main program, initialization, etc. */
/* -------------------------------------- */

#include        <ctype.h>
#include        <signal.h>
#include        <sys/types.h>
#include        <sys/socket.h>
#include        <netinet/in.h>
extern "C" {
#include        <netdb.h>
}
#include        <sys/time.h>
#include        <setjmp.h>
#include        "types.h"
#include        "../SIMULATOR/display.h"      // Borrowed from friends

#if     ZZ_OWNLIB == 1
long    MemoryUsed;
#endif

/* ------------------------------------- */
/* Global variables (actual definitions) */
/* ------------------------------------- */

Template        *TempList;              // The list of templates

NewWindow       *Updates = NULL;        // The list of added windows

TItem   *HTable [HASHSIZE];             // The object dictionary
TItem   *Tree;                          // Root pointer

int     MonSocket,                      // Socket to the monitor
	SmrSocket,                      // Socket to the displayed smurph

	TDesc,                          // Terminal file descriptor

	CurCol, CurRow,                 // Cursor position
	NCols, NRows,                   // Screen parameters

	TermFlag;                       // Simulator termination flag

char    CBuffer [SANE];                 // Character buffer for a textual item

/* ---------------- */
/* Action functions */
/* ---------------- */

Action  ScRefresh,      // Screen refresh
	ScUp,           // Move the cursor up
	ScDown,         // Move the cursor down
	ScLeft,         // Move the cursor left
	ScRight,        // Move it right
	ScNext,         // Next screen
	ScPrev,         // Previous screen
	ScHelp,         // Display a help screen
	ScExit,         // Exit current mode
	ScQuit;         // Quit the program

OAction ScOther;        // Special

int     NStations;      // The number of stations

Window  *Active;        // The list of active windows

int     BigPrecision;

long    DisplayInterval;

Rectangle *FRectangles; // The list of free rectangles

int     FError;         // i/o error flag

char    *stfname = NULL,        // System template file name
	*utfname = NULL;        // User template file name

int     WCNAdmitted,    // Flag == WCN phrases are legal
	Stepped;        // Stepped window updated


#include        "../MONITOR/smdesc.c"

static  char    *cpname;                // Program call name

jmp_buf         restart;                // Recovery context

static  int     recvFlag = NO;          // YES if processing display request

#ifdef  DEBUG
ostream *dbgf;
#endif

#ifdef  DEBUG
int     tdspc, kcn;
static  void    dumpTItem (TItem *t) {

/* --------------------------------- */
/* Dumps one TItem to the debug file */
/* --------------------------------- */

	TItem   *ct;
	
	if (t != NULL) {
		for (kcn = 0; kcn < tdspc; kcn++) (*dbgf) << ' ';
		(*dbgf) << t->SName << '\n';
		tdspc++;
		for (ct = t->ChList; ct != NULL; ct = ct->lnext) dumpTItem (ct);
		tdspc--;
	}
}

void    dumpTItems () {

/* -------------------------------------- */
/* Dumps the dictionary to the debug file */
/* -------------------------------------- */

	tdspc = 0;
	dumpTItem (Tree);
}
#endif

void    clearActions () {

/* ----------------------------------- */
/* Reset the action functions to NULLs */
/* ----------------------------------- */

	ScRefresh =
	ScUp =
	ScDown =
	ScLeft =
	ScRight =
	ScNext =
	ScPrev =
	ScHelp =
	ScExit =
	ScQuit = NULL;
	ScOther = NULL;
}

static  void    badArgs () {

/* ----------------- */
/* Call format error */
/* ----------------- */

	cerr << "Usage: " << cpname << " [-l] [-s] or\n";
	cerr << "       " << cpname << " [-t fname] [-u fname]\n";
	exit (3);
}

void    setTimer () {

/* ------------- */
/* Set the timer */
/* ------------- */

	struct  itimerval       td;
	td.it_value.tv_sec = CDELAY;
	td.it_value.tv_usec = 0;
	td.it_interval.tv_sec = 0;
	td.it_interval.tv_usec = 0;
	setitimer (ITIMER_REAL, &td, NULL);
}

void    clearTimer () {

/* --------------------- */
/* Set the timeout timer */
/* --------------------- */

	struct  itimerval       td;
	td.it_value.tv_sec = 0;
	td.it_value.tv_usec = 0;
	td.it_interval.tv_sec = 0;
	td.it_interval.tv_usec = 0;
	setitimer (ITIMER_REAL, &td, NULL);
}

static  void    timeOut () {

/* ------------------------- */
/* Timeout interrupt service */
/* ------------------------- */

	if (recvFlag)
		abtDisplay ("Connection timed out");
	excptn ("connection timed out");
}

static  void    connectToMonitor () {

/* ----------------------------------- */
/* Establish connection to the monitor */
/* ----------------------------------- */

	struct  sockaddr_in     monaddr;
	struct  hostent         *host;
	int                     w;

	if ((host = gethostbyname (ZZ_MONHOST)) == NULL)
		excptn ("monitor connection failed -- no monitor host");

	if ((MonSocket = socket (AF_INET, SOCK_STREAM, 0)) < 0)
		excptn ("monitor connection problems -- cannot create socket");

	bzero (&monaddr, sizeof (monaddr));
	bcopy (host->h_addr, &(monaddr.sin_addr), host->h_length);
	monaddr.sin_family = host->h_addrtype;
	monaddr.sin_port = htons (ZZ_MONSOCK);

	if (connect (MonSocket, (sockaddr*)(&monaddr), sizeof (monaddr)))
		excptn ("monitor connection refused");

	// Send your signature

	setTimer ();
	w = write (MonSocket, SERVRSIGN, SIGNLENGTH);
	clearTimer ();
	if (w < 0)
		excptn ("write to monitor failed");
}

static  void    connectToSmurph (SmDesc *sm, long port) {

/* --------------------------------------------- */
/* Establish connection with the selected smurph */
/* --------------------------------------------- */

	struct  sockaddr_in     smphaddr;
	struct  hostent         *host;
	int                     w;
	char    sign [SIGNLENGTH];

	if ((host = gethostbyname (sm->HName)) == NULL)
		abtDisplay ("Cannot locate the smurph's host");

	if ((SmrSocket = socket (AF_INET, SOCK_STREAM, 0)) < 0)
		abtDisplay ("Cannot create socket to smurph");

	bzero (&smphaddr, sizeof (smphaddr));
	bcopy (host->h_addr, &(smphaddr.sin_addr), host->h_length);
	smphaddr.sin_family = host->h_addrtype;
	smphaddr.sin_port = (unsigned short) port;
	// Note that port already is in the network format

	if (connect (SmrSocket, (sockaddr*)(&smphaddr), sizeof (smphaddr)))
		abtDisplay ("Smurph refuses connection");

	// Send your signature
	setTimer ();
	w = write (SmrSocket, SERVRSIGN, SIGNLENGTH);
	clearTimer ();
	if (w < 0)
		abtDisplay ("Write to smurph failed");

	// Get the smurph's signature in response

	setTimer ();
	w = read (SmrSocket, sign, SIGNLENGTH);
	clearTimer ();
	if (w < 0)
		abtDisplay ("Conection broken by smurph");

	// Check the signature
	if (strncmp (sign, SMRPHSIGN, SIGNLENGTH))
		abtDisplay ("The responding party is not a smurph");
}

static  void    closeMonitorConnection () {

/* -------------------------------- */
/* Closes the socket to the monitor */
/* -------------------------------- */

	close (MonSocket);
}

static  void    sendRequest (int rq, long arg = NONE) {

/* ----------------------------- */
/* Send a request to the monitor */
/* ----------------------------- */

	char    c;
	long    dummy;

	c = rq;

	setTimer ();
	if (write (MonSocket, &c, 1) < 0) excptn ("write to monitor failed");
	if (arg != NONE) {
		dummy = htonl (arg);
		if (write (MonSocket, RWCAST &dummy, sizeof (long)) < 0)
			excptn ("write to monitor failed");
	}
	clearTimer ();
}

static  int     getByteFromMonitor () {

/* ------------------------------- */
/* Reads one byte from the monitor */
/* ------------------------------- */

	char    c;

	if (read (MonSocket, &c, 1) < 1) excptn ("read from monitor failed");
	return (((int)c) & 0377);
}

static  void    getTextFromMonitor (char *t, int nc) {

/* -------------------------------------------- */
/* Read a string of characters from the monitor */
/* -------------------------------------------- */

	char    c;

	for (nc--; ;) {
		if ((c = getByteFromMonitor ()) == '\0') break;
		if (nc > 0) {
			*t++ = c;
			nc--;
		}
	}

	*t = '\0';
}

static  long    getIntFromMonitor () {

/* ---------------------------------------------- */
/* Reads a (long) integer number from the monitor */
/* ---------------------------------------------- */

	long    dummy;

	if (read (MonSocket, RWCAST &dummy, sizeof (long)) < 0)
		excptn ("read from monitor failed");

	return (ntohl (dummy));
}

static  void    getSmurphs () {

/* ---------------------------------------- */
/* Read in information about active smurphs */
/* ---------------------------------------- */

	int     b;
	SmDesc  *sm;

	while (Smurphs != NULL) delete (Smurphs); // Cleanup

	connectToMonitor ();
	sendRequest (DR_INFALL);
	setTimer ();

	while (1) {
		if ((b = getByteFromMonitor ()) == PH_ESP)
			// End of list
			break;
		// b is the first character of host name

		sm = new SmDesc;
		sm->HName [0] = b;
		getTextFromMonitor (&(sm->HName [1]), 127);
		getTextFromMonitor (sm->PArgs, 128);
		getTextFromMonitor (sm->DTime, 32);
		sm->PId = getIntFromMonitor ();
		sm->Handle = getIntFromMonitor ();
	}
	clearTimer ();
	closeMonitorConnection ();
}

char    *rightAdjust (char *t, int size) {

/* ---------------------------------------------------------- */
/* Right  adjust the text pointed to by t so that it occupies */
/* exactly size characters                                    */
/* ---------------------------------------------------------- */

	register  int   k, l;
	register  char  *cp;

	cp = CBuffer;
	// Ignore trailing blanks
	for (l = strlen (t); l > 0 && t[l-1] == ' '; l--);
	if (l < size)
		for (k = size - l; k > 0; k--) *cp++ = ' ';
	else
		l = size;
	if (l > 0) strncpy (cp, t, l);
	cp [l] = '\0';

	return (CBuffer);
}

char    *leftAdjust (char *t, register int size) {

/* ---------------------------------------------------------- */
/* Left  adjust  the text pointed to by t so that it occupies */
/* exactly size characters                                    */
/* ---------------------------------------------------------- */

	register char   *cp;

	while (*t == ' ') t++;          // Ignore leading blanks
	for (cp = CBuffer; *t != '\0' && size > 0; *cp++ = *t++, size--);
	while (size-- > 0) *cp++ = ' ';
	*cp = '\0';
	return (CBuffer);
}

static  void    printLine (SmDesc *sm) {

/* ---------------------------------------------------------- */
/* Print  one  line  containing  information about one active */
/* smurph                                                     */
/* ---------------------------------------------------------- */

	int     i;

	cout << rightAdjust (sm->HName, 16) << ' ';
	cout << rightAdjust (sm->DTime, 19) << ' ';
	cout << form ("%6d", sm->PId) << ' ';
	leftAdjust (sm->PArgs, 55);
	for (i = 54; i > 0 && CBuffer [i] == ' '; i--);
	CBuffer [i+1] = '\0';   // Strip trailing blanks
	cout << CBuffer << '\n';
}

static  void    processList () {

/* ----------------------- */
/* List all active smurphs */
/* ----------------------- */

	SmDesc  *sm;

	if ((sm = Smurphs) == NULL) {
		cout << "No active smurphs\n";
		return;
	}
	cout << "        Hostname   Startup Date/Time    PID    Called as\n";

	for ( ; sm != NULL; sm = sm->next) printLine (sm);
}

static  void    processStatus () {

/* ------------------------------------- */
/* List the status of all active smurphs */
/* ------------------------------------- */

	SmDesc  *sm;
	char    txt [16], txu [16], txv [16];
	int     k;

	if ((sm = Smurphs) == NULL) {
		cout << "No active smurphs\n";
		return;
	}
	cout << "        Hostname   Startup Date/Time    PID    Called as\n";

	for ( ; sm != NULL; sm = sm->next) {
		printLine (sm); cout << " ... ";
		// Get the status information
		connectToMonitor ();
		sendRequest (DR_STATUS, sm->Handle);
		setTimer ();
		k = getByteFromMonitor ();
		if (k == PH_ESP) {
			clearTimer ();
			closeMonitorConnection ();
			// The smurph has disappeared
			cout << "completed";
		} else {
			txt [0] = k;
			getTextFromMonitor (&(txt[1]), 15);
			getTextFromMonitor (txu, 16);
			getTextFromMonitor (txv, 16);
			clearTimer ();
			closeMonitorConnection ();
			cout << "RM = " << leftAdjust (txt, 16) << "  ";
			cout << "VT = " << leftAdjust (txu, 16) << "  ";
			cout << "CP = " << leftAdjust (txv, 16);
		}
		cout << '\n';
	}
}

static  void    processDisplay () {

/* ----------------------------- */
/* Processes the display request */
/* ----------------------------- */

	SmDesc  *sm;
	long    port;

	Active = NULL;
	DisplayInterval = NONE;
	initTemplates ();
	initDisplay ();
	if (NCols < 64)
		excptn ("Terminal has too few columns");
	if (NRows < 16)
		excptn ("Terminal has too few rows");

	SmrSocket = -1;

	if (setjmp (restart)) {
		clearTimer ();
		close (SmrSocket);
		clearScreen ();
	}

	getSmurphs ();

	if (Smurphs == NULL) {
		clearScreen ();
		closeDisplay ();
		cout << "No active smurphs\n";
		exit (0);
	}

	NStations = 0;
	initDict ();            // Re-initialize object dictionary
	initWind ();            // Clear window data structures
	initUpdt ();            // And updates

	recvFlag = NO;                  // Abort on timeouts

	// sm returns the pointer to the selected smurph
	sm = smurphMenu ();

	clearScreen ();
	connectToMonitor ();
	sendRequest (DR_DISPLY, sm->Handle);

	// Read the response
	setTimer ();
	port = getIntFromMonitor ();
	clearTimer ();
	closeMonitorConnection ();

	// Now port tells the smurph's port number (it is already in
	// the network format)
	recvFlag = YES;
	// From now on will recover from timeouts
	connectToSmurph (sm, port);
	// Read the first spurt from smurph
	WCNAdmitted = NO;
	readSmurphSpurt ();
#ifdef  DEBUG
	dumpTItems ();
#endif
	// Do not respond immediately; the user may wish to open some windows
	// and/or put smurph into the step mode
	while (NCols) { // Always true
		// Selecting new windows
		objectMenu ();
		if (displayWindows ()) longjmp (restart, 1);
	}
}
		
void    errMessage (char *t) {

/* --------------------- */
/* Display error message */
/* --------------------- */

	outRLC (t);
}

void    abtDisplay (char *t) {

/* ---------------------------------------- */
/* Fatal error aborting the display session */
/* ---------------------------------------- */

	errMessage (t);
	close (SmrSocket);
	sleep (2);
	longjmp (restart, 1);
}

void    processUserInput () {

/* --------------------------------------- */
/* Function distributor for terminal input */
/* --------------------------------------- */

	char    c;

	while (NCols) { // Always true

		switch (c = inChar ()) {


			case 'r' :
			case '\014' :

				if (ScRefresh != NULL) {
					(*ScRefresh) ();
					continue;
				}
				break;

			case 'k' :

				if (ScUp != NULL) {
					(*ScUp) ();
					continue;
				}
				break;

			case 'j' :

				if (ScDown != NULL) {
					(*ScDown) ();
					continue;
				}
				break;

			case 'h' :

				if (ScLeft != NULL) {
					(*ScLeft) ();
					continue;
				}
				break;

			case 'l' :

				if (ScRight != NULL) {
					(*ScRight) ();
					continue;
				}
				break;

			case 'n' :

				if (ScNext != NULL) {
					(*ScNext) ();
					continue;
				}
				break;

			case 'p' :

				if (ScPrev != NULL) {
					(*ScPrev) ();
					continue;
				}
				break;

			case 'x' :

				if (ScExit != NULL) {
					(*ScExit) ();
					continue;
				}
				break;

			case 'q' :

				if (ScQuit != NULL) {
					(*ScQuit) ();
					continue;
				}
				break;

			case '?' :

				if (ScHelp != NULL) {
					(*ScHelp) ();
					continue;
				}
				break;

			default:
				break;
		}

		if (ScOther != NULL) (*ScOther) (c);
	}
}

main (int argc, char *argv []) {

	int     mode;

#ifdef  DEBUG
	dbgf = new ostream ("debug", "w");
#endif

	cpname = *argv;

	// Call format:
	//
	//      ds [-l] [-s]
	//
	//      -l  -- list all smurphs currently running
	//      -s  -- give status information of all smurphs
	//
	//      -t fname -- system template file name
	//      -u fname -- user template file name

	for (mode = CM_DISPLAY, argc--, argv++; argc > 0; argc--, argv++) {

		if (**argv != '-' || *((*argv)+2) != '\0') badArgs ();
		if (*((*argv)+1) == 'l') {
			if (mode != CM_DISPLAY || stfname != NULL ||
				utfname != NULL) badArgs ();
			mode = CM_LIST;
		} else if (*((*argv)+1) == 's') {
			if (mode != CM_DISPLAY || stfname != NULL ||
				utfname != NULL) badArgs ();
			mode = CM_STATUS;
		} else if (*((*argv)+1) == 't') {
			if (argc < 2 || mode != CM_DISPLAY || stfname != NULL)
				badArgs ();
			argc--;
			stfname = *++argv;
		} else if (*((*argv)+1) == 'u') {
			if (argc < 2 || mode != CM_DISPLAY || utfname != NULL)
				badArgs ();
			argc--;
			utfname = *++argv;
		} else
			badArgs ();
	}

	recvFlag = NO;
	signal (SIGALRM, (SIGARG) timeOut);

	// Start with reading in information about active smurphs

	switch (mode) {

		case    CM_LIST:

			getSmurphs ();
			processList ();
			exit (0);

		case    CM_STATUS:

			getSmurphs ();
			processStatus ();
			exit (0);

		case    CM_DISPLAY:

			BigPrecision = -1;      // Undefined
			// Initialize hash table
			for (int i = 0; i < HASHSIZE; i++) HTable [i] = NULL;
			Tree = NULL;
			processDisplay ();
			exit (0);
	}
}
