/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */

  // #define    DEBUG   1

#include        "../version.h"

/* ---------------------------------------------------------- */
/* This  is  the  basic include file for the terminal display */
/* server. It contains declarations of basic data types.      */
/* ---------------------------------------------------------- */

#ifdef  DEBUG
#define debug(a)        do { (*dbgf) << a << '\n'; dbgf->flush (); } while (0)
class   ostream;
extern  ostream         *dbgf;
#else
#define debug(a)        
#endif

#ifdef	ZZ_ATT_SUN
#define	RWCAST 	(char*)
#else
#define	RWCAST
#endif

/* ---------------------------------------------------------- */
/* Constants  of  standard  types  (machine-dependent: assume */
/* 32-bit long and integer                                    */
/* ---------------------------------------------------------- */
#define         MAXlong                 2147483647
#define         MINlong                 (-2147483648)
#define         MAXshort                32767
#define         MINshort                (-32768)
#define         MAXint                  MAXlong
#define         MINint                  MINlong
#define         MAXdft                  2147483648.0

#define         YES             1       // Boolean constants
#define         NO              0
#define         NONE            (-1)

#define         SANE            512     // A 'sane' limit on scratch array sizes
#define         CDELAY          240     // Default connection delay (seconds)

/* ---------------------------------------------------------- */
/* The  maximum  precision  of  BIG  numbers. Increase if the */
/* server starts complaining.                                 */
/* ---------------------------------------------------------- */
#define         MAX_BIG_PRECISION       16

/* ---------------------------------------------------------- */
/* Hash  table  size;  it must be a power of two. The maximum */
/* sensible size is 32K.                                      */
/* ---------------------------------------------------------- */
#define         HASHSIZE        4096

/* ---------------------------------------------------------- */
/* The  root  object  name. It should be the same as the type */
/* name of the System  station  in  smurph  (with  the  'ZZ_' */
/* prefix stripped off).                                      */
/* ---------------------------------------------------------- */
#define         ROOTNAME        "SYSTEM"

/* ----------------- */
/* Server call modes */
/* ----------------- */

#define         CM_LIST         0       // List active smurphs
#define         CM_STATUS       1       // Give status of all active smurphs
#define         CM_DISPLAY      2       // A display request

/* -------------------------------------------------- */
/* Constants describing relative status of a template */
/* -------------------------------------------------- */

#define         TS_ABSOLUTE     -1      // No station-relative
#define         TS_RELATIVE     -2      // Mandatory relative
#define         TS_FLEXIBLE     -3      // Relative or not

/* ----------- */
/* Field types */
/* ----------- */

#define         FT_FIXED        0       // Fixed contents
#define         FT_REPEAT       1       // Replication of last line
#define         FT_REGION       2       // Region description
#define         FT_FILLED       3       // Filled dynamically
#define         FT_HEADER       4       // Header field

/* ------------- */
/* Display modes */
/* ------------- */

#define         DM_NORMAL       0
#define         DM_REVERSE      1
#define         DM_BLINKING     2
#define         DM_BRIGHT       3

/* ----------------------------- */
/* Justification for fixed field */
/* ----------------------------- */

#define         LEFT            0
#define         RIGHT           1

/* ------------------- */
/* Template line types */
/* ------------------- */

#define         LT_LAY          0       // Layout
#define         LT_ESC          1       // Escape
#define         LT_REP          2       // Replication

/* ------------ */
/* Escape codes */
/* ------------ */

#define         ES_NSP          1       // Non-special character
#define         ES_FLD          2       // Field separation

/* ---------------------------- */
/* Display item attribute types */
/* ---------------------------- */

#define         AT_END          0       // No more attributes
#define         AT_DMO          1       // Display mode
#define         AT_DCH          2       // Region display character
#define         AT_DCS          3       // Region display characters
#define         AT_SCA          4       // Scaling factors

/* ------------------ */
/* Select I/O parties */
/* ------------------ */

#define         IO_USER         0       // Terminal i/o
#define         IO_SMURPH       1       // Smurph i/o

#ifndef NULL
#define         NULL            0
#endif

#if ZZ_OWNLIB == 1
#include        "../LIB/lib.h"
#else
#include        <stream.h>
#endif
#include        "../MONITOR/monitor.h"
#include        "big.h"

typedef void (*Action) ();              // Pointer to action function
typedef void (*OAction) (char);         // Pointer to 'other' action function

class   Field           {

/* ----------------------------------- */
/* The internal description of a field */
/* ----------------------------------- */

	friend  void    fieldFlags ();

#ifdef  DEBUG
	friend  void    dumpTemplates ();
#endif
	private:

	Field   *next;                  // List pointer

	public:

	friend  class   Window;

	char    type;                   // The type of this field

	Field ();
};

class   FixedField : public Field {

/* ------------------------------ */
/* A field with constant contents */
/* ------------------------------ */

	public:

	FixedField (short, short);

	char    dmode;                  // Display mode
	short   x, y;                   // Starting coordinates
	char    *contents;              // The stuff to be displayed
};

class   HeaderField : public Field {

/* ------------------------------ */
/* A field with constant contents */
/* ------------------------------ */

	public:

	char    dmode;                  // Display mode
	short   x, y;                   // Starting coordinates
	short   length;                 // The header length

	HeaderField (short, short);
};

class   FilledField : public Field {

/* -------------------------------- */
/* A field to be filled dynamically */
/* -------------------------------- */

	public:

	FilledField (short, short);

	char    dmode,                  // Display mode
		just;                   // Justification (LEFT, RIGHT)
	short   x, y,                   // Starting coordinates
		length;                 // The size
};

class   RegionField : public Field {

/* --------------------------- */
/* A field describing a region */
/* --------------------------- */

	public:

	RegionField (short, short, short);

	char    dmode;                  // Display mode
	short   x, y, x1, y1;           // Corner coordinates
	char    pc, lc;                 // Point/line display characters

	float   xo, xs, yo, ys;         // Default scaling
};

class   RepeatField : public Field {

/* ---------------------------------------------------------- */
/* A  dummy  field  describing  replication  of a sequence of */
/* previous fields                                            */
/* ---------------------------------------------------------- */

	public:

	Field   *from, *to;             // Replication boundaries
	int     count;                  // Replication count 0 == unlimited

	// Note: regions cannot occur within the replication boundaries

	RepeatField (int);
};

class   TItem;                  // Forward declaration

class   Template        {

/* ----------------------------------------------------- */
/* This is the internal description of a window template */
/* ----------------------------------------------------- */

	private:

	Template         *next; // List pointer

	public:

	friend  void     Add ();
	friend  Template *findTemp (TItem*, int&, int k = -2);
	friend  int      tempStatus (TItem*, int);
#ifdef  DEBUG
	friend  void     dumpTemplates ();
#endif

	char    *typename;      // The object type
	int     mode;           // The exposure mode
	short   status,         // How the window can be related to a station
		unbounded;      // Flag == height unbounded

	// Legal values for 'status' are:
	//
	//         -1  -- no station-relative
	//         -2  -- only station-relative (any station)
	//         -3  -- optionally station relative

	Field   *fieldlist;     // The list of fields

	short   *xclip,         // Clipping columns
		*yclip;         // Clipping rows

	short   nxclip,
		nyclip;         // Sizes of the clipping arrays

	short   xclipd,         // Default clipping column
		yclipd;         // Default clipping row

	Template ();
	void    layoutLine (int, short*);
};

class   Window          {

/* ---------------------------------------------------------- */
/* This  data  structure describes an active window displayed */
/* on the screen                                              */
/* ---------------------------------------------------------- */

	friend          void    removeObject ();
	friend          void    updateWindow ();
	friend          void    initWind ();
	friend          int     isFree (int, int, int, int, Window *ign = NULL);
	friend          void    buildRectangles (int, int, Window *ign = NULL);
	friend          void    selectFirst ();
	friend          void    moveToSelected ();
	friend          void    refreshScreen (int);
	friend          void    Move ();
	friend          void    Clip ();
	friend          void    Step ();
	friend          void    UnStep ();
	friend          void    Updd ();
	friend          void    Downdd ();
	friend          void    Leftdd ();
	friend          void    Rightdd ();
	friend          void    Remove ();
	friend          void    Add ();

	private:

	Window          *next, *prev;   // List pointers

	public:

	Template        *templ;         // Template pointer
	TItem           *objct;         // Object pointer
	short           x, y,           // Coordinates of left top corner
			xs,             // Current (possibly clipped) width
			ys;             // Current (possibly clipped) height

	int             sttn;           // The station the window is related to

	char            *lff;           // Index of the last filled field

	char            *contents;      // The image

	Window  (TItem*, Template*, int, int, int, int, int st = NONE);
	~Window ();

	int     activeFields ();        // How many fields are to be filled
	long    contentsLength ();      // Calculates the buffer size
	void    requestDisplay ();      // Request the window to be displayed
	void    displayOld ();          // Redisplays the window
	void    displayNew ();          // Updated the window
	char    *makeHeader (int, int d=NO);      // Creates the header
};

class   NewWindow      {

/* ---------------------------------------------------------- */
/* This  object  describes  a  new  window to be added to the */
/* active list                                                */
/* ---------------------------------------------------------- */

	public:

	NewWindow       *next;          // Pointer to next addition
	Window          *wind;          // The added window

	NewWindow (Window*);
};

class   TItem           {

/* --------------------------------------------- */
/* Describes a single item of the ownership tree */
/* --------------------------------------------- */

	public:

	TItem           *lnext, *lprev, // Ownership list pointers
			*hnext, *hprev, // Hash list pointers
			*ChList;        // Children

	char            *SName,         // Standard name of the object
			*BName,         // Base name
			*NName;         // Nickname

	TItem (char*, char*, char*, char*);
	TItem () {};
	~TItem ();

	char    *dName ();              // Returns displayable name
	char    *tName (char*);         // Returns type name
};

class   Rectangle       {

/* ------------------------------------------------- */
/* Describes one element of the free rectangles list */
/* ------------------------------------------------- */

	public:

	Rectangle       *next;          // List pointer

	int             x, y, xx, yy;   // Attributes

	Rectangle (int, int, int, int);
};

#if     ZZ_ATT
int     hash (register char*);  // Cfront doesn't like loops in inlines
TItem   *findObj (char*);
#else
static  inline  int  hash (register char *t) {

/* ------------------------------------ */
/* The hash function for standard names */
/* ------------------------------------ */

	register int val;

	for (val = 0; *t != '\0'; t++)
		val = ((val << 3) + *t) & (HASHSIZE-1);
	return (val);
};

extern  TItem   *HTable [HASHSIZE];     // Forward declaration

static  inline  TItem   *findObj (char *t) {

/* --------------------------------------- */
/* Find the object given its standard name */
/* --------------------------------------- */

	register TItem  *p;

	for (p = HTable [hash (t)]; p != NULL; p = p->hnext)
		if (strcmp (p->SName, t) == 0)
			return (p);
	return (NULL);
};
#endif

/* -------------------------- */
/* Global function prototypes */
/* -------------------------- */
void    clearActions ();
void    setTimer ();
void    clearTimer ();
char    *rightAdjust (char*, int);
char    *leftAdjust (char*, register int);
void    errMessage (char*);
void    abtDisplay (char*);
void    processUserInput ();
SmDesc  *smurphMenu ();
void    objectMenu ();
void    initDict ();
void    initWind ();
void    initUpdt ();
TItem   *findParent (TItem*);
void    buildRectangles (int, int, Window*);
void    destroyRectangles ();
int     fitWindow (Template*, int&, int&, int&, int&);
void    removeWindow (Window*);
char    *ltoa (long, char*, int);
int     displayWindows ();
int     getOctet ();
int     peekOctet ();
void    sendOctet (int);
void    sendText (char*);
long    getInt ();
BIG     getBIG ();
double  getFloat ();
void    sendInt (long);
void    sendBIG (BIG);
void    sendFloat (double);
void    readSmurphSpurt ();
void    setCurrentWindow ();
void    setCurrentWindow (int, int, int, int);
void    setCurrentWindow (Window*);
void    initDisplay ();
void    closeDisplay ();
void    absMove (int, int);
void    moveCursor (int, int);
void    clearScreen ();
void    excptn (char*);
void    outChar (char);
int     inChar ();
char    *inNumber (char*, int, int&);
int     isKill (char c);
int     isErase (char c);
void    startField (int);
void    endField ();
void    outText (char*);
void    outText (int, char*);
void    outText (int, char*, int);
void    blankOut (int, int, int);
void    clearLUC (int er = YES);
void    clearRUC (int er = YES);
void    clearLLC (int er = YES);
void    clearRLC (int er = YES);
void    outLUC (char*);
void    outRUC (char*);
void    outLLC (char*);
void    outRLC (char*);
void    refreshCorners ();
int     readyParty ();
void    initTemplates ();
void    skipWindow ();
void    tabt (char*);
TItem   *findByBName (char*);


#ifdef  sgi
typedef void (*SIGARG) (int, ...);      // Type of the signal function
#else
#ifdef	ZZ_ATT_SUN
#define	SIGARG	SIG_PF
#else
typedef void (*SIGARG) ();
#endif
#endif
