/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ------------------------------------------------- */
/* Functions connecting smurph to the display server */
/* ------------------------------------------------- */

#include        <sys/types.h>
#include        <sys/socket.h>
#include        <sys/time.h>
extern "C" {
#include        <netinet/in.h>
#include        <netdb.h>
}
#include        <setjmp.h>
#include        <fcntl.h>
#include        <signal.h>
#include        "display.h"
#include        "system.h"

/* --------------------------- */
/* Macros to set timeout clock */
/* --------------------------- */
#define         SETTIMER                zz_settimer (zz_ctimeout, zz_tmoutenv)
#define         CLEARTIMER              zz_cleartimer ();

static  int             monsock = -1,   // Monitor socket
			mastsock = -1,  // To be used by server to establish
					// connection
			linesock = -1;  // To communicate with the server

struct  sockaddr_in     master_id;

extern  jmp_buf         zz_tmoutenv;

void    *retjmp;                // Context to resume after timeout

void    zz_setasyncio (SIGARG action) {

/* ------------------------------------------ */
/* Assumes asynchronous input from the server */
/* ------------------------------------------ */

	signal (SIGIO, action);
	fcntl (linesock, F_SETOWN, zz_processId);
	fcntl (linesock, F_SETFL, FASYNC);
}

void    zz_setsyncio () {

/* ----------------------------------------- */
/* Assumes synchronous input from the server */
/* ----------------------------------------- */

	fcntl (linesock, F_SETFL, 0);
	signal (SIGIO, SIG_DFL);
}

void    zz_monitorwait (SIGARG action) {

/* ----------------------------- */
/* Assumes the monitor wait mode */
/* ----------------------------- */

	if (monsock >= 0) {
		signal (SIGIO, action);
		fcntl (monsock, F_SETOWN, zz_processId);
		fcntl (monsock, F_SETFL, FASYNC);
	}
}

void    zz_monitorstop () {

/* ----------------------------- */
/* Cancels the monitor wait mode */
/* ----------------------------- */

	if (monsock >= 0) {
		fcntl (monsock, F_SETFL, 0);
		signal (SIGIO, SIG_DFL);
	}
}

void    zz_closeMonitor () {

/* ----------------------------- */
/* Closes the monitor connection */
/* ----------------------------- */

	close (monsock);
}

static  void    timeout () {

/* ---------------------- */
/* Timeout signal service */
/* ---------------------- */

	longjmp ((int*)retjmp, YES);
}

void    zz_settimer (long delay, jmp_buf context) {

/* ------------------ */
/* Set interval timer */
/* ------------------ */

	struct  itimerval       td;
	td.it_value.tv_sec = delay;
	td.it_value.tv_usec = 0;
	td.it_interval.tv_sec = 0;
	td.it_interval.tv_usec = 0;
	retjmp = (void*)(&(context[0]));
	signal (SIGALRM, (SIGARG) timeout);
	signal (SIGPIPE, (SIGARG) timeout);
	setitimer (ITIMER_REAL, &td, NULL);
}

void    zz_cleartimer () {

/* -------------------- */
/* Clear interval timer */
/* -------------------- */

	struct  itimerval       td;
	td.it_value.tv_sec = 0;
	td.it_value.tv_usec = 0;
	td.it_interval.tv_sec = 0;
	td.it_interval.tv_usec = 0;
	setitimer (ITIMER_REAL, &td, NULL);
	signal (SIGALRM, SIG_DFL);
	signal (SIGPIPE, SIG_DFL);
}

static  jmp_buf stmc;
static  struct  hostent  *host = NULL;

int     zz_connectToMonitor () {

/* ---------------------------------------------------------- */
/* Tries  to  connect the simulator to the monitor; returns a */
/* negative value if the attempt fails                        */
/* ---------------------------------------------------------- */

	struct  sockaddr_in     maddr;
	char                    txt [128], **args;
	int                     argc, cnt;
	long                    dummy;

	if ((monsock = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
		// Rather unlikely
		return (monsock = -1);
	}

	if (host == NULL) host = gethostbyname (ZZ_MONHOST);
	bzero (&maddr, sizeof (maddr));
	bcopy (host->h_addr, &(maddr.sin_addr), host->h_length);
	maddr.sin_family = host->h_addrtype;
	maddr.sin_port = htons (ZZ_MONSOCK);

	if (connect (monsock, (struct sockaddr*)(&maddr), sizeof (maddr))) {
		// Connection refused
		return (monsock = -1);
	}

	if (setjmp (stmc)) {
		// After write timeout
		zz_cleartimer ();
		close (monsock);
		return (monsock = -1);
	}

	zz_settimer (30, stmc);

	// Send the signature
	if (write (monsock, SMRPHSIGN, SIGNLENGTH) < 0) longjmp (stmc, 1);

	// Send the hostname
	if (gethostname (txt, 128) < 0) longjmp (stmc, 1);
	txt [127] = '\0';               // Just in case
	if (write (monsock, txt, strlen (txt)+1) < 0) longjmp (stmc, 1);

	// Now for the program call line
	args = zz_callArgs;
	argc = zz_callArgc;

	for (cnt = 0; argc > 0 && cnt < 127; argc--, args++) {

		int l;

		l = strlen (*args);     // The length of the arg
		if (l > 127 - cnt) l = 127 - cnt;

		strncpy (&(txt[cnt]), *args, l);
		cnt += l;
		if (cnt < 127 && argc > 1) txt [cnt++] = ' ';
	}
	txt [cnt] = '\0';
	if (write (monsock, txt, strlen (txt)+1) < 0) longjmp (stmc, 1);

	// Startup date time
	strcpy (txt, tDate ());
	if (write (monsock, txt, strlen (txt)+1) < 0) longjmp (stmc, 1);

	// Process ID
	dummy = htonl (zz_processId);
	if (write (monsock, RWCAST &dummy, sizeof (long)) < 0) longjmp (stmc,1);
	zz_cleartimer ();
		
	// OK, this part is done
	return (0);
}

int     zz_getByteFromMonitor () {

/* ---------------------------------------------------------- */
/* Reads  one  byte  from  the  monitor;  called  after SIGIO */
/* interrupt in the MWAIT mode                                */
/* ---------------------------------------------------------- */
	char    c;

	if (read (monsock, &c, 1) < 1) {
		// Return unrecognizable value
		c = 0;
	}

	return (((int) c) & 0377);
}

void    zz_sendTextToMonitor (char *t) {

/* ---------------------------------------- */
/* Writes a character string to the monitor */
/* ---------------------------------------- */

	write (monsock, t, strlen (t)+1);
}


int     zz_openLine () {

/* ------------------------- */
/* Opens a server connection */
/* ------------------------- */

	long    dummy, pcn;
	char    sign [SIGNLENGTH];

	linesock = -1;

	// Find a free port for the master socket

	if ((mastsock = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
		// This shouldn't really happen
		dummy = htonl (-1);
		// Notify the server about failure
		write (monsock, RWCAST &dummy, sizeof (long));
		return (-1);
	}

	for (pcn = ZZ_MONSOCK + zz_processId%100 + 1; pcn < ZZ_MONSOCK + 10000;
		pcn++) {
		// Try to find a port (quite a few times)
		master_id.sin_family = AF_INET;
		master_id.sin_addr.s_addr = htonl (INADDR_ANY);
		master_id.sin_port = (unsigned short) htons (pcn);

		if (!bind (mastsock, (struct sockaddr*) &master_id,
		    sizeof (master_id)))
			// Done -- stay here
			break;
	}

	if (pcn >= ZZ_MONSOCK + 10000) {
		// Failure
		dummy = htonl (-1);
		// Notify the server about failure
		write (monsock, RWCAST &dummy, sizeof (long));
		return (-1);
	}

	listen (mastsock, 1);

	// OK, we've got the port, send the number
	dummy = htonl (pcn);
	if (write (monsock, RWCAST &dummy, sizeof (long)) < 0) return (-1);

	// Wait for connection from server
	
	while (1) {

		linesock = accept (mastsock, (struct sockaddr*) NULL,
			(int*)(&dummy));
		// This shouldn't happen
		if (linesock < 0) return (-1);

		// We have a connection -- time to exchange signatures

		// Read the server's signature
		if (read (linesock, sign, SIGNLENGTH) < SIGNLENGTH) return (-1);
		if (strncmp (SERVRSIGN, sign, SIGNLENGTH)) {
			// The wrong signature
			close (linesock);
			linesock = -1;
			continue;               // Keep on trying
		}

		// OK, send your own signature in response
		if (write (linesock, SMRPHSIGN, SIGNLENGTH) < 0) return (-1);
		break;
	}

	close (mastsock);                       // No longer needed
	mastsock = -1;
	return (1);
}

int     zz_getByteFromServer () {

/* ----------------------------- */
/* Read one byte from the server */
/* ----------------------------- */

	char    u;

	if (read (linesock, &u, 1) < 1) {
		debug ("Read from server failed");
		longjmp (zz_tmoutenv, YES);
	}
	return (((unsigned int) u) & 0377);
}

void    zz_sendByteToServer (int byte) {

/* --------------------------- */
/* Send one byte to the server */
/* --------------------------- */

	char    u;

	u = byte;
	if (write (linesock, &u, 1) < 0) {
		debug ("Write to server failed");
		longjmp (zz_tmoutenv, YES);
	}
}

void    zz_closeLine () {

/* -------------------- */
/* Close the connection */
/* -------------------- */

	if (mastsock >= 0) close (mastsock);
	if (linesock >= 0) close (linesock);
	mastsock = linesock = -1;
}
