/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ---------------------------------------------------------- */
/* This  is  the  standard  include  file  for  user protocol */
/* modules. It also constitutes  the  basic  part  of  system */
/* include files.                                             */
/* ---------------------------------------------------------- */

#include "../version.h"

#ifdef  ZZ_ATT
#define INLINE
#ifdef  ZZ_ATT_SUN
#define volatile
#endif
#else
#define INLINE  inline
#endif

/* ------------------------ */
/* #define ZZ_SMURPHDEBUG 1 */
/* ------------------------ */

#pragma once

#ifdef  ZZ_OLB
#include        "../LIB/lib.h"
#else
#include        <math.h>
#include        <stream.h>
#endif

/* ---------------------------------------------------------- */
/* Constants  of  standard  types  (machine-dependent: assume */
/* 32-bit long and integer)                                   */
/* ---------------------------------------------------------- */
#define         MAX_long                2147483647
#define         MIN_long                (-2147483648)
#define         MAX_short               32767
#define         MIN_short               (-32768)
#define         MAX_int                 MAX_long
#define         MIN_int                 MIN_long

/* ------------------- */
/* Scratch array sizes */
/* ------------------- */
#ifdef  sun
#define         ZZ_ACSSIZE              512
#else
#define         ZZ_ACSSIZE              1024
#endif
#define         ZZ_EETSIZE              1024
/* --------------------------------------------------------------------- */
/* ACSSIZE is the size of a scratch array (of type TIME) used by some of */
/* the  link  subroutines.  It is allocated on the stack and with static */
/* size - to avoid fragmentation that could have occurred, had  it  been */
/* allocated  by  malloc.  EETSIZE  is the size of another scratch array */
/* used (by get_event) and allocated on the stack for  similar  reasons. */
/* ACSSIZE  can  be increased if 'stack overflow' is signalled by a link */
/* subroutine. For some computers, e.g. Sun's, allocating  large  arrays */
/* on  the  stack  is impossible and ACSSIZE may have to be reduced. The */
/* amount of actual stack space required is equal  to  ACSSIZE  *  (8  * */
/* BIG_precision  +  8)  bytes.                                          */
/* --------------------------------------------------------------------- */

/* --------------------------------------------------------------------- */
/* The size of hash array for the QUEUE activity interpreter. If changed */
/* the corresponding changes must be made to the  AI  which  uses  fixed */
/* binary masks to build the hash index.                                 */
/* --------------------------------------------------------------------- */
#define         ZZ_QQSIZE               64

/* -------------------- */
/* Clock tick frequency */
/* -------------------- */
#ifndef         HZ
#define         ZZ_HZ                   60      // North America frequency
#else
#define         ZZ_HZ                   HZ
#endif

/* ------------------- */
/* User-visible macros */
/* ------------------- */

/* --------- */
/* Assertion */
/* --------- */
#if     ZZ_ASR
#define assert(a,b)     do { if (!(a)) excptn (b); } while (0)
#else
#define assert(a,b)     /* Void */
#endif
#define Assert(a,b)     do { if (!(a)) excptn (b); } while (0)

#if     ZZ_SMURPHDEBUG == 1
#define         zz_debug(s)     do { if (def (TracingTime) && (Time >=\
				TracingTime)) Ouf << "Time: " << Time <<\
				" --> " << s << '\n'; } while (0)
#else
#if     ZZ_SMURPHDEBUG == 2
#define         zz_debug(s)     do { (*zz_dbgf) << "Time: " << Time << " --> "\
				<< s << '\n'; zz_dbgf->flush (); } while (0)
#else
#define         zz_debug(s)
#endif
#endif

/* ------------------------------------------------------ */
/* A macro telling whether the protocol is being debugged */
/* ------------------------------------------------------ */
#define Debugging        (def (TracingTime) && (Time >= TracingTime))

/* --------------- */
/* Pointers to AIs */
/* --------------- */
#define         Timer   (&zz_AI_timer)
#define         Client  (&zz_AI_client)
#define         Queue   (&zz_AI_queue)
#define         Signal  (&zz_AI_signal)

/* ---------------------------------------------- */
/* A scratch pointer for SMURPH object generation */
/* ---------------------------------------------- */
#if     ZZ_ATT
extern  void    *zz_COBJ[];
extern  int     zz_clv;
#endif

/* ------------------------------ */
/* Converts station id to station */
/* ------------------------------ */
#if     ZZ_ASR
#define         idToStation(id)         (isStationId (id) ? zz_st [id] :\
			(Station*) excptn ("idToStation: illegal station id"))
#else
#define         idToStation(id)         zz_st [id]
#endif

/* ------------------------------ */
/* Converts Traffic id to Traffic */
/* ------------------------------ */
#if     ZZ_ASR
#define         idToTraffic(id)         (isTrafficId (id) ? zz_tp [id] :\
		     (Traffic*) excptn ("idToTraffic: illegal Traffic id"))
#else
#define         idToTraffic(id)         (zz_tp [id])
#endif

/* ------------------------ */
/* Converts Link id to Link */
/* ------------------------ */
#if     ZZ_ASR
#define         idToLink(id)            (isLinkId (id) ? zz_lk [id] :\
			(Link*) excptn ("idToLink: illegal Link id"))
#else
#define         idToLink(id)            (zz_lk [id])
#endif

/* ----------------------------------- */
/* Converts whatever to the identifier */
/* ----------------------------------- */
#define         ident(a)                ((a)->getId ())

/* ----------------------------- */
/* Aliases for Info01 and Info02 */
/* ----------------------------- */
#define         TheTraffic              ((int)Info02)
#define         TheValue1               ((long)Info01)
#define         TheValue2               ((long)Info02)
#define         ThePacket               ((Packet*)(Info01))
#define         ThePort                 ((Port*)(Info02))
#define         TheMessage              ((Message*)(Info01))
#define         TheSignal               ((int)(Info01))
#define         TheSender               ((int)(Info02))
#define         TheEvent                ((int)(Info02))
#define         TheItem                 Info01
#define         TheQueue                ((int)(Info02))
#define         TheExitCode             ((int)(Info01))

/* ---------------------------------------------------------- */
/* One  more  alias of the same type, but for something else. */
/* This is the pointer to  the  current  window's  associated */
/* object                                                     */
/* ---------------------------------------------------------- */
#define         TheWFrame               (zz_the_window->frame)

/* ----------------------------------------- */
/* Checks if a station identifier is correct */
/* ----------------------------------------- */
#define         isStationId(id)         ((id) >= 0 && (id) < ::NStations)

/* ----------------------------------------- */
/* Checks if a Traffic identifier is correct */
/* ----------------------------------------- */
#define         isTrafficId(id)         ((id) >= 0 && (id) < ::NTraffics)

/* -------------------------------------- */
/* Checks if a Link identifier is correct */
/* -------------------------------------- */
#define         isLinkId(id)            ((id) >= 0 && (id) < ::NLinks)

/* ------------------- */
/* Operations on flags */
/* ------------------- */
#define         setFlag(f,i)            (f |= (1 << (i)))
#define         clearFlag(f,i)          (f &= ~(1 << (i)))
#define         flagSet(f,i)            ((f & (1 << (i))) != 0)
#define         flagCleared(f,i)        ((f & (1 << (i))) == 0)

/* ---------------------------------------------------------- */
/* Operations  on  queues,  requests, and events (global, but */
/* invisible to the user)                                     */
/* ---------------------------------------------------------- */
#define         zz_eq           (zz_fsent.next) // Event queue pointer

#define         zz_queue_head(item,hd,type)     do {(item)->next = (hd);\
						 (item)->prev = (type*)(&(hd));\
	if ((hd) != NULL) (hd)->prev = (item);\
	(hd) = (item);} while (0)

#define         zz_queue_out(item)                      do {\
	if ((item)->next != NULL) ((item)->next)->prev = (item)->prev;\
	((item)->prev)->next = (item)->next;} while (0)

/* ---------------------------------- */
/* Convert stream pointers to streams */
/* ---------------------------------- */
#define         Inf                     (*zz_ifpp)
#define         Ouf                     (*zz_ofpp)

/* ------------------------- */
/* Global symbolic constants */
/* ------------------------- */

/* -------------- */
/* Boolean values */
/* -------------- */
#define         NO              0       // Logical value 'false'
#define         YES             1       // Logical value 'true'
#define         ON              1       // A flag is on
#define         OFF             0       // A flag is off
#define         UNDEFINED       2       // A flag is undefined

/* -------------------------- */
/* Traffic distribution flags */
/* -------------------------- */
#define         MIT_exp         1
#define         MIT_unf         2
#define         MLE_exp         4
#define         MLE_unf         8
#define         BIT_exp         16
#define         BIT_unf         32
#define         BSI_exp         64
#define         BSI_unf         128
#define         SCL_on          256
#define         SCL_off         512
#define         SPF_on          1024
#define         SPF_off         2048

/* ---------------------- */
/* Function return status */
/* ---------------------- */
#define         OK              0       // No error
#define         ERROR           1       // Error

/* ---------------------------- */
/* Simulation termination codes */
/* ---------------------------- */
#define         EXIT_msglimit   0       // Message number limit
#define         EXIT_stimelimit 1       // Simulated time limit
#define         EXIT_rtimelimit 2       // Real (CPU) time limit
#define         EXIT_noevents   3       // Event queue empty
#define         EXIT_user       4       // User-requested
#define         EXIT_abort      5       // Aborted

/* --------- */
/* Wildcards */
/* --------- */
#define         NONE            (-1)    // Dummy parameter
#define         ANY             (-1)    // For observers

/* -------------------------------- */
/* Integer (and TIME) type ordinals */
/* -------------------------------- */
#define         TYPE_double     (-1)    // A double used as TIME
#define         TYPE_long       0       // Long integer (standard)
#define         TYPE_short      1       // Short integer
#define         TYPE_BIG        2       // TIME

/* ---------------------------------- */
/* Seeds for random number generators */
/* ---------------------------------- */
#define         SEED_traffic    0       // Traffic seed
#define         SEED_delay      1       // Spacing seed
#define         SEED_toss       2       // Toss-a-coin seed
#define         ZZ_N_SEEDS      (SEED_toss+1)

/* ----------- */
/* Link events */
/* ----------- */
#define         SILENCE         0
#define         ACTIVITY        1
#define         COLLISION       2
#define         BOJ             3       // Start of jam
#define         EOJ             4       // End of jam
#define         BOT             5       // Start of packet
#define         EOT             6       // End of transfer
#define         BMP             7       // Start of packet addressed to me
#define         EMP             8       // End of packet addressed to me
#define         ANYEVENT        9       // Any status change of the carrier
#define         ZZ_MY_COLLISION 10      // Invisible to the user

#define         ZZ_N_LINK_EVENTS        (ZZ_MY_COLLISION+1)

/* ---------------------------------------------------------- */
/* User-visible  Bus  activity  types  (returned  in TheEvent */
/* after ANYEVENT)                                            */
/* ---------------------------------------------------------- */
#define         JAM                     0
#define         TRANSFER                1
#define         ENDTRANSFER             2
#define         ABTTRANSFER             3
#define         ENDJAM                  4

/* --------------------------------------- */
/* Events generated by the standard Client */
/* --------------------------------------- */
#define         ARRIVAL                 (-1)
#define         INTERCEPT               (-2)

/* ---------------------------------- */
/* Events generated by the process AI */
/* ---------------------------------- */
#define         DEATH                   (-1)    // One more in system.h

/* ------------------------------ */
/* Client time distribution modes */
/* ------------------------------ */
#define         EXPONENTIAL             0
#define         UNIFORM                 1

/* ---------------------------- */
/* Forward definitions of types */
/* ---------------------------- */
class ZZ_Object;
class EObject;
class RVariable;
class AI;
class SGroup;
class CGroup;
class Message;
class Packet;
class Traffic;
class Port;
class Link;
class Station;
class Process;
class Observer;

class ZZ_QITEM;
class ZZ_SIG;
class ZZ_EVENT;
class ZZ_REQUEST;
class ZZ_LINK_ACTIVITY;
class ZZ_INSPECT;
class ZZ_WINDOW;
class ZZ_SYSTEM;
class ZZ_PFItem;

#define         ZZ_EMD_0        4096    // Display expose offset

/* ------------------- */
/* Packet status flags */
/* ------------------- */
#define         PF_broadcast    31      // A broadcast packet
#define         PF_last         30      // The last packet of a message
#define         PF_bsignal      29      // Signal to the receiver on BOP
#define         PF_esignal      28      // Signal to the receiver on EOP
#define         PF_full         27      // The packet buffer is full

#define         PF_usr0         0       // User-accessible flags
#define         PF_usr1         1
#define         PF_usr2         2
#define         PF_usr3         3
#define         PF_usr4         4
#define         PF_usr5         5
#define         PF_usr6         6
#define         PF_usr7         7
#define         PF_usr8         8
#define         PF_usr9         9

/* ------------------- */
/* Station group types */
/* ------------------- */
#define         GT_selection    0
#define         GT_broadcast    NONE    // -1

/* ----------------------------------------------- */
/* Special constants for addSender and addReceiver */
/* ----------------------------------------------- */
#define         BROADCAST       NONE    // -1
#define         ALL             ((Station*)NULL)

/* ------------------ */
/* Predefined signals */
/* ------------------ */
#define         RECEIVE_b       (MIN_long)      // Receiving a packet BOP
#define         RECEIVE_e       (MIN_long+1)    // Receiving a packet EOP

/* ------------------------- */
/* Response after signalling */
/* ------------------------- */
#define         ACCEPTED        (-1)    // A process is awaiting the Signal
#define         PENDING         1       // No, the Signal becomes pending
#define         REJECTED        0       // The Signal is already pending

/* ---------- */
/* AI classes */
/* ---------- */
#define         AIC_timer       0
#define         AIC_queue       1
#define         AIC_signal      2
#define         AIC_link        3
#define         AIC_port        4
#define         AIC_client      5
#define         AIC_traffic     6
#define         AIC_process     7
#define         AIC_observer    8

#define         OBJ_rvariable   9
#define         OBJ_station     10      // Three honorary AIs
#define         OBJ_eobject     11      // User-created displayable

/* ---------- */
/* Link types */
/* ---------- */
#define         LT_broadcast            0       // Standard ether-type
#define         LT_cpropagate           1       // With propagating collisions
#define         LT_unidirectional       2       // Segmented fiber optic
#define         LT_pointtopoint         3       // Straight unidirectional

/* ------------------------- */
/* A few function prototypes */
/* ------------------------- */

/* ----------------------------------------- */
/* Simulation startup (provided by the user) */
/* ----------------------------------------- */
void    ZZRoot ();

/* ------------------------------------------------- */
/* Forces topology construction at the point of call */
/* ------------------------------------------------- */
void    buildNetwork ();

/* ------------------ */
/* Actual termination */
/* ------------------ */
void    zz_done ();
void    zz_noevents ();                 // Run out of events

/* ---------------------------------------------------------- */
/* Sets  the queue flush flag. A call of this function before */
/* the simulation is started has the same effect as selecting */
/* the -c call option of the simulator.                       */
/* ---------------------------------------------------------- */
void    setFlush ();

/* ------------------- */
/* Sets the time units */
/* ------------------- */
void    setEtu (double);
void    setItu (double);

/* ---------------------------------------------------------- */
/* Sets  the  clock  tolerance parameters. They are in effect */
/* for all stations defined since the  call  until  the  next */
/* call.                                                      */
/* ---------------------------------------------------------- */
void    setTolerance (double t = 0.0, int q = 2);

/* ---------------------------------------------------------- */
/* Sets  the  limit  on  the  number  of messages that can be */
/* queued at messages awaiting transmission. If this limit is */
/* reached, new generated messages are ignored.               */
/* ---------------------------------------------------------- */
void    setQSLimit (long lim = MAX_long);

/* ------------------------------------------- */
/* Cleanup after a nonstandard packet creation */
/* ------------------------------------------- */
void    zz_mkpclean ();

/* -------------------- */
/* Exception processing */
/* -------------------- */
int     excptn (char*);                 // Aborts the simulation
#if     ZZ_AER
int     zz_aerror (char*);              // Same, but after arithmetic error
#endif
void    zz_ierror (char*);              // Fatal error before starting

double  cpuTime ();                     // Gives the CPU execution time
char    *tDate ();                      // Gives current date and time

/* --------------------------------------------------- */
/* Encodes long numbers with optional decimal exponent */
/* --------------------------------------------------- */
void    encodeLong (long, char*, int nc = 10);
void    encodeInt (int i, char *s, int nc = 5);

/* ------------------------ */
/* Combine random variables */
/* ------------------------ */
void   combineRV (RVariable*, RVariable*, RVariable*);
inline void combineRV (RVariable *a, RVariable *b) { combineRV (a, b, b); };
inline void combineRV (RVariable &a, RVariable &b, RVariable &c) {
		combineRV (&a, &b, &c);
	    };
inline void combineRV (RVariable &a, RVariable &b) { combineRV (&a, &b, &b); };

/* ----------- */
/* Basic types */
/* ----------- */

/* ------------------------------------- */
/* Binary flags (e.g. packet attributes) */
/* ------------------------------------- */
typedef         long            FLAGS;
#if     ZZ_ATT
typedef         int             Boolean;
#else
typedef         char            Boolean;
#endif

/* ---------- */
/* BIG / TIME */
/* ---------- */

#if     BIG_precision > 1

/* ----------------------------------------------------- */
/* Definitions for multiple precision arithmetic package */
/* ----------------------------------------------------- */

class   BIG     {

	protected:

	long x [BIG_precision];

	public:

	inline  BIG () { };                     // Dummy constructor

	inline  BIG (const BIG& a) {            // Assignment of big to big
#if     BIG_precision > 5
		bcopy (a.x, x, sizeof (long) * BIG_precision);
#else
		x [0] = a.x [0];
		x [1] = a.x [1];
#if     BIG_precision > 2
		x [2] = a.x [2];
#if     BIG_precision > 3
		x [3] = a.x [3];
#if     BIG_precision > 4
		x [4] = a.x [4];
#endif
#endif
#endif
#endif
	};

/* ----------- */
/* Conversions */
/* ----------- */

	INLINE  BIG (long i);                   // From long integer
	BIG     (double);                       // Conversion from double/float

	// The following dummy conversion is used as a means of initializing
	// the package which must be done before any user generated conversion
	// (e.g. at static initialization of BIG from int or double) is
	// attempted (see the beginning of mpartihm.c)
	BIG     (BIG*);

	operator        long () const;          // Conversion to long/int
	operator        double () const;        // Conversion to float/double

	// To check convertibility to double
	friend  int     convertible (const BIG&);
	// Internal
	friend  void    zz_shiftB (BIG&, int);

/* ----------------------------------------- */
/* Basic arithmetic relations and operations */
/* ----------------------------------------- */

	friend  INLINE  int  operator== (const BIG &a, const BIG &b);
	friend  INLINE  int  operator!= (const BIG &a, const BIG &b);

	friend  inline  int  operator== (const BIG &a, long b) {
		return (a == (BIG) b);
	};
	friend  inline  int  operator!= (const BIG &a, long b) {
		return (a != (BIG) b);
	};
	friend  inline  int  operator== (long a, const BIG &b) {
		return ((BIG) a == b);
	};
	friend  inline  int  operator!= (long a, const BIG &b) {
		return ((BIG) a != b);
	};
	friend  inline  int  operator== (const BIG &a, double b) {
		return (a == (BIG) b);
	};
	friend  inline  int  operator!= (const BIG &a, double b) {
		return (a != (BIG) b);
	};
	friend  inline  int  operator== (double a, const BIG &b) {
		return ((BIG) a == b);
	};
	friend  inline  int  operator!= (double a, const BIG &b) {
		return ((BIG) a != b);
	};

	friend  int     operator<  (const BIG&, const BIG&);  // Inequalities
	friend  int     operator<= (const BIG&, const BIG&);
	friend  int     operator>  (const BIG&, const BIG&);
	friend  int     operator>= (const BIG&, const BIG&);

	friend  inline int operator< (const BIG &a, long b) {
		return (a < (BIG) b);
	};
	friend  inline int operator< (long a, const BIG &b) {
		return ((BIG) a < b);
	};
	friend  inline int operator< (const BIG &a, double b) {
		return (a < (BIG) b);
	};
	friend  inline int operator< (double a, const BIG &b) {
		return ((BIG) a < b);
	};
	friend  inline int operator<= (const BIG &a, long b) {
		return (a <= (BIG) b);
	};
	friend  inline int operator<= (long a, const BIG &b) {
		return ((BIG) a <= b);
	};
	friend  inline int operator<= (const BIG &a, double b) {
		return (a <= (BIG) b);
	};
	friend  inline int operator<= (double a, const BIG &b) {
		return ((BIG) a <= b);
	};
	friend  inline int operator> (const BIG &a, long b) {
		return (a > (BIG) b);
	};
	friend  inline int operator> (long a, const BIG &b) {
		return ((BIG) a > b);
	};
	friend  inline int operator> (const BIG &a, double b) {
		return (a > (BIG) b);
	};
	friend  inline int operator> (double a, const BIG &b) {
		return ((BIG) a > b);
	};
	friend  inline int operator>= (const BIG &a, long b) {
		return (a >= (BIG) b);
	};
	friend  inline int operator>= (long a, const BIG &b) {
		return ((BIG) a >= b);
	};
	friend  inline int operator>= (const BIG &a, double b) {
		return (a >= (BIG) b);
	};
	friend  inline int operator>= (double a, const BIG &b) {
		return ((BIG) a >= b);
	};

	friend  BIG     operator+ (const BIG&, const BIG&); // Addition
	friend  BIG     operator+ (const BIG&, long);       // Addition of long
	friend  double  operator+ (const BIG&, double);     // Addition of dble

	friend  inline BIG operator+ (long a, const BIG &b) {
		return (b + a);
	};
	friend  inline double operator+ (double a, const BIG &b) {
		return (b + a);
	};

	friend  BIG     operator- (const BIG&, const BIG&); // Subtraction
	friend  BIG     operator- (const BIG&, long);       // ... of int
	friend  double  operator- (const BIG&, double);     // ... float
	friend  double  operator- (double, const BIG&);     // ... from float

	inline  BIG     operator++ () {                 // Autoincrement
		return (*this = *this + 1L);
	};
	inline  BIG     operator-- () {                 // Autodecrement
		return (*this = *this - 1L);
	};
	inline  BIG     operator+= (const BIG a) {
		return (*this = *this + a);
	};
	inline  BIG     operator-= (const BIG a) {
		return (*this = *this - a);
	};
	inline  BIG     operator+= (long a) {
		return (*this = *this + a);
	};
	inline  BIG     operator-= (long a) {
		return (*this = *this - a);
	};

	friend  BIG     operator* (const BIG&, const BIG&); // Multiplication
	friend  BIG     operator* (const BIG&, long);       // By long
	friend  double  operator* (const BIG&, double);     // By double

	friend  inline BIG operator* (long a, const BIG &b) {
		return (b * a);
	};
	friend  inline double operator* (double a, const BIG &b) {
		return (b * a);
	};

	friend  BIG     operator/ (const BIG&, const BIG&); // Division
	friend  double  operator/ (const BIG&, double);     // By double
	friend  double  operator/ (double, const BIG&);     // Of double

	friend  inline BIG operator/ (const BIG &a, long b) {
		return (a / (BIG) b);
	};

	friend  BIG     operator% (const BIG&, const BIG&); // Modulo
	friend  long    operator% (const BIG&, long);       // By long

	inline  BIG     operator*= (const BIG a) {

		return (*this = *this * a);
	};
	inline  BIG     operator*= (long a) {

		return (*this = *this * a);
	};
	inline  BIG     operator/= (const BIG a) {

		return (*this = *this / a);
	};
	inline  BIG     operator/= (long a) {

		return (*this = *this / a);
	};
	inline  BIG     operator%= (const BIG a) {

		return (*this = *this % a);
	};
	inline  BIG     operator%= (long a) {

		return (*this = *this % a);
	};

	friend  BIG     atob (const char*);     // String to BIG conversion

						// And vice versa
	friend  char    *btoa (const BIG&, char *s = NULL, int nc = 15);

#if     ZZ_ATT
	INLINE  BIG (int i);

	inline  BIG (float a) { *this = (double) a; };

	inline  operator float () const { return ((float)((double) (*this))); };

	friend  inline  int  operator== (const BIG &a, int b) {
		return (a == (BIG) b);
	};
	friend  inline  int  operator!= (const BIG &a, int b) {
		return (a != (BIG) b);
	};
	friend  inline  int  operator== (int a, const BIG &b) {
		return ((BIG) a == b);
	};
	friend  inline  int  operator!= (int a, const BIG &b) {
		return ((BIG) a != b);
	};
	friend  inline  int  operator== (const BIG &a, float b) {
		return (a == (BIG) b);
	};
	friend  inline  int  operator!= (const BIG &a, float b) {
		return (a != (BIG) b);
	};
	friend  inline  int  operator== (float a, const BIG &b) {
		return ((BIG) a == b);
	};
	friend  inline  int  operator!= (float a, const BIG &b) {
		return ((BIG) a != b);
	};
	friend  inline int operator< (const BIG &a, int b) {
		return (a < (BIG) b);
	};
	friend  inline int operator< (int a, const BIG &b) {
		return ((BIG) a < b);
	};
	friend  inline int operator< (const BIG &a, float b) {
		return (a < (BIG) b);
	};
	friend  inline int operator< (float a, const BIG &b) {
		return ((BIG) a < b);
	};
	friend  inline int operator<= (const BIG &a, int b) {
		return (a <= (BIG) b);
	};
	friend  inline int operator<= (int a, const BIG &b) {
		return ((BIG) a <= b);
	};
	friend  inline int operator<= (const BIG &a, float b) {
		return (a <= (BIG) b);
	};
	friend  inline int operator<= (float a, const BIG &b) {
		return ((BIG) a <= b);
	};
	friend  inline int operator> (const BIG &a, int b) {
		return (a > (BIG) b);
	};
	friend  inline int operator> (int a, const BIG &b) {
		return ((BIG) a > b);
	};
	friend  inline int operator> (const BIG &a, float b) {
		return (a > (BIG) b);
	};
	friend  inline int operator> (float a, const BIG &b) {
		return ((BIG) a > b);
	};
	friend  inline int operator>= (const BIG &a, int b) {
		return (a >= (BIG) b);
	};
	friend  inline int operator>= (int a, const BIG &b) {
		return ((BIG) a >= b);
	};
	friend  inline int operator>= (const BIG &a, float b) {
		return (a >= (BIG) b);
	};
	friend  inline int operator>= (float a, const BIG &b) {
		return ((BIG) a >= b);
	};
	friend  inline BIG operator+ (const BIG &a, int b) {
		return (a + (long)b);
	};
	friend  inline float operator+ (const BIG &a, float b) {
		return ((float)(((double) a) + b));
	};
	friend  inline BIG operator+ (int a, const BIG &b) {
		return (b + (long) a);
	};
	friend  inline float operator+ (float a, const BIG &b) {
		return ((float)(((double) b) + a));
	};
	friend  inline BIG operator- (const BIG &a, int b) {
		return (a - (long)b);
	};
	friend  inline float operator- (const BIG &a, float b) {
		return ((float)(((double) a) - b));
	};
	friend  inline float operator- (float a, const BIG &b) {
		return ((float)(((double)a) - b));
	};

	inline  BIG     operator+= (int a) {
		return (*this = *this + (long) a);
	};
	inline  BIG     operator-= (int a) {
		return (*this = *this - (long) a);
	};

	friend  inline BIG operator* (const BIG &a, int b) {
		return (a * (long) b);
	};
	friend  inline double  operator* (const BIG &a, float b) {
		return (a * (double) b);
	};
	friend  inline BIG operator* (int a, const BIG &b) {
		return (b * (long) a);
	};
	friend  inline double operator* (float a, const BIG &b) {
		return (((double)b) * a);
	};
	friend  inline float operator/ (const BIG &a, float b) {
		return ((float)(a / (double)b));
	};
	friend  inline float operator/ (float a, const BIG &b) {
		return ((float)((double)a / b));
	};
	friend  inline BIG operator/ (const BIG &a, int b) {
		return (a / (BIG) b);
	};
	friend  inline int operator% (const BIG &a, int b) {
		return ((int)(a % (long) b));
	};

	inline  BIG     operator*= (int a) {
		return (*this = *this * (long) a);
	};
	inline  BIG     operator/= (int a) {
		return (*this = *this / (long) a);
	};
	inline  BIG     operator%= (int a) {
		return (*this = (BIG) (*this % (long) a));
	};
#endif
};

#endif
#if     BIG_precision == TYPE_double

/* ---------------------------------------------------------- */
/* Definitions  for  multiple  precision  arithmetic  package */
/* using double representation                                */
/* ---------------------------------------------------------- */

class   BIG     {

	protected:

	double x;

	public:

	inline  BIG () { };                             // Dummy constructor

	inline  BIG (const BIG &a) {
		x = a.x;
	};

/* -------------------- */
/* Arithmetic relations */
/* -------------------- */

	friend  inline  int  operator== (const BIG a, const BIG b) {
		return (a.x == b.x);
	};
	friend  inline  int  operator!= (const BIG a, const BIG b) {
		return (a.x != b.x);
	};
	friend  inline  int  operator== (const BIG a, long b) {
		return (a.x == (double) b);
	};
	friend  inline  int  operator!= (const BIG a, long b) {
		return (a.x != (double) b);
	};
	friend  inline  int  operator== (long a, const BIG b) {
		return ((double) a == b.x);
	};
	friend  inline  int  operator!= (long a, const BIG b) {
		return ((double) a != b.x);
	};
	friend  inline  int  operator== (const BIG a, double b) {
		return (a.x == floor (b));
	};
	friend  inline  int  operator!= (const BIG a, double b) {
		return (a.x != floor (b));
	};
	friend  inline  int  operator== (double a, const BIG b) {
		return (floor (a) == b.x);
	};
	friend  inline  int  operator!= (double a, const BIG b) {
		return (floor (a) != b.x);
	};
	friend  inline  int operator<  (const BIG a, const BIG b) {
		return (a.x < b.x);
	};
	friend  inline int operator< (const BIG a, long b) {
		return (a.x < (double) b);
	};
	friend  inline int operator< (long a, const BIG b) {
		return ((double) a < b.x);
	};
	friend  inline int operator< (const BIG a, double b) {
		return (a.x < floor (b));
	};
	friend  inline int operator< (double a, const BIG b) {
		return (floor (a) < b.x);
	};
	friend  inline  int operator<= (const BIG a, const BIG b) {
		return (a.x <= b.x);
	};
	friend  inline int operator<= (const BIG a, long b) {
		return (a.x <= (double) b);
	};
	friend  inline int operator<= (long a, const BIG b) {
		return ((double) a <= b.x);
	};
	friend  inline int operator<= (const BIG a, double b) {
		return (a.x <= floor (b));
	};
	friend  inline int operator<= (double a, const BIG b) {
		return (floor (a) <= b.x);
	};
	friend  inline int operator>  (const BIG a, const BIG b) {
		return (a.x > b.x);
	};
	friend  inline int operator> (const BIG a, long b) {
		return (a.x > (double) b);
	};
	friend  inline int operator> (long a, const BIG b) {
		return ((double) a > b.x);
	};
	friend  inline int operator> (const BIG a, double b) {
		return (a.x > floor (b));
	};
	friend  inline int operator> (double a, const BIG b) {
		return (floor (a) > b.x);
	};
	friend  inline int operator>= (const BIG a, const BIG b) {
		return (a.x >= b.x);
	};
	friend  inline int operator>= (const BIG a, long b) {
		return (a.x >= (double) b);
	};
	friend  inline int operator>= (long a, const BIG b) {
		return ((double) a >= b.x);
	};
	friend  inline int operator>= (const BIG a, double b) {
		return (a.x >= floor (b));
	};
	friend  inline int operator>= (double a, const BIG b) {
		return (floor (a) >= b.x);
	};

/* ---------- */
/* Operations */
/* ---------- */

	friend  inline BIG operator+ (const BIG a, const BIG b) {
		BIG     r;   r.x = a.x + b.x;   return r;
	};
	friend  inline BIG operator+ (const BIG a, long b) {
		BIG     r;   r.x = a.x + (double) b;   return r;
	};
	friend  inline double operator+ (const BIG a, double b) {
		return (a.x + b);
	};
	friend  inline BIG operator+ (long a, const BIG b) {
		BIG     r;   r.x = b.x + (double) a;   return r;
	};
	friend  inline double operator+ (double a, const BIG b) {
		return (b.x + a);
	};
	friend  inline BIG operator- (const BIG a, const BIG b) {
		BIG     r;   r.x = a.x - b.x;   return r;
	};
	friend  inline BIG operator- (const BIG a, long b) {
		BIG     r;   r.x = a.x - (double) b;   return r;
	};
	friend  inline double operator- (const BIG a, double b) {
		return (a.x - b);
	};
	friend  inline double operator- (double a, const BIG b) {
		return (a - b.x);
	};
	inline  BIG     operator++ () {                 // Autoincrement
		x = x + (double) 1;
		return (*this);
	};
	inline  BIG     operator-- () {                 // Autodecrement
		x = x - (double) 1;
		return (*this);
	};
	inline  BIG     operator+= (const BIG a) {
		x += a.x;
		return (*this);
	};
	inline  BIG     operator+= (long a) {
		x += (double) a;
		return (*this);
	};
	inline  BIG     operator-= (const BIG a) {
		x -= a.x;
		return (*this);
	};
	inline  BIG     operator-= (long a) {
		x -= (double) a;
		return (*this);
	};
	friend  inline BIG operator* (const BIG a, const BIG b) {
		BIG     r;   r.x = a.x * b.x;   return r;
	};
	friend  inline BIG operator* (const BIG a, long b) {
		BIG     r;   r.x = a.x * (double) b;   return r;
	};
	friend  inline double operator* (const BIG a, double b) {
		return (a.x * b);
	};
	friend  inline BIG operator* (long a, const BIG b) {
		BIG     r;   r.x = (double) a * b.x;   return r;
	};
	friend  inline double operator* (double a, const BIG b) {
		return (b.x * a);
	};
	friend  inline BIG operator/ (const BIG a, const BIG b) {
		BIG     r;   r.x = floor ( a / b );   return r;
	};
	friend  inline double operator/ (const BIG a, double b) {
		return (a.x / b);
	};
	friend  inline double operator/ (double a, const BIG b) {
		return (a / b.x);
	}
	friend  inline BIG operator/ (const BIG a, long b) {
		BIG     r;   r.x = floor (a.x / (double) b);   return r;
	};
	friend  inline BIG operator% (const BIG a, const BIG b) {
		BIG     r;
		double  f;

		f = floor (a.x / b.x);
		r.x = a.x - (f * b.x);
		return r;
	};
	friend  inline long operator% (const BIG a, long b) {
		double  f, ff;

		f = floor (a.x / (ff = (double) b));
		return ((long) (a.x - (f * ff)));
	};
	inline  BIG     operator*= (const BIG a) {
		x *= a.x;
		return (*this);
	};
	inline  BIG     operator*= (long a) {
		x *= (double) a;
		return (*this);
	};
	inline  BIG     operator/= (const BIG a) {
		x = floor (x / a.x);
		return (*this);
	};
	inline  BIG     operator/= (long a) {
		x = floor (x / (double) a);
		return (*this);
	};
	inline  BIG     operator%= (const BIG a) {
		double  f;

		f = floor (x / a.x);
		x = x - (f * a.x);
		return (*this);
	};
	inline  BIG     operator%= (long a) {
		double  f, ff;

		f = floor (x / (ff = (double) a));
		x = x - (f * ff);
		return (*this);
	};

/* ----------- */
/* Conversions */
/* ----------- */

	inline BIG      (long i) {      // Conversion from long

		x = (double) i;
	};
	inline BIG      (double d) {    // Conversion from double

		x = floor (d);
	};
	inline operator long () const { // Conversion to long

		return ((long) x);
	};
	inline operator double () const { // Conversion to double

		return x;
	};

	// To check convertibility to double

	friend  inline int convertible (const BIG) { return (1); };

	friend  BIG     atob (const char*);     // String to BIG conversion

						// And vice versa
	friend  char    *btoa (const BIG&, char *s = NULL, int nc = 15);

#if     ZZ_ATT
	friend  inline  int  operator== (const BIG a, int b) {
		return (a.x == (double) b);
	};
	friend  inline  int  operator!= (const BIG a, int b) {
		return (a.x != (double) b);
	};
	friend  inline  int  operator== (int a, const BIG b) {
		return ((double) a == b.x);
	};
	friend  inline  int  operator!= (int a, const BIG b) {
		return ((double) a != b.x);
	};
	friend  inline  int  operator== (const BIG a, float b) {
		return (a.x == floor (b));
	};
	friend  inline  int  operator!= (const BIG a, float b) {
		return (a.x != floor (b));
	};
	friend  inline  int  operator== (float a, const BIG b) {
		return (floor (a) == b.x);
	};
	friend  inline  int  operator!= (float a, const BIG b) {
		return (floor (a) != b.x);
	};
	friend  inline int operator< (const BIG a, int b) {
		return (a.x < (double) b);
	};
	friend  inline int operator< (int a, const BIG b) {
		return ((double) a < b.x);
	};
	friend  inline int operator< (const BIG a, float b) {
		return (a.x < floor (b));
	};
	friend  inline int operator< (float a, const BIG b) {
		return (floor (a) < b.x);
	};
	friend  inline int operator<= (const BIG a, int b) {
		return (a.x <= (double) b);
	};
	friend  inline int operator<= (int a, const BIG b) {
		return ((double) a <= b.x);
	};
	friend  inline int operator<= (const BIG a, float b) {
		return (a.x <= floor (b));
	};
	friend  inline int operator<= (float a, const BIG b) {
		return (floor (a) <= b.x);
	};
	friend  inline int operator> (const BIG a, int b) {
		return (a.x > (double) b);
	};
	friend  inline int operator> (int a, const BIG b) {
		return ((double) a > b.x);
	};
	friend  inline int operator> (const BIG a, float b) {
		return (a.x > floor (b));
	};
	friend  inline int operator> (float a, const BIG b) {
		return (floor (a) > b.x);
	};
	friend  inline int operator>= (const BIG a, int b) {
		return (a.x >= (double) b);
	};
	friend  inline int operator>= (int a, const BIG b) {
		return ((double) a >= b.x);
	};
	friend  inline int operator>= (const BIG a, float b) {
		return (a.x >= floor (b));
	};
	friend  inline int operator>= (float a, const BIG b) {
		return (floor (a) >= b.x);
	};

	friend  inline BIG operator+ (const BIG a, int b) {
		BIG     r;   r.x = a.x + (double) b;   return r;
	};
	friend  inline double operator+ (const BIG a, float b) {
		return (a.x + b);
	};
	friend  inline BIG operator+ (int a, const BIG b) {
		BIG     r;   r.x = b.x + (double) a;   return r;
	};
	friend  inline double operator+ (float a, const BIG b) {
		return (b.x + a);
	};
	friend  inline BIG operator- (const BIG a, int b) {
		BIG     r;   r.x = a.x - (double) b;   return r;
	};
	friend  inline double operator- (const BIG a, float b) {
		return (a.x - b);
	};
	friend  inline double operator- (float a, const BIG b) {
		return (a - b.x);
	};
	inline  BIG     operator+= (int a) {
		x += (double) a;
		return (*this);
	};
	inline  BIG     operator-= (int a) {
		x -= (double) a;
		return (*this);
	};
	friend  inline BIG operator* (const BIG a, int b) {
		BIG     r;   r.x = a.x * (double) b;   return r;
	};
	friend  inline double operator* (const BIG a, float b) {
		return (a.x * b);
	};
	friend  inline BIG operator* (int a, const BIG b) {
		BIG     r;   r.x = (double) a * b.x;   return r;
	};
	friend  inline double operator* (float a, const BIG b) {
		return (b.x * a);
	};
	friend  inline double operator/ (const BIG a, float b) {
		return (a.x / b);
	};
	friend  inline double operator/ (float a, const BIG b) {
		return (a / b.x);
	}
	friend  inline BIG operator/ (const BIG a, int b) {
		BIG     r;   r.x = floor (a.x / (double) b);   return r;
	};
	friend  inline long operator% (const BIG a, int b) {
		double  f, ff;

		f = floor (a.x / (ff = (double) b));
		return ((long) (a.x - (f * ff)));
	};
	inline  BIG     operator*= (int a) {
		x *= (double) a;
		return (*this);
	};
	inline  BIG     operator/= (int a) {
		x = floor (x / (double) a);
		return (*this);
	};
	inline  BIG     operator%= (int a) {
		double  f, ff;

		f = floor (x / (ff = (double) a));
		x = x - (f * ff);
		return (*this);
	};

	inline BIG      (int i) {       // Conversion from int

		x = (double) i;
	};
	inline BIG      (float d) {     // Conversion from float

		x = floor ((double)d);
	};
	inline operator int () const {  // Conversion to int

		return ((int) x);
	};
	inline operator float () const {// Conversion to float

		return ((float) x);
	};
#endif
};

#endif
#if     BIG_precision == 1

/* ----------------------------------------------- */
/* Definitions for single precision BIG arithmetic */
/* ----------------------------------------------- */

typedef long    BIG;

inline  int  convertible (const BIG&) { return 1; };

/* ---------------------------------- */
/* Prototypes of some other functions */
/* ---------------------------------- */

BIG     atob (const char*);                     // String to BIG conversion

char    *btoa (const BIG a, char *s = NULL, int nc = 15);  // And vice versa

#endif

extern         BIG   BIG_0, BIG_1, BIG_inf;     // Some constants of type big

#define TIME_0          BIG_0
#define TIME_1          BIG_1
#define TIME_inf        BIG_inf

#if     BIG_precision != 1
/* -------------------------------------- */
/* Overload "<<" to print out BIG numbers */
/* -------------------------------------- */
ostream &operator<< (ostream&, const BIG&);

/* --------------------------------- */
/* Overload ">>" to read BIG numbers */
/* --------------------------------- */
istream &operator>> (istream&, BIG&);
#endif

/* ---------------------------------------------------------- */
/* Two macros for recognizing infinite (undefined) BIG values */
/* ---------------------------------------------------------- */
#define         def(a)          ((a) != BIG_inf)
#define         undef(a)        ((a) == BIG_inf)

typedef         BIG             TIME;                   // An alias

/* ------------------------------- */
/* Distances, bit counts and rates */
/* ------------------------------- */

#if     ZZ_DST
	// Distances should be of type BIG
typedef BIG     DISTANCE;
#define TYPE_DISTANCE   TYPE_BIG
#define MAX_DISTANCE    BIG_inf         // Maximum distance
#define DISTANCE_inf    BIG_inf         // Maximum distance
#define DISTANCE_0      BIG_0
#define DISTANCE_1      BIG_1
#else
typedef long    DISTANCE;               // Default: type long
#define TYPE_DISTANCE   TYPE_long
#define MAX_DISTANCE    MAX_long
#define DISTANCE_inf    MAX_long        // Maximum distance
#define DISTANCE_0      (0L)
#define DISTANCE_1      (1L)
#endif

#if     ZZ_BTC
	// Counters should be of type BIG
typedef BIG     BITCOUNT;
#define TYPE_BITCOUNT   TYPE_BIG
#define MAX_BITCOUNT    BIG_inf
#define BITCOUNT_inf    BIG_inf
#define BITCOUNT_0      BIG_0
#define BITCOUNT_1      BIG_1
#else
typedef long    BITCOUNT;               // Default: type long
#define TYPE_BITCOUNT   TYPE_long
#define MAX_BITCOUNT    MAX_long
#define BITCOUNT_inf    MAX_long
#define BITCOUNT_0      (0L)
#define BITCOUNT_1      (1L)
#endif

#if     ZZ_RTS
	// Transfer rates should be BIG
typedef BIG     RATE;
#define TYPE_RATE       TYPE_BIG
#define MAX_RATE        BIG_inf
#define RATE_inf        BIG_inf
#define RATE_0          BIG_0
#define RATE_1          BIG_1
#else
typedef long    RATE;                   // Default: type long
#define TYPE_RATE       TYPE_long
#define MAX_RATE        MAX_long
#define RATE_inf        MAX_long
#define RATE_0          (0L)
#define RATE_1          (1L)
#endif

/* ------------------------ */
/* More function prototypes */
/* ------------------------ */

/* ------------------------- */
/* Setting simulation limits */
/* ------------------------- */
void   setLimit (long nm, TIME mt, double ct);
void   setLimit (long nm, TIME mt);
void   setLimit (long nm);

/* ------------------------ */
/* Standard input functions */
/* ------------------------ */
void    readIn (long&);
void    readIn (double&);
#if     BIG_precision != 1
void    readIn (BIG&);
#endif
void    readIn (int&);
void    readIn (float&);

/* ------------------------- */
/* Standard output functions */
/* ------------------------- */
void print (long n, char *h = NULL, int ns = 15, int hs = 0);
inline void print (long n, int ns) { print (n, NULL, ns); };
void print (double n, char *h = NULL, int ns = 15, int hs = 0);
inline void print (double n, int ns) { print (n, NULL, ns); };
#if     ZZ_ATT
inline void print (int n, char *h = NULL, int ns = 15, int hs = 0) {
	print ((long) n, h, ns, hs);
};
inline void print (int n, int ns) { print ((long) n, NULL, ns); };
inline void print (float n, char *h = NULL, int ns = 15, int hs = 0) {
	print ((double) n, h, ns, hs);
};
inline void print (float n, int ns) { print ((double) n, NULL, ns); };
#endif
#if     BIG_precision != 1
void print (const BIG &n, char *h = NULL, int ns = 15, int hs = 0);
inline void print (const BIG &n, int ns) { print (n, NULL, ns); };
#endif
void print (char *n, int ns = 0);

/* ---------------------------------------------------------- */
/* User-invisible global function to issue an inspect request */
/* ---------------------------------------------------------- */
void    zz_inspect_rq (Station*, void*, char*, int, int);

/* ------------------------ */
/* Global display functions */
/* ------------------------ */

/* -------------------------------- */
/* Request display from the program */
/* -------------------------------- */
int     requestDisplay (char *msg = NULL);

/* ------------------ */
/* Refresh the screen */
/* ------------------ */
void    refreshDisplay ();

/* ---------------------------------------------------------- */
/* Add  an  object  to the display server pool of displayable */
/* objects                                                    */
/* ---------------------------------------------------------- */
void    zz_DADD (ZZ_Object*, ZZ_Object*);

/* --------------------------------------------- */
/* Remove an object from the display server pool */
/* --------------------------------------------- */
void    zz_DREM (ZZ_Object*);

/* ------------------------------------------------ */
/* User-callable dynamic display interface routines */
/* ------------------------------------------------ */

char    *displayNote (char*);           // Send a message to display server
void    display (long);                 // Display an integer value
void    display (double);               // Display a fp number
#if     ZZ_ATT
inline  void    display (int a) { display ((long)a); };
inline  void    display (float a) { display ((double)a); };
#endif
#if     BIG_precision != 1
void    display (BIG);                  // Display a BIG number
#endif
void    display (char*);                // Display text
					// Display a single-segment region
void    display (double, double, double, double, long, int, double*, double*);
					// Unscaled single-segment region
void    display (long, int, double*, double*);
void    display (int, double*, double*);// As above with default attributes
					// Start a multi-segment region
void    display (char);                // For a single character

void    startRegion (double, double, double, double);
void    startRegion ();                 // Unscaled multi-segment
void    endRegion ();                   // Terminate a multi-segment region
					// Display a single segment of a msr
void    displaySegment (long, int, double*, double*);
					// The same with default attributes
void    displaySegment (int n, double*, double*);
void    startSegment (long att = NONE); // Start a segment (point mode)
void    displayPoint (double, double);  // Display a single point
void    endSegment ();                  // Close a segment (point mode)

/* ---------------------------------------------------------- */
/* Data   structures   and   intrinsic  functions  for  event */
/* processing                                                 */
/* ---------------------------------------------------------- */

extern  Station                *TheStation;    // Announcements (see below)
extern  Process                *TheProcess;
extern  Process                *ZZ_Main;
extern  ZZ_REQUEST             *zz_c_other;
extern  ZZ_EVENT               *zz_pe;
extern  ZZ_REQUEST             *zz_pr;
extern  long                   zz_npee;
extern  TIME                   Time;
extern  long                   NStations, NLinks, NTraffics;
extern  Link                   **zz_lk;
extern  Station                **zz_st;
extern  Traffic                **zz_tp;
#if     ZZ_OBS
extern  ZZ_INSPECT             *zz_inlist_tail;
extern  Observer               *zz_current_observer;
extern  int                    zz_observer_running;
#endif
extern  int                    TheState;
extern  void                   *Info01, *Info02;
extern  ZZ_EVENT               zz_fsent, zz_rsent, *zz_sentinel_event;
extern  ZZ_REQUEST             *zz_nqrqs;
extern  TIME                   TracingTime;
extern  volatile int           DisplayActive;
extern  int                    zz_flg_otreeAdjusted;
extern  ZZ_SYSTEM              *System;
extern  int                    zz_dispfound;    // A flag for screen exposures
#if     ZZ_ATT
extern  long                   zz_Stroustrup;   // Garbage
#endif

/* ------------------------------------------- */
/* General displayable and/or printable object */
/* ------------------------------------------- */

class   ZZ_Object  {

	friend  class   EObject;
	friend  class   Station;
	friend  class   ZZ_REQUEST;
	friend  class   zz_signal;
	friend  class   ZZ_INSPECT;
	friend  class   Message;
	friend  class   AI;
	friend  class   Traffic;
	friend  class   zz_timer;
	friend  class   zz_client;
	friend  class   zz_queue;
	friend  class   Port;
	friend  class   Link;
	friend  class   Process;
	friend  class   ZZ_KERNEL;
	friend  class   Observer;
	friend  class   RVariable;
#if     ZZ_DBG
	friend  void    zz_print_debug_info ();
#endif
	friend  main (int, char *[]);
	friend  void    zz_adjust_ownership ();
	friend  void    zz_start_client ();
	friend  void    zz_sendTree (ZZ_Object*);
#if     ZZ_SMURPHDEBUG == 2
	friend  void    zz_dumpTree (ZZ_Object*);
#endif
	friend  ZZ_Object *zz_findAdvance (ZZ_Object*);
	friend  ZZ_Object *zz_findWAdvance (ZZ_Object*);
	friend  void    zz_processWindowPhrase ();
	friend  void    zz_DREM (ZZ_Object*);
	friend  void    zz_inspect_rq (Station*, void*, char*, int, int);
	friend  void    terminate (Process*);
	friend  void    terminate (Observer*);

	// All objects whose contents are to be displayable and/or
	// printable in a natural way must be declared as subobjects of
	// type Object.

	protected:

	ZZ_Object       *next,  // Pointers to put the object on the owner's
			*prev;  // list. Note: these attributes must be the
				// first attributes of Object.

	long    Id;             // Numerical identifier
	int     Class;          // Object's class

	public:

	char    *zz_nickname;   // The object's nickname (or NULL)

	ZZ_Object () { zz_nickname = NULL; };

	inline  void    setup () { };

	void    zz_start ();

	// Note: the function zz_start is supposed to play the role of a non-
	//       standard constructor. Regular constructors cannot be used
	//       sometimes, as a virtual function called from a superclass
	//       constructor is taken from that superclass rather than the
	//       subclass.

	inline  long    getId () {
		return (Id);
	};

	inline  int     getClass () {
		return (Class);
	};

	// Exposable
	virtual void zz_expose (int, char *h = NULL, long s = NONE);

	// Returns the object type identifier
	virtual char    *getTName () { return ("object"); };
	
	// Returns the user-assigned nickname of the object
	char    *getNName ();

	// Returns the standard name of the object. This name consists of the
	// object's type name combined with the Id, provided that the Id is not
	// NONE.
	char    *getSName ();

	// Returns a printable name of an object, i.e. the object's nickname, if
	// one is defined, or the standard name, otherwise.
	char    *getOName ();

	// Returns the base standard name of the object. This name is built
	// similarly to the regular standard name with exception that the
	// base type name is used instead of the actual type name. For
	// instance, the base standard name of any station (of a possibly
	// derived type) is "Station nn".
	char    *getBName ();

	inline  void    printOut (int mode, char *hdr = NULL, long sid = NONE) {
		zz_expose (mode, hdr, sid);
	};

	inline  void    printOut (int mode, long sid) {
		zz_expose (mode, NULL, sid);
	};

	inline  void    displayOut (int mode, long sid = NONE) {
		zz_expose (mode+ZZ_EMD_0, NULL, sid);
	};

	int     displayInit ();
};

/* -------------------------------------------------------- */
/* User-visible superclass for defining displayable objects */
/* -------------------------------------------------------- */

class   EObject : public ZZ_Object  {

	private:

	static  long    sernum;                 // Object serial number

	public:
	void zz_start ();                       // Nonstandard constructor
	virtual char *getTName () { return ("EObject"); };
#if     ZZ_ATT
	EObject () { };                         // To keep Cfront happy
#endif
	~EObject ();
};

/* --------- */
/* RVariable */
/* --------- */

class   RVariable : public ZZ_Object {

	friend  class   ZZ_TRVariable;          // Temporary RVariable
						// (used internally)
	private:

	// Exposure functions

	void    exPrint0 (char*);               // Contents
	void    exPrint1 (char*, int);          // Abbreviated/short contents
	void    exDisplay0 (int reg = YES);     // Full contents region optional
	void    exDisplay1 ();                  // Abbreviated contents

	static long sernum;                     // Serial number

	protected:

	long            *s;                     // The storage

	// Note: this type is implemented in a somewhat tricky way -- mainly
	//       for efficiency reasons.

	public:

	// Exposable
	virtual void zz_expose (int, char *h = NULL, long s = NONE);

	char    *getTName () {
		return ("RVariable");
	};

	void zz_start ();                       // Nonstandard constructor

	void setup (int type = TYPE_long, int moments = 2);

#if     ZZ_ATT
	RVariable () { };                       // To keep Cfront happy
#endif
	~RVariable      ();                     // Deallocates memory

	// The following operator intentionally returns nothing -- to avoid
	// problems with unnecessarily invoked destructors
	void operator= (RVariable&);

	void    update (double, long inc = 1);  // Add a sample
	void    calculate (double &min, double &max, double *mom, long &count);

#if     BIG_precision != 1
	// BIG is different from long
	void    update (double, BIG inc);
	void    calculate (double &min, double &max, double *mom, BIG &count);
#endif
	friend  void combineRV (RVariable*, RVariable*, RVariable*);

	inline void     printCnt (char *hd = NULL) {
		// Print contents
		printOut (0, hd);
	};

	inline void     printACnt (char *hd = NULL) {
		// Print abbreviated contents
		printOut (1, hd);
	};

	inline void     printSCnt (char *hd = NULL) {
		// Print short contents
		printOut (2, hd);
	};
};

#if     ZZ_ATT
	// Builder needed
inline  void    zz_bld_RVariable (char *nn = NULL) {
	register RVariable *p;
	zz_COBJ [++zz_clv] = (void*) (p = new RVariable);
	p->zz_start ();
	if (nn != NULL) {
		p->zz_nickname = new char [strlen (nn) + 1];
		strcpy (p->zz_nickname, nn);
	}
}
#endif

/* ------------------ */
/* Station definition */
/* ------------------ */

class   Station : public ZZ_Object {

	friend  class   ZZ_REQUEST;
	friend  class   ZZ_signal;
	friend  class   SGroup;
	friend  class   CGroup;
	friend  class   Traffic;
	friend  class   zz_client;
	friend  class   zz_queue;
	friend  class   zz_signal;
	friend  class   zz_timer;
	friend  class   Packet;
	friend  class   Port;
	friend  class   Systat;         // The system station
	friend  class   AI;
	friend  class   Observer;
	friend  class   Process;
	friend  class   RVariable;
	friend  class   ZZ_SYSTEM;
	friend  class   ZZ_INSPECT;
	friend  class   EObject;
	friend  main (int, char *[]);
	friend  void    zz_adjust_ownership ();
	friend  void    zz_sendTree (ZZ_Object*);
	friend  Port    *idToPort (int);
#if     ZZ_SMURPHDEBUG == 2
	friend  void    zz_dumpTree (ZZ_Object*);
#endif
	friend  ZZ_Object *zz_findAdvance (ZZ_Object*);
	friend  ZZ_Object *zz_findWAdvance (ZZ_Object*);

	public:

	Message         **MQHead,       // Message queue heads
			**MQTail;       // Message queue tails
	private:

	ZZ_PFItem       *Buffers;       // The list of buffers
	Port            *Ports;         // The list of Ports

	ZZ_REQUEST      *CWList,        // Client wait requests
			*SWList;        // Signal wait requests

	ZZ_SIG          *RSList;        // Received signals

	ZZ_Object       *ChList;        // Child list (objects owned)

#if     ZZ_TOL
	double          CTolerance;     // Tolerance parameters
	int             CQuality;
#endif
#if     ZZ_QSL
	long            QSLimit,        // Message queue size limit
			QSize;
#endif
	static long sernum;             // Serial number

	// Exposure functions

	void    exPrint0 (char*);       // Processes
	void    exPrint1 (char*);       // Buffers
	void    exPrint2 (char*);       // Activities

	void    exDisplay0 ();          // Processes
	void    exDisplay1 ();          // Buffers
	void    exDisplay2 ();          // Activities

	static  void    pttrav (ZZ_Object*);    // Recursive printer
	static  void    dttrav (ZZ_Object*);    // Recursive displayer

	public:

	// Exposable
	virtual void zz_expose (int, char *h = NULL, long s = NONE);

	void zz_start ();               // Nonstandard constructor

#if     ZZ_ATT
	Station () { };                 // To keep Cfront happy
#endif
	~Station () {

		excptn ("Station: once created, a station cannot be destroyed");
	};

	void    setQSLimit (long l = MAX_long); // Sets QSLimit (see above)

	INLINE  Port    *idToPort (int);

	inline  void    printBuf (char *hd = NULL) {
		// Print buffers
		printOut (1, hd);
	};

	inline  void    printPrc (char *hd = NULL) {
		// Print processes
		printOut (0, hd);
	};

	inline  void    printAct (char *hd = NULL) {
		// Print activities
		printOut (2, hd);
	};
};

#if     ZZ_ATT
	// Builder needed
inline  void    zz_bld_Station (char *nn = NULL) {
	register Station *p;
	zz_COBJ [++zz_clv] = (void*) (p = new Station);
	p->zz_start ();
	if (nn != NULL) {
		p->zz_nickname = new char [strlen (nn) + 1];
		strcpy (p->zz_nickname, nn);
	}
}
#endif

class   ZZ_REQUEST {

/* ------------------------------------ */
/* Description of a single wait request */
/* ------------------------------------ */

	friend  class   ZZ_EVENT;
	friend  class   ZZ_SIG;
	friend  class   zz_timer;
	friend  class   zz_queue;
	friend  class   zz_signal;
	friend  class   zz_link;
	friend  class   zz_client;
	friend  class   Traffic;
	friend  class   Port;
	friend  class   Link;
	friend  class   Process;
	friend  class   Station;
	friend  class   ZZ_KERNEL;
	friend  main (int, char *[]);
	friend  void    terminate (Process*);

	private:

	ZZ_REQUEST      *next,          // For linking requests at AIs
			*prev,

			*other;         // To link requests at events
	ZZ_EVENT        *event;         // Event backpointer
	void            *what;          // A general purpose fullword
	TIME            when;           // Time when the event will occur
	long            id;             // Link-relative station id
	int             pstate;         // The state
	AI              *ai;            // AI pointer
	long            event_id;       // Event identifier
	void            *Info01,        // Corresponding to Info01 and Info02
			*Info02;        // for event_struct

	// Note: the id attribute for Signal and process requests is
	//       -1 and -2 respectively, to speed up identification
	//       of 'special processing upon release' for these types
	//       of requests

	INLINE ZZ_REQUEST *min_request ();

	public:

	// The constructors

	inline ZZ_REQUEST      (ZZ_REQUEST **q,         // The head
					 AI *a,         // The AI pointer
				      long eid,         // Event identifier
				       int act,         // The state
			       long sid = NONE,         // L-R station id
				     void *in1 = NULL,  // Info 1
				     void *in2 = NULL,  // Info 2
				     void *wht = NULL)  // Aux
	{
		what = wht;
		id = (sid == NONE) ? TheStation->Id : sid;
		pstate = act;
		ai = a;
		event_id = eid;
		Info01 = in1;
		Info02 = in2;
		if (zz_c_other != NULL) zz_c_other -> other = this;
		zz_c_other = this;
		zz_queue_head (this, *q, ZZ_REQUEST);
	};

	inline ZZ_REQUEST      (         AI *a,         // The AI pointer
				      long eid,         // Event identifier
				       int act,         // The state
			       long sid = NONE,         // L-R station id
				     void *in1 = NULL,  // Info 1
				     void *in2 = NULL,  // Info 2
				     void *wht = NULL)  // Aux
	{
		what = wht;
		id = (sid == NONE) ? TheStation->Id : sid;
		pstate = act;
		ai = a;
		event_id = eid;
		Info01 = in1;
		Info02 = in2;
		if (zz_c_other != NULL) zz_c_other -> other = this;
		zz_c_other = this;
		// Use a dummy queue
		zz_queue_head (this, zz_nqrqs, ZZ_REQUEST);
	};
};

class   ZZ_EVENT {

	friend  class   zz_timer;
	friend  class   zz_queue;
	friend  class   zz_signal;
	friend  class   zz_client;
	friend  class   Traffic;
	friend  class   Port;
	friend  class   Link;
	friend  class   Process;
	friend  class   Station;
	friend  class   ZZ_KERNEL;
	friend  class   ZZ_REQUEST;
	friend  class   ZZ_SIG;
	friend  class   Observer;
	friend  class   ObserverService;
	friend  main (int, char *[]);
	friend  void    terminate (Observer*);
	friend  void    terminate (Process*);
	friend  void    zz_print_event_info (ostream&);
	friend  void    zz_noevents ();
#if     ZZ_DBG
	friend  void    zz_print_debug_info ();
#endif

	private:

	ZZ_EVENT        *next,          // To link events into event queue
			*prev;

	ZZ_REQUEST      *chain;         // Request chain
	TIME            waketime;       // Time the event is scheduled at
	Station         *station;       // Station to be awakened or NULL
	Process         *process;       // The station's process
	int             pstate;         // Action triggered
	AI              *ai;            // AI pointer
	long            event_id;       // Event identifier
	void            *Info01,        // Two general-purpose fullwords to
			*Info02;        // return various information

	INLINE  void    enqueue ();
	INLINE  void    enqueue (ZZ_EVENT *h);

	inline  void    store () {

	// Puts (stores) the event at the end of the time queue

		zz_npee++;

		assert (undef (waketime),
			"EVENT->store: definite time event to store");
		zz_rsent.prev->next = this;
		prev = zz_rsent.prev;
		next = &zz_rsent;
		zz_rsent.prev = this;
	};

	inline  void    remove () {

	// Removes event from the queue without purging it

		zz_queue_out (this);
		zz_npee--;
	};

	INLINE  void    cancel ();

	inline  void    reschedule () {

	// Reschedule the event

		remove ();
		enqueue ();
	}

	inline  void    new_top_request (ZZ_REQUEST *r) {

	// Change the current request of the event

		waketime = r -> when;
		pstate   = r -> pstate;
		ai       = r -> ai;
		event_id = r -> event_id;
		Info01   = r -> Info01;
		Info02   = r -> Info02;
		chain    = r;

		reschedule ();
	}

	static  INLINE  ZZ_EVENT  *get ();

	public:                         // Constructors

	inline  ZZ_EVENT () { };        // Dummy constructor

#if     BIG_precision == 1
	inline  ZZ_EVENT (TIME wk,
#else
	inline  ZZ_EVENT (TIME &wk,
#endif
				Station *st,
				void *i_1,
				void *i_2,
				Process *prc,
				AI   *a,
				int  e,
				int  pst,
				ZZ_REQUEST *ch) {

	// Generates a new event


		waketime = wk;
		station  = st;
		Info01   = i_1;
		Info02   = i_2;
		process  = prc;
		ai       = a;
		event_id = e;
		pstate   = pst;
		chain    = ch;

		enqueue ();
	};

	inline  ZZ_EVENT (
				Station *st,
				void *i_1,
				void *i_2,
				Process *prc,
				AI   *a,
				int  e,
				int  pst,
				ZZ_REQUEST *ch) {

	// Generates a new "stored" event with undefined waketime


		waketime = TIME_inf;
		station  = st;
		Info01   = i_1;
		Info02   = i_2;
		process  = prc;
		ai       = a;
		event_id = e;
		pstate   = pst;
		chain    = ch;

		store ();
	};

#if     BIG_precision == 1
	inline  ZZ_EVENT (ZZ_EVENT *h, TIME wk,
#else
	inline  ZZ_EVENT (ZZ_EVENT *h, TIME &wk,
#endif
				Station *st,
				void *i_1,
				void *i_2,
				Process *prc,
				AI   *a,
				int  e,
				int  pst,
				ZZ_REQUEST *ch) {

	// Generates a new event with "hint" h

		waketime = wk;
		station  = st;
		Info01   = i_1;
		Info02   = i_2;
		process  = prc;
		ai       = a;
		event_id = e;
		pstate   = pst;
		chain    = ch;

		enqueue (h);
	};
};

/* --------------------- */
/* Activity interpreters */
/* --------------------- */

/* ------------------------------------------ */
/* Common definitions for AI event processing */
/* ------------------------------------------ */

class   AI : public ZZ_Object { // Activity interpreter

	public:

	// Exposable
	virtual void zz_expose (int, char *h = NULL, long s = NONE);

	char *zz_eid (long);    // Return event identifier
};

/* -------- */
/* Timer AI */
/* -------- */

class   zz_timer : public AI {

	private:

	// Exposure functions
	void exPrint0 (char*, int);
	void exPrint1 (char*, int);

	void exDisplay0 (int);
	void exDisplay1 (int);

	public:

	// Exposable
	virtual void zz_expose (int, char *h = NULL, long s = NONE);

	zz_timer ();    // Constructor

	void zz_wait (TIME, int);
#if     BIG_precision != 1
	void zz_wait (long, int);
#endif
	// Two special cases of wait -- for 0 and 1 ITU
	void zz_proceed (int);
	void zz_skipto (int);

	char    *getTName () { return ("Timer"); };

	inline  void    printRqs (char *hd = NULL, long s = NONE) {
		// Print full request list
		printOut (0, hd, s);
	};

	inline  void    printARqs (char *hd = NULL, long s = NONE) {
		// Print abbreviated request list
		printOut (1, hd, s);
	};
};

extern  zz_timer  zz_AI_timer;                         // Announcement

/* -------- */
/* Queue AI */
/* -------- */

class   zz_queue : public AI {

	private:

	// Exposure funtions

	void    exPrint0 (char*, int);
	void    exPrint1 (char*, int);
	void    exPrint2 (char*, int);

	void    exDisplay0 (int);
	void    exDisplay1 (int);
	void    exDisplay2 (int);

	protected:

	ZZ_QITEM        *queuesh [ZZ_QQSIZE];           // Headers
	ZZ_QITEM        *queuest [ZZ_QQSIZE];           // Trailers
	ZZ_REQUEST      *queuewr [ZZ_QQSIZE];           // Requests

	public:

	// Exposable
	virtual void zz_expose (int, char *h = NULL, long s = NONE);

	void zz_wait (long, int);

	char    *getTName () { return ("Queue"); };

	inline  void    printRqs (char *hd = NULL, long s = NONE) {
		// Print full request list
		printOut (0, hd, s);
	};

	inline  void    printARqs (char *hd = NULL, long s = NONE) {
		// Print abbreviated request list
		printOut (1, hd, s);
	};

	inline  void    printCnt (char *hd = NULL, long s = NONE) {
		// Print Queue contents
		printOut (2, hd, s);
	};

	zz_queue ();            // The constructor

	void    *get (int itype = ANY);

	int     nonempty (int itype = ANY);

	inline  int     empty (int itype = ANY) {
		return (!nonempty (itype));
	};

	int     put (long, void*, int itype = ANY);

	inline  int     put (Station *st, void *it, int itype = ANY) {
		return (put (st->Id, it, itype));
	};

	inline  int     put (void *it, int itype = ANY) {
		return (put (TheStation->Id, it, itype));
	};
};

class ZZ_QITEM {        // Queue item for the Queue AI

	protected:

	ZZ_QITEM        *next, *prev;           // Links
	TIME            RTime;                  // Time when received
	void            *item;                  // The item pointer
	int             id,                     // Station id
			itype;                  // Item type

	public:

	inline ZZ_QITEM (int sid, int qid, void *it) {
		next = NULL;
		RTime = Time;
		item = it;
		id = sid;
		itype = qid;
	};

	friend class zz_queue;
};

/* ----------------------- */
/* Signal AI (definitions) */
/* ----------------------- */

class   zz_signal : public AI {

	private:

	// Exposure functions

	void exPrint0 (char*, int);
	void exPrint1 (char*, int);
	void exPrint2 (char*, int);

	void exDisplay0 (int);
	void exDisplay1 (int);
	void exDisplay2 (int);

	public:

	// Exposable
	virtual void zz_expose (int, char *h = NULL, long s = NONE);

	zz_signal ();           // Constructor

	char    *getTName () { return ("Signal"); };

	void    zz_wait (long, int);

	inline  void    printRqs (char *hd = NULL, long s = NONE) {
		// Print full request list
		printOut (0, hd, s);
	};

	inline  void    printARqs (char *hd = NULL, long s = NONE) {
		// Print abbreviated request list
		printOut (1, hd, s);
	};

	inline  void    printCnt (char *hd = NULL, long s = NONE) {
		// Print Signal list contents
		printOut (2, hd, s);
	};

	int     send (long, Station*, void*, void*);  // Send Signal

	inline  int     send (long sig, long st, void *i1, void *i2) {
		// Station id instead of pointer
		if (!isStationId (st))
			excptn ("Signal->send: illegal station Id");
		return (send (sig, zz_st [st], i1, i2));
	};

	inline  int     send (long sig, long sid) {
		return (send (sig, sid, (void*) sig, (void*) sid));
	};

	inline  int     send (long sig, Station *st) {
		return (send (sig, st, (void*) sig, (void*) (st->Id)));
	};

	inline  int     send (long sig) {
		return (send (sig, TheStation, (void*) sig,
			(void*) (TheStation->Id)));
	};

	int  sendPS (long);                     // Send priority Signal

	int  clear (long);                      // Clear a Signal

	int  clear ();                          // Clear all Signals
};

class   ZZ_SIG {        // Description of a Signal

	friend  main (int, char *[]);

	private:

	ZZ_SIG          *next,
			*prev;                  // For linking

	long            signum;                 // Signal number
	TIME            RTime;                  // Time when received
	ZZ_REQUEST      *request;               // Request pointer

	void            *Info01;                // Additional information
	void            *Info02;

	void            cleanup ();             // Cleanup after removing the
						// Signal
	public:

	ZZ_SIG (long sig, void *i1, void *i2) { // Constructor
		signum = sig;
		RTime = Time;
		Info01 = i1;
		Info02 = i2;
		request = NULL;
	};

	friend  class   zz_signal;
};

/* ------- */
/* Link AI */
/* ------- */

class   Port : public AI {

	friend  class   Link;
	friend  class   zz_client;
	friend  class   Traffic;
	friend  class   Station;
	friend  class   ZZ_SYSTEM;
	friend  ZZ_LINK_ACTIVITY *zz_gen_activity (int, Port*, int,
			Packet *p = (Packet*) NULL);
	friend  class   LinkService;
	friend  void    setD (Port*, Port*, DISTANCE);
	friend  void    zz_adjust_ownership ();
	friend  INLINE  Port *idToPort (int);

	private:

	Port             *nextp;        // Next on the station's list

	DISTANCE         *DV,           // Distance vector to other ports
			 MaxDistance;   // Maximum value in DV

	RATE             TRate;         // Transmission rate
	Link             *Lnk;          // Link pointer
	int              LRId;          // Link-relative port id

	ZZ_LINK_ACTIVITY *Activity,     // Station's activity at the port

			 *Interpolator; // Last-sensed activity interpolator
					// (unidirectional links only)

	TIME             IJTime;        // Estimated time when the interpolator
					// will leave the link 

	// Determines the earliest silence period heard on the port

	TIME    silenceTime ();

	// Determines the earliest collision heard on the port

	TIME    collisionTime (int);

	// Determines when the port will hear the beginning of any
	// activity

	TIME    activityTime ();

	// Determines when the port will hear the nearest beginning of
	// a jamming signal

	TIME    bojTime ();

	// Determines when the port will hear the end of the earliest
	// jamming signal

	TIME    eojTime ();

	// Determines when the port will hear the nearest beginning of a
	// packet

	TIME    botTime ();

	// Determines when the port will hear the nearest end of packet

	TIME    eotTime ();

	// Determines when the port will hear the nearest beginning of
	// a packet addressed to this station

	TIME    bmpTime ();

	// determines when the port will hear the nearest end of a
	// packet addressed to this station

	TIME    empTime ();

	// Determines when the port will perceive the nearest beginning/end
	// of any activity

	TIME    aevTime ();

	// Signals receiver stations upon reception of a packet
	// flagged as RECEIVE_b or RECEIVE_e

	void    port_signal (long, Station*, Packet*);

	// Rescheduling functions

	void    reschedule_act (int);
	void    reschedule_aev (int);
	void    reschedule_mpa ();
	void    reschedule_emp ();
	void    reschedule_col (int);
	void    reschedule_sil (int);
	void    reschedule_boj ();
	void    reschedule_eoj (int);
	void    reschedule_eot ();

	// To update activity interpolator
#if     BIG_precision != TYPE_double
	INLINE
#endif
		ZZ_LINK_ACTIVITY    *initLAI ();

	// A function to find the nearest hint event for the given activity
	INLINE  ZZ_EVENT *findHint (ZZ_LINK_ACTIVITY*);

	// Exposure functions

	void    exPrint0 (char*);       // Request list
	void    exPrint1 (char*);       // Activities
	void    exPrint2 (char*);       // Predicted events

	void    exDisplay0 ();          // Request list
	void    exDisplay1 ();          // Activities
	void    exDisplay2 ();          // Predicted events

	public:

	// Exposable
	virtual void zz_expose (int, char *h = NULL, long s = NONE);

	// Set distance

	void    setDTo (Port*, DISTANCE);
	void    setDTo (Port&, DISTANCE);
	void    setDFrom (Port*, DISTANCE);
	void    setDFrom (Port&, DISTANCE);

	// Connect to a link

	void    connect (Link *lk, int lrid = NONE);
	void    connect (int lk, int lrid = NONE);

	// Start packet transmission

	void    startTransfer (Packet*);

	// Transmit packet

	inline void zz_transmit (Packet*, int);
	inline void zz_transmit (Packet&, int);

	// Convert bits to time

	INLINE TIME bitsToTime (long);

	// Start jamming

	void    startJam ();

	// Terminate transfer or jamming signal

	void    stop ();

	// Abort transfer (or stop jamming signal)

	void    abort ();

	// Port inquiries

	TIME    lastCOL ();
	TIME    lastEOA ();

	inline TIME lastBOS () { return (lastEOA ()); };

	TIME    lastBOT ();
	TIME    lastEOT ();
	TIME    lastBOJ ();
	TIME    lastEOJ ();
	TIME    lastBOA ();

	inline TIME lastEOS () { return (lastBOA ()); };

	int     activities (int&, int&);

	inline int  activities (int &t) {
		int junk;
		return (activities (t, junk));
	};

	inline int  activities () {
		int junk1, junk2;
		return (activities (junk1, junk2));
	};

	inline int  busy () {
		return (undef (lastEOA ()));
	};

	inline int  idle () {
		return (def (lastEOA ()));
	};

	int     events (int);
	Packet  *anotherPacket ();

	// Dumps link activities to the standard output

	inline  void    dump () {
		printOut (1);
	};

	// Constructors

#if     ZZ_ATT
	Port (RATE, char *nn = NULL);
	Port ();
#else
	Port (RATE r = RATE_0, char *nn = NULL);
#endif

	Port (char *nn);                        // Nickname only

	void    zz_start () { };                // Nonstandard constructor

	~Port () {
		excptn ("Port: once created, a port cannot be destroyed");
	};

	char    *getTName () { return ("Port"); };

	// Setup functions

#if     ZZ_ATT
	void    setup (RATE r);
	inline  void setup () { setup (RATE_0); };
#else
	void    setup (RATE r = RATE_0);
#endif
	void zz_wait (int, int);

	inline  void    printRqs (char *hd = NULL) {
		// Print request list
		printOut (0, hd);
	};

	inline  void    printAct (char *hd = NULL) {
		// Print activities
		printOut (1, hd);
	};

	inline  void    printEvs (char *hd = NULL) {
		// Print anticipated event list
		printOut (2, hd);
	};
};

#if     ZZ_ATT
	// Builder needed
inline  void    zz_bld_Port (char *nn = NULL) {
	register Port *p;
	zz_COBJ [++zz_clv] = (void*) (p = new Port);
	p->zz_start ();
	if (nn != NULL) {
		p->zz_nickname = new char [strlen (nn) + 1];
		strcpy (p->zz_nickname, nn);
	}
}
#endif

class   Link : public AI {

	friend  class   Port;
	friend  class   Station;
	friend  class   zz_client;
	friend  class   LinkService;
	friend  class   ZZ_SYSTEM;
	friend  ZZ_LINK_ACTIVITY *zz_gen_activity (int, Port*, int, Packet*);
	friend  void    setD (Port*, Port*, DISTANCE);

	private:

	ZZ_LINK_ACTIVITY  *Alive,               // Alive link activities
			  *Archived,            // Archived activities

			  *AliveTail,           // LT_pointtopoint only
			  *ArchivedTail;        // >= LT_unidirectional

	static long sernum;                     // Serial number

	// Exposure functions

	void    exPrint0 (char*, int);          // Requests
	void    exPrint1 (char*);               // Performance
	void    exPrint2 (char*, int);          // Activities

	void    exDisplay0 (int);               // Requests
	void    exDisplay1 ();                  // Performance
	void    exDisplay2 (int);               // Activities

	public:

	// Exposable
	virtual void zz_expose (int, char *h = NULL, long s = NONE);

	// Tells whether standard performance measures are to be
	// calculated

	char            FlgSPF;                 // = ON

	int             Type,                   // = LT_broadcast;

	// Note: during the link initialization phase (while not all of its
	//       ports are defined), NAlive contains -1 to indicate the
	//       incomplete status of the link. During that phase, Alive
	//       stores the pointer to the link's distance matrix. This
	//       matrix is later turned into port distance vectors and
	//       deallocated.
	//       During dynamic initialization of the distance matrix,
	//       NAliveTransfers stores the last "i" position (row) and
	//       NAliveJams -- the last "j" (column) position.
	//       NArchived tells the number of ports already defined

			NPorts,                 // Number of ports

			NAlive,                 // Number of alive activities
			NAliveTransfers,        // Number of alive transfers
			NAliveJams,             // Number of alive jams

			NArchived,              // Number of archived activities
			NArchivedTransfers,     // Number of archived transfers
			NArchivedJams;          // Number of archived jams

	long            NTotalJams,             // Total number of jams
			NTotalTAttempts,        // ... of transfer attempts
			NTotalTransfers;        // ... of successful transfers

	BITCOUNT        NRBits,                 // Total bits received
			NTBits;                 // Total bits transmitted

	long            NRPackets,              // Total packets received
			NTPackets,              // Total packets transmitted

			NRMessages,             // Total messages received
			NTMessages;             // Total messages transmitted

	TIME            ArchiveTime;            // Archiving period

	ZZ_REQUEST      *RQueue [ZZ_N_LINK_EVENTS];     // Request queues

	void zz_start ();                       // Nonstandard constructor

#if     ZZ_ATT
	Link () { };                            // To keep Cfront happy
	void setup (long np, TIME at, int spf = ON);
	inline void setup (long np) { setup (np, TIME_0); };
#else
	void setup (long np, TIME at = TIME_0, int spf = ON);
#endif
	~Link () {
		excptn ("Link: once created, a link cannot be destroyed");
	};

	virtual char* getTName () { return ("Link"); };

	inline  void    printRqs (char *hd = NULL, long s = NONE) {
		// Print request list
		printOut (0, hd, s);
	};

	inline  void    printPfm (char *hd = NULL) {
		// Print performance measures
		printOut (1, hd);
	};

	inline  void    printAct (char *hd = NULL, long s = NONE) {
		// Print activities
		printOut (2, hd, s);
	};

	// Set one element in the distance matrix
	void setD     (int, int, DISTANCE);
	void setDFrom (int, DISTANCE);
	void setDTo   (int, DISTANCE);

	// User-definable extensions

	virtual void    pfmMTR (Packet*) {};
	virtual void    pfmMRC (Packet*) {};
	virtual void    pfmPTR (Packet*) {};
	virtual void    pfmPRC (Packet*) {};
	virtual void    pfmPAB (Packet*) {};

	private:

	// Link performance measuring functions

	// Standard functions

	inline void     stdpfmMTR (Packet *p);
	inline void     stdpfmMRC (Packet *p);
	inline void     stdpfmPTR (Packet *p);
	inline void     stdpfmPRC (Packet *p);
	inline void     stdpfmPAB (Packet *p);

	// Called each time the last packet of a message is transmitted

	inline  void    spfmMTR (Packet *p) {
		if (FlgSPF == ON)
			stdpfmMTR (p);
		pfmMTR (p);
	};

	// Called each time a packet is transmitted

	inline  void    spfmPTR (Packet *p) {
		if (FlgSPF == ON)
			stdpfmPTR (p);
		pfmPTR (p);
	};

	// Called each time a packet is received

	inline  void    spfmPRC (Packet *p) {
		if (FlgSPF == ON)
			stdpfmPRC (p);
		pfmPRC (p);
	};

	// Called each time a packet is aborted

	inline  void    spfmPAB (Packet *p) {
		if (FlgSPF == ON)
			stdpfmPAB (p);
		pfmPAB (p);
	};

	// Called each time a message is received

	inline  void    spfmMRC (Packet *p) {
		if (FlgSPF == ON)
			stdpfmMRC (p);
		pfmMRC (p);
	};
};

/* ---------------------------------------------------------- */
/* Link-independent  functions  for  defining  link  distance */
/* matrices. These functions operate directly on ports.       */
/* ---------------------------------------------------------- */

void setD     (Port*, Port*, DISTANCE);
void setD     (Port&, Port&, DISTANCE);
void setDFrom (Port*, DISTANCE);
void setDFrom (Port&, DISTANCE);
void setDTo   (Port*, DISTANCE);
void setDTo   (Port&, DISTANCE);

/* ------------------- */
/* Standard link types */
/* ------------------- */
#if     ZZ_ATT
	// Builder needed
inline  void    zz_bld_Link (char *nn = NULL) {
	register Link *p;
	zz_COBJ [++zz_clv] = (void*) (p = new Link);
	p->zz_start ();
	if (nn != NULL) {
		p->zz_nickname = new char [strlen (nn) + 1];
		strcpy (p->zz_nickname, nn);
	}
}
#endif

/* -------------------------- */
/* A broadcast link (default) */
/* -------------------------- */
class   BLink : public Link {
	public:
	virtual char *getTName () { return ("BLink"); };
};
#if     ZZ_ATT
	// Builder needed
inline  void    zz_bld_BLink (char *nn = NULL) {
	register BLink *p;
	zz_COBJ [++zz_clv] = (void*) (p = new BLink);
	p->zz_start ();
	if (nn != NULL) {
		p->zz_nickname = new char [strlen (nn) + 1];
		strcpy (p->zz_nickname, nn);
	}
}
#endif

/* --------------------- */
/* A unidirectional link */
/* --------------------- */
class   ULink : public Link {
	public:
	virtual char *getTName () { return ("ULink"); };
#if     ZZ_ATT
	inline void setup (long np, TIME at, int spf = ON) {
		Type = LT_unidirectional;
		Link::setup (np, at, spf);
	};
	inline void setup (long np) { setup (np, TIME_0); };
#else
	inline void setup (long np, TIME at = TIME_0, int spf = ON) {
		Type = LT_unidirectional;
		Link::setup (np, at, spf);
	};
#endif
};
#if     ZZ_ATT
	// Builder needed
inline  void    zz_bld_ULink (char *nn = NULL) {
	register ULink *p;
	zz_COBJ [++zz_clv] = (void*) (p = new ULink);
	p->zz_start ();
	if (nn != NULL) {
		p->zz_nickname = new char [strlen (nn) + 1];
		strcpy (p->zz_nickname, nn);
	}
}
#endif

/* ------------------------------------ */
/* A unidirectional point-to-point link */
/* ------------------------------------ */
class   PLink : public Link {
	public:
	virtual char *getTName () { return ("PLink"); };
#if     ZZ_ATT
	inline void setup (long np, TIME at, int spf = ON) {
		Type = LT_pointtopoint;
		Link::setup (np, at, spf);
	};
	inline void setup (long np) { setup (np, TIME_0); };
#else
	inline void setup (long np, TIME at = TIME_0, int spf = ON) {
		Type = LT_pointtopoint;
		Link::setup (np, at, spf);
	};
#endif
};
#if     ZZ_ATT
	// Builder needed
inline  void    zz_bld_PLink (char *nn = NULL) {
	register PLink *p;
	zz_COBJ [++zz_clv] = (void*) (p = new PLink);
	p->zz_start ();
	if (nn != NULL) {
		p->zz_nickname = new char [strlen (nn) + 1];
		strcpy (p->zz_nickname, nn);
	}
}
#endif

/* ---------------------------------------------- */
/* A broadcast link in which collisions propagate */
/* ---------------------------------------------- */
class   CLink : public Link {
	public:
	virtual char *getTName () { return ("CLink"); };
#if     ZZ_ATT
	inline void setup (long np, TIME at, int spf = ON) {
		Type = LT_cpropagate;
		Link::setup (np, at, spf);
	};
	inline void setup (long np) { setup (np, TIME_0); };
#else
	inline void setup (long np, TIME at = TIME_0, int spf = ON) {
		Type = LT_cpropagate;
		Link::setup (np, at, spf);
	};
#endif
};
#if     ZZ_ATT
	// Builder needed
inline  void    zz_bld_CLink (char *nn = NULL) {
	register CLink *p;
	zz_COBJ [++zz_clv] = (void*) (p = new CLink);
	p->zz_start ();
	if (nn != NULL) {
		p->zz_nickname = new char [strlen (nn) + 1];
		strcpy (p->zz_nickname, nn);
	}
}
#endif

/* ----------------------------------- */
/* Type of test function for getPacket */
/* ----------------------------------- */
typedef         int(*ZZ_MTTYPE)(Message*);

/* --------- */
/* Client AI */
/* --------- */

class   zz_client : public AI {

	private:

	// Exposure functions

	void exPrint0 (char*, int);     // Request list
	void exPrint1 (char*);          // Performance measures
	void exPrint2 (char*, int);     // Contents (message queues)
	void exPrint3 (char*);          // Definitions (traffic)

	void exDisplay0 (int);          // Request list
	void exDisplay1 ();             // Performance measures
	void exDisplay2 (int);          // Contents (message queues)
	void exDisplay3 ();             // Sent-received statistics

	public:

	// Exposable
	virtual void zz_expose (int, char *h = NULL, long s = NONE);

	zz_client ();           // Constructor

	char    *getTName () { return ("Client"); };

	void zz_wait (int, int);

	inline  void    printRqs (char *hd = NULL, long s = NONE) {
		// Print request list
		printOut (0, hd, s);
	};

	inline  void    printPfm (char *hd = NULL) {
		// Print performance measures
		printOut (1, hd);
	};

	inline  void    printCnt (char *hd = NULL, long s = NONE) {
		// Print queues
		printOut (2, hd, s);
	};

	inline  void    printDef (char *hd = NULL) {
		// Print definition
		printOut (3, hd);
	};

	// To be called upon reception of a packet

	void receive (Packet *p, Link *l = (Link*) NULL);

	inline void receive (Packet &p, Link &l) {
		receive (&p, &l);
	};

	inline void receive (Packet *p, Port *pp) {
		receive (p, pp->Lnk);
	};

	inline void receive (Packet &p, Port &pp) {
		receive (&p, (&pp)->Lnk);
	};

	// Client (global) versions of getPacket

	int getPacket (Packet *p, long mi = 0, long mx = 0,
		    long fr = 0);

	inline int getPacket (Packet &p, long mi = 0, long mx = 0,
		    long fr = 0);

	int getPacket (Packet *p, ZZ_MTTYPE, long mi = 0, long mx = 0,
		    long fr = 0);

	inline int getPacket (Packet &p, ZZ_MTTYPE, long mi = 0, long mx = 0,
		    long fr = 0);

	// Checks if there is a packet of any type to be acquired from the
	// Client

	int isPacket ();
	int isPacket (ZZ_MTTYPE);

	// Releases a packet that has been successfully transmitted

	void release (Packet *p);

	inline void release (Packet &p);
};

extern  zz_client  zz_AI_client;         // Announcement

class   SGroup {                // A station group

	friend  class   zz_client;
	friend  class   Traffic;
	friend  class   CGroup;
	friend  class   Port;
	friend  void    zz_sort_group (SGroup*, float* = NULL);
	friend  void    zz_start_client ();

	private:

	int     NStations;      // The number of stations in the group
	short   *SIdents;       // The array of station identifiers
	char    sorted;         // YES, if the group is sorted

	public:

	void    zz_start () { sorted = NO; };
	void    setup (int, int*);      // List of stations given explicitely

	void    setup () {              // All stations
	
		NStations = (int) ::NStations;
		SIdents = NULL;
	};

	void    setup (int, int, ...);  // Explicit list of stations
	void    setup (int, Station*, ...);

#if     ZZ_ATT
	SGroup () { };                  // To keep Cfront happy
#endif
	~SGroup () {

		excptn ("SGroup: once created, cannot be destroyed");
	};

	// Checks if a givent station occurs within the group
	int     occurs (long);
	inline  int occurs (Station *s) {
		return (occurs (s->Id));
	};
};

#if     ZZ_ATT
	// Builder needed
inline  void    zz_bld_SGroup () {
	register SGroup *p;
	zz_COBJ [++zz_clv] = (void*) (p = new SGroup);
	p->zz_start ();
}
#endif

class   CGroup {                // A communication group

	friend  class   zz_client;
	friend  class   Traffic;
	friend  void    zz_start_client ();

	private:

	SGroup  *Senders,       // Legitimate transmitters
		*Receivers;     // Legitimate receivers

	float   *SWeights,      // Sender weights
		*RWeights,      // Receiver weights

		CGWeight;       // Group weight (to speed up things a little)

	public:

	// The most general case
	void    setup (SGroup*, float*, SGroup*, float*);

	// No receiver weights
	void    setup (SGroup *snd, float *snw, SGroup *rcv) {
		setup (snd, snw, rcv, NULL);
	};

	// No sender weights (receiver weights given)
	void    setup (SGroup *snd, SGroup *rcv, float *rcw) {
		setup (snd, (float*) NULL, rcv, rcw);
	};

	// No weights at all
	void    setup (SGroup *snd, SGroup *rcv) {
		setup (snd, (float*) NULL, rcv, (float*) NULL);
	};

	// Broadcast traffic (the last parameter can only be GT_broadcast)
	void    setup (SGroup*, float*, SGroup*, int);

	// Broadcast without sender weights
	void    setup (SGroup *snd, SGroup *rcv, int b) {
		setup (snd, (float*) NULL, rcv, b);
	};

	void    zz_start () { };

#if     ZZ_ATT
	CGroup () { };          // To keep Cfront happy
#endif
	~CGroup () {

		excptn ("CGroup: once created, cannot be destroyed");
	};
};

#if     ZZ_ATT
	// Builder needed
inline  void    zz_bld_CGroup () {
	register CGroup *p;
	zz_COBJ [++zz_clv] = (void*) (p = new CGroup);
	p->zz_start ();
}
#endif

class   Message {                       // Message skeleton

	friend  class   Traffic;
	friend  class   zz_client;
	friend  class   Packet;

	public:

	Message         *next, *prev;   // For putting messages into queue

	long            Receiver;       // The recipient
	TIME            QTime;          // The time the message was queued
	long            Length;         // The number of (information) bits
	int             TP;             // Traffic pattern identifier

	virtual void setup () {
		// Void -- to be provided by the user (in the unlikely case
		// he wants to generate messages by hand)
	};

	virtual int  frameSize () {
		// Returns the size of the subtype
		return (sizeof (Message));
	};

	// The function below is defined for compatibility. Nobody sane
	// will 'create' messages by hand
	void    zz_start () { TP = NONE; };     // See Packet

	inline  void operator= (Message &m) {
		// Intentionally void -- just in case
		bcopy ((void*)(&m), (void*)(this), m.frameSize ());
	};

	// Acquires a packet from this specific message

	int getPacket (Packet *p, long mi = 0, long mx = 0, long fr = 0);

	inline int      getPacket (Packet &p, long mi = 0, long mx = 0,
		long fr = 0) {

		return (getPacket (&p, mi, mx, fr));
	};
};

#if     ZZ_ATT
	// Builder needed
inline  void    zz_bld_Message () {
	register Message *p;
	zz_COBJ [++zz_clv] = (void*) (p = new Message);
	p->zz_start ();
}
#endif

class   Packet  {                       // Packet skeleton

	public:

	TIME            QTime;          // Message queue time
	TIME            TTime;          // The time the packet appeared on
					// top of the message queue
	long            Sender,         // The transmitter
			Receiver;       // The receiver

	int             TP;             // Qualification
	FLAGS           Flags;          // E.g. the last packet of message

	long            ILength;        // The length of the information part
	long            TLength;        // Total length (incl. hdr+trailer)

	// A nasty little trick. zz_setptr is used to reference the location
	// containing pointer to the virtual function table. Do not add any
	// new attributes to 'Packet' past this point.

#define zz_setptr(p)    (*(char**)((&((p)->TLength))+1))

	// Don't put anything between zz_setptr and setup

	// For copying non-standard message information (from Message subtypes)
	// to user-defined Packet subtypes

	virtual void setup (Message *m); // From message
	virtual void setup () { };	 // Raw

	virtual int  frameSize () {
		// Returns the size of the subtype
		return (sizeof (Packet));
	};

	inline  void setup (Message &m) { setup (&m); };

	// Checks if the packet is addressed to this station
	int     isMy ();

	// The packet version of release
	inline  void    release () {
		Client->release (this);
	};

	// Converts standard flags to printable format
	char    *zz_pflags ();

	// The following function is defined for 'create'. A packet created
	// this way is assumed to be nonstandard and its TP attribute is
	// set to NONE
	void    zz_start () { TP = NONE; zz_mkpclean (); };

	inline  void operator= (Packet &p) {
		// Intentionally void -- just in case
		bcopy ((void*)(&p), (void*)(this), p.frameSize ());
	};

	void    fill (Station*, Station*, long, long il = NONE, long tp = NONE);
	void    fill (long, long, long, long il = NONE, long tp = NONE);

	// Constructor for creating statically-declared packet buffers
	Packet ();

	// Methods to check the packet status

	inline  int     isFull () { return (flagSet (Flags, PF_full)); };
	inline  int     isEmpty () { return (flagCleared (Flags, PF_full)); };
	inline  int     isNonstandard () { return (!isTrafficId (TP)); };
	inline  int     isStandard () { return (isTrafficId (TP)); };
	inline  int     isLast () { return (flagSet (Flags, PF_last)); };
	inline  int     isNonlast () { return (flagCleared (Flags, PF_last)); };
	inline  int     isBroadcast () {
		return (flagSet (Flags, PF_broadcast));
	};
	inline  int     isNonbroadcast () {
		return (flagCleared (Flags, PF_broadcast));
	};
};

#if     ZZ_ATT
	// Builder needed
inline  void    zz_bld_Packet () {
	register Packet *p;
	zz_COBJ [++zz_clv] = (void*) (p = new Packet);
	p->zz_start ();
}
#endif

class   Traffic : public AI {

	public:

	// Exposable
	virtual void zz_expose (int, char *h = NULL, long s = NONE);

	friend  class   zz_client;
	friend  class   zz_client_service;
	friend  class   ZZ_LINK_ACTIVITY;
	friend  ZZ_LINK_ACTIVITY *zz_gen_activity (int, Port*, int, Packet *p);
	friend  class   Message;
	friend  class   Packet;
	friend  void    zz_adjust_ownership ();
	friend  void    zz_start_client ();
	friend  void    zz_sendTree (ZZ_Object*);
#if     ZZ_SMURPHDEBUG == 2
	friend  void    zz_dumpTree (ZZ_Object*);
#endif
	friend  ZZ_Object *zz_findAdvance (ZZ_Object*);
	friend  ZZ_Object *zz_findWAdvance (ZZ_Object*);

	// Flags

	char    DstMIT,         // Message interarrival time distribution
		DstMLE,         // Message length distribution
		DstBIT,         // Burst interarrival time distribution
		DstBSI,         // Burst size distribution

		FlgSCL,         // Standard Client ON/OFF
		FlgSPF;         // Standard performance measures ON/OFF

	// Note: A Dst (distribution flag) can be set to:
	//
	//              UNDEFINED       - (inconsitencies detected in run time)
	//              EXPONENTIAL     - Poisson distribution
	//              UNIFORM         - easy to guess
	//
	// More distribution types will be added in the future

	double  ParMnMIT,       // Min (or Mean) message interarrival time
		ParMxMIT,       // Max message interarrival time
		ParMnMLE,       // Min (or Mean) message length
		ParMxMLE,       // Max message length
		ParMnBIT,       // Min (or Mean) burst interarrival time
		ParMxBIT,       // Max burst interarrival time
		ParMnBSI,       // Min (or Mean) burst size
		ParMxBSI;       // Max burst size

	int     NCGroups;       // The number of communication groups
	CGroup  **CGCluster;    // The cluster of communication groups
	double  TCGWeight;      // Total sender weight

	// Note: the LANSF concept of a "flood" group and cluster has been
	//       abandoned. Nobody could find a reasonable application for
	//       it. Anyway, now it would not be difficult to define it in
	//       a non-standard way.

	// Slots for standard RVariables used to calculate standard
	// performance measures

	RVariable      *RVAMD,  // Absolute message delay
		       *RVAPD,  // Absolute packet delay
		       *RVWMD,  // Weighted message delay
		       *RVMAT,  // Message access time
		       *RVPAT,  // Packet access time
		       *RVMLS;  // Message length statistics

	// The following functions are declared as virtual. This way the
	// user is able to redefine them in his/her own favorite way

	virtual TIME    genMIT ();      // Generate message interarrival time
	virtual long    genMLE ();      // Generate message length
	virtual TIME    genBIT ();      // Generate burst interarrival time
	virtual long    genBSI ();      // Generate burst size
	virtual int     genSND ();      // Generate sender
	virtual long    genRCV ();      // Generate receiver

	// Functions for dynamic initialization of communication groups

	void    addSender (Station *s = ALL, double w = 1.0, int gr = 0);
	void    addSender (long s, double w = 1.0, int gr = 0);

	void    addReceiver (Station *s = ALL, double w = 1.0, int gr = 0);
	void    addReceiver (long s, double w = 1.0, int gr = 0);

	// Generate a communication group including a given sender. The
	// group is selected according to its probability of being used
	// to generate the specified sender.

	virtual CGroup *genCGR (long);

	inline  CGroup *genCGR (Station *s) {
		return (genCGR (s->Id));
	};

	// Generate message and queue it at the sender

	virtual Message *genMSG (long,  // The sender identifier
				long,   // The receiver identifier
				long);  // The message length

	inline  Message *genMSG (long s, SGroup *r, long l) {
		return (genMSG (s, (long)r, l));
	};

	inline  Message *genMSG (Station *s, Station *r, long l) {
		return (genMSG (s->Id, r->Id, l));
	};

	inline  Message *genMSG (Station *s, SGroup *r, long l) {
		return (genMSG (s->Id, (long)r, l));
	};

	void    zz_wait (int, int);      // Wait request


	inline  void    printRqs (char *hd = NULL, long s = NONE) {
		// Print request list
		printOut (0, hd, s);
	};

	inline  void    printPfm (char *hd = NULL) {
		// Print performance measures
		printOut (1, hd);
	};

	inline  void    printCnt (char *hd = NULL, long s = NONE) {
		// Print queues
		printOut (2, hd, s);
	};

	inline  void    printDef (char *hd = NULL) {
		// Print definition
		printOut (3, hd);
	};

	virtual char   *getTName () {
		return ("Traffic");
	};

	private:

	// A preprocessor for addSender and addReceiver

	void    zz_addinit (int);
#if     ZZ_QSL
	long    QSLimit,                        // Pending message count limit
		QSize;                          // Current count
#endif
	ZZ_Object  *ChList;                     // For RVariables owned

	// Exposure functions

	void    exPrint0 (char*, int);          // Request list
	void    exPrint1 (char*);               // Performance
	void    exPrint2 (char*, int);          // Contents (message queues)
	void    exPrint3 (char*);               // Traffic pattern definition

	void    exDisplay0 (int);               // Request list
	void    exDisplay1 ();                  // Performance
	void    exDisplay2 (int);               // Message queues
	void    exDisplay3 ();                  // Sent-received statistics

	// This junk should be invisible

	virtual Message *zz_mkm () {            // Create a (subclass) message
	    return ((Message*)(Info01 = (void*) new Message));
				// TheMessage
	};

	virtual Packet  *zz_mkp ();             // Create a (subclass) packet
						// used for tricks only

	// Pointer to the virtual function table of the associated packet type.
	// We play some tricks here. This pointer is used to initialize 
	// packets that are created dynamically in a non-standard way.
	char    *ptviptr;

	static long sernum;             // Serial number

	long    last_sender;            // To avoid generating receiver the
	CGroup  *last_CGroup;           // same as sender

	// Called each time a message is queued at a station

	inline  void    spfmMQU (Message *m) {
		if (FlgSPF == ON)
			stdpfmMQU (m);  // The standard part
		pfmMQU (m);             // The user-definable part
	};

	// Called each time a full message is transmitted

	inline  void    spfmMTR (Packet *p) {
		if (FlgSPF == ON)
			stdpfmMTR (p);
		pfmMTR (p);
	};

	// Called each time a full message is received

	inline  void    spfmMRC (Packet *p) {
		if (FlgSPF == ON)
			stdpfmMRC (p);
		pfmMRC (p);
	};

	// Called each time a packet is transmitted

	inline  void    spfmPTR (Packet *p) {
		if (FlgSPF == ON)
			stdpfmPTR (p);
		pfmPTR (p);
	};

	// Called each time a packet is received

	inline  void    spfmPRC (Packet *p) {
		if (FlgSPF == ON)
			stdpfmPRC (p);
		pfmPRC (p);
	};

	// Standard performance measuring functions

	void    stdpfmMQU (Message*);
	void    stdpfmMTR (Packet*);
	void    stdpfmMRC (Packet*);
	void    stdpfmPTR (Packet*);
	void    stdpfmPRC (Packet*);

	// To convert absolute weights into accumulated weights
	void    preprocess_weights ();

	public:

	// User-definable extensions

	virtual void    pfmMQU (Message*) {};
	virtual void    pfmMTR (Packet*) {};
	virtual void    pfmMRC (Packet*) {};
	virtual void    pfmPTR (Packet*) {};
	virtual void    pfmPRC (Packet*) {};

	void setQSLimit (long l = MAX_long);    // Sets QSLimit (see above)

	// Counters

	long            NQMessages,     // Number of queued messages
			NTMessages,     // Number of transmitted messages
			NRMessages,     // Number of received messages
			NTPackets,      // Number of transmitted packets
			NRPackets;      // Number of received packets

	BITCOUNT        NQBits,         // Number of queued bits
			NTBits,         // Number of transmitted bits
			NRBits;         // Number of received bits

	// Traffic version of receive (checks against type consistency)

	inline void receive (Packet *p, Link *l = (Link*) NULL);

	inline void receive (Packet &p, Link &l) {
		receive (&p, &l);
	};

	inline void receive (Packet &p) {
		receive (&p);
	};

	inline void receive (Packet *p, Port *pp) {
		receive (p, pp->Lnk);
	};

	inline void receive (Packet &p, Port &pp) {
		receive (&p, (&pp)->Lnk);
	};

	// Gets a packet from a message queue and puts it into a buffer.
	// Returns NO, if there is no packet to get

	int     getPacket (Packet *b, long mi = 0, long mx = 0, long fr = 0);

	inline  int     getPacket (Packet &p, long mi = 0, long mx = 0,
		    long fr = 0) {

		return (getPacket (&p, mi, mx, fr));
	};

	int     getPacket (Packet *b, ZZ_MTTYPE, long mi = 0, long mx = 0,
		    long fr = 0);

	inline  int     getPacket (Packet &p, ZZ_MTTYPE tf, long mi = 0,
		    long mx = 0, long fr = 0) {

		return (getPacket (&p, tf, mi, mx, fr));
	};

	// Checks if there is a packet of "this" type to be acquired from
	// the Client

	int isPacket ();
	int isPacket (ZZ_MTTYPE);

	// Member version of release (checks against type consistency)

	inline void release (Packet *p);

	inline void release (Packet &p) {
		release (&p);
	};

	long    RemBSI;         // Residual burst size


	void    setup (CGroup**, int, int, ...);

	//      Arguments:
	//
	//      1. The array of communication groups
	//      2. The number of communication groups
	//      3. The flag interpreted as sum of the following
	//              
	//              MIT_exp
	//              MIT_unf
	//              MLE_exp
	//              MLE_unf
	//              BIT_exp
	//              BIT_unf
	//              BSI_exp
	//              BSI_unf
	//              SCL_on
	//              SCL_off
	//              SPF_on
	//              SPF_off
	//
	//      The remaining arguments are double precision floating
	//      point numbers in the order of declaration of "ParXXX"
	//      variables (see above). Only the relevant values are
	//      expected and should be specified
	//

	// An alias for a single communication group

	void    setup (CGroup*, int, ...);

	// An alias for a nonstandard pattern

	void    setup (int, ...);

	void    zz_start ();                    // Nonstandard constructor

#if     ZZ_ATT
	Traffic () { };                         // To keep Cfront happy
#endif
	~Traffic () {

		excptn ("Traffic: once created, cannot be destroyed");
	};
};

#if     ZZ_ATT
	// Builder needed
inline  void    zz_bld_Traffic (char *nn = NULL) {
	register Traffic *p;
	zz_COBJ [++zz_clv] = (void*) (p = new Traffic);
	p->zz_start ();
	if (nn != NULL) {
		p->zz_nickname = new char [strlen (nn) + 1];
		strcpy (p->zz_nickname, nn);
	}
}
#endif

class   ZZ_LINK_ACTIVITY {

	friend  class   Port;
	friend  class   Link;
	friend  class   Station;
	friend  ZZ_LINK_ACTIVITY *zz_gen_activity (int, Port*, int, Packet *p);
	friend  class   LinkService;

	private:

	ZZ_LINK_ACTIVITY   *next,               // Next on list
			   *prev;               // Previous on list

	int                LRId;                // Link-relative generator id
	Port               *GPort;              // Generator port
	TIME               STime,               // Time when started
			   FTime;               // Time when finished
	int                Type;                // Transfer, jam, etc.
	ZZ_EVENT           *ae;                 // Archiving event pointer
	Packet             Pkt;                 // Information buffer

	// Note: when the activity is generated, the amount of memory
	// needed is determined by the actual size of the packet data
	// structure, possibly extended by the user

	public:

	ZZ_LINK_ACTIVITY  () {
		excptn ("LINK_ACTIVITY: cannot create directly");
	};
};

/* ------------------ */
/* Process definition */
/* ------------------ */

class   Process : public AI {

	friend  class   ZZ_Object;
	friend  class   EObject;
	friend  class   Observer;
	friend  class   Station;
	friend  class   RVariable;
	friend  class   Traffic;
	friend  class   Link;
	friend  class   Port;
	friend  class   ZZ_KERNEL;
	friend  class   ZZ_INSPECT;
	friend  main (int, char *[]);
	friend  void    zz_adjust_ownership ();
	friend  void    zz_sendTree (ZZ_Object*);
#if     ZZ_SMURPHDEBUG == 2
	friend  void    zz_dumpTree (ZZ_Object*);
#endif
	friend  ZZ_Object *zz_findAdvance (ZZ_Object*);
	friend  ZZ_Object *zz_findWAdvance (ZZ_Object*);

	private:

	static long sernum;             // Serial number

	Station         *Owner;         // The station the proceess runs at
	ZZ_REQUEST      *TWList,        // Processes waiting for termination
			*SWList;        // And for states
	ZZ_Object       *ChList;        // Objects owned by the process
	long            Ret1, Ret2;     // To store return values

	// Exposure functions

	void    exPrint0 (char*, int);  // Request list
	void    exPrint1 (char*);       // Wait list

	void    exDisplay0 (int);       // Request list
	void    exDisplay1 ();          // Wait list

	public:

	// Exposable
	virtual void zz_expose (int, char *h = NULL, long s = NONE);

	void    zz_start ();            // The constructor

	// The address of zz_"typename"_prcs, where "typename" is the name
	// of the process class uniquely identifies the
	// process type and is required to implement observers.
	// This address is kept in the process attribute zz_typeid

	static int zz_Process_prcs;     // Equivalent to ANY

	void    *zz_typeid;

	// The process code

	virtual void zz_code () {
		excptn (form ("Process: process %s code undefined",
			getTName ()));
	};

	// Returns the name of the state number i
	virtual char *zz_sn (int);

	void    zz_wait (int, int);

	inline  void    rtn (long a, long b = 0L) {
		// Returns values to the caller
		Ret1 = a;
		Ret2 = b;
	};

	friend void terminate (Process *p);

	inline void terminate () {
		::terminate (this);
	};

	inline  void    printRqs (char *hd = NULL, long sid = NONE) {
		// Print request list
		printOut (0, hd, sid);
	};

	inline  void    printWai (char *hd = NULL) {
		// Print wait requests of this process
		printOut (1, hd);
	};
};

#if     ZZ_ATT
	// Builder needed
inline  void    zz_bld_Process (char *nn = NULL) {
	register Process *p;
	zz_COBJ [++zz_clv] = (void*) (p = new Process);
	p->zz_start ();
	if (nn != NULL) {
		p->zz_nickname = new char [strlen (nn) + 1];
		strcpy (p->zz_nickname, nn);
	}
}
#endif

/* --------- */
/* Observers */
/* --------- */

class   ZZ_INSPECT      {

/* --------------------------------------- */
/* Description of a single inspect request */
/* --------------------------------------- */

	friend  class   Observer;
	friend  class   ObserverService;
	friend  void    terminate (Observer*);
	friend  main (int, char *[]);

	ZZ_INSPECT      *next;          // To link inspect requests into lists
	Station         *station;       // The inspected station
	void            *typeid;        // Process type identifier

	char            *nickname;      // Process nickname

	int             pstate,         // Process state
			ostate;         // Observer's state

	char            *getptype (),   // Return process type name
			*pttrav (ZZ_Object*); // A recursive helper for getptype
	public:

	inline ZZ_INSPECT (Station*, void*, char*, int, int);
};

class   Observer : public AI {

/* ------------------ */
/* The observer frame */
/* ------------------ */

	friend  class   ZZ_INSPECT;
	friend  class   ZZ_Object;
	friend  class   EObject;
	friend  class   RVariable;
	friend  class   ObserverService;
	friend  main (int, char *[]);
	friend  void    zz_timeout_rq (TIME, int);
	friend  void    zz_inspect_rq (Station*, void*, char*, int, int);

	private:

	// Besides belonging to its creator, an observer is kept on a
	// separate list, to speed up searching. Note: the following two
	// pointers must be the first attributes of Observer.

	Observer        *next, *prev;

	static long sernum;             // Serial number

	ZZ_Object       *ChList;        // The objects owned

	// Hash masks -- to speed up things a little bit
	long            smask,          // Station mask
			pmask,          // Process mask
			nmask,          // Nickname mask
			amask;          // State mask

	ZZ_INSPECT      *inlistHead;    // The list of inspect requests

	ZZ_EVENT        *tevent;        // Timeout event pointer

	// Exposure functions

	void    exPrint0 (char*);       // Global info
	void    exPrint1 (char*);       // Inspect list

	void    exDisplay0 ();          // Global info
	void    exDisplay1 ();          // Inspect list

	public:

	// Exposable
	virtual void zz_expose (int, char *h = NULL, long s = NONE);

	void    zz_start ();            // The constructor

	// The observer's code -- to be provided by the user
	virtual void zz_code () {
		excptn (form ("Observer: observer %s code undefined",
			getTName ()));
	};

	// Returns the name of the observer state number i
	virtual char *zz_sn (int);

	friend void terminate (Observer *o);

	inline void terminate () {
		::terminate (this);
	};

	inline  void    printAll (char *hd = NULL) {
		// Information about all observers
		printOut (0, hd);
	};

	inline  void    printIns (char *hd = NULL) {
		// Print inspect list of this observer
		printOut (1, hd);
	};
};

#if     ZZ_ATT
	// Builder needed
inline  void    zz_bld_Observer (char *nn = NULL) {
	register Observer *p;
	zz_COBJ [++zz_clv] = (void*) (p = new Observer);
	// Note: zz_start for Observers is called by smpp
	if (nn != NULL) {
		p->zz_nickname = new char [strlen (nn) + 1];
		strcpy (p->zz_nickname, nn);
	}
}
#endif

/* ------------------------- */
/* Packet buffer description */
/* ------------------------- */

class   ZZ_PFItem       {

	public:

	ZZ_PFItem       *next;

	Packet          *buf;

	ZZ_PFItem (Packet *p);
};

/* --------------------------------------------- */
/* Data structures used for displaying, stepping */
/* --------------------------------------------- */

/* ---------------- */
/* An active window */
/* ---------------- */

class   ZZ_WINDOW       {

	public:                         // This type is invisible anyway

	ZZ_WINDOW       *next, *prev;   // List pointers
	ZZ_Object       *obj;           // The displayable object
	short           mode,           // Display mode
			flag;           // E.g. starting, terminating

	int             stid,           // Station id
			count;          // Item number limit

	void            *frame;         // Associated object

	// Constructors

	ZZ_WINDOW (ZZ_Object*, int, int, int);
};

/* -------------------------------------- */
/* An element of the stepped windows list */
/* -------------------------------------- */

class   ZZ_SIT          {

	public:                         // Invisible anyway

	ZZ_SIT          *next, *prev;   // List pointers

	ZZ_WINDOW       *win;           // Window pointer
	Station         *station;       // Station to be awakened
	Process         *process;       // Process to be run
	AI              *ai;            // AI generating the event
	Observer        *obs;           // For stepping observers
};

/* ------------------------------------------------------- */
/* An element of the waiting list of windows to be stepped */
/* ------------------------------------------------------- */

class   ZZ_WSIT         {

	public:                         // Invisible anyway

	ZZ_WSIT         *next, *prev;   // List pointers

	long            cevent;         // Starting event number
	TIME            ctime;          // Starting time

	ZZ_WINDOW       *win;           // Window pointer
	Station         *station;       // Station to be awakened
	Process         *process;       // Process to be run
	AI              *ai;            // AI generating the event
	Observer        *obs;           // For stepping observers
};

/* ---------------------------------------------------------- */
/* An  element  of  the  insertion  supplement  list built by */
/* zz_DADD. This list is used  to  save  some  time  if  many */
/* displayable   objects   are  dynamically  created  between */
/* display updates.                                           */
/* ---------------------------------------------------------- */

class   ZZ_SUPPLI       {

	public:

	ZZ_SUPPLI       *next, *prev;   // List pointers

	ZZ_Object       *child,         // The object
			*parent;        // Its parent (or NULL for removal)
};
/* ---------------------------------------------------------- */
/* An  element  of  the  deletion  supplement  list  built by */
/* zz_DREM. This list is used  to  save  some  time  if  many */
/* displayable   objects   are  dynamically  removed  between */
/* display updates.                                           */
/* ---------------------------------------------------------- */

class   ZZ_SUPPLD       {

	public:

	ZZ_SUPPLD       *next, *prev;   // List pointers

	char            *tname;         // The object's type name
	long            identf;         // And its serial number

	// The two attributes above are used to reconstruct the standard
	// name of the deleted object. Note that port names are different
	// from the rest. Fortunately, once created, a port is never
	// removed, so ports never appear on supplement lists.
};

/* ---------------------------------------------------------- */
/* A  special  station type representing the system agent. It */
/* is a numberless station which is created before any  other */
/* station.                                                   */
/* ---------------------------------------------------------- */

class ZZ_SYSTEM : public Station {

	public:

	virtual char *getTName () {
		return ("SYSTEM");
	};

	friend  class   Process;
	friend  void    buildNetwork ();

	private:

	// Exposure functions

	void    exPrint0 (char*);       // Network topology
	void    exPrint1 (char*);       // Abbreviated topology

	static void makeTopology ();    // Topology cleanup

	public:

	void setup () {
		Id = NONE;
		sernum = 0;
		NStations = 0;
	};

	// For exposing special things that don't belong to any
	// specific AI's or other objects

	// Exposable
	virtual void zz_expose (int, char *h = NULL, long s = NONE);

	inline  void    printTop (char *hd = NULL) {
		// Print network topology
		printOut (0, hd);
	};

	inline  void    printATop (char *hd = NULL) {
		// Print abbreviated topology
		printOut (1, hd);
	};
};

/* ------------------- */
/* System process type */
/* ------------------- */

class   ZZ_SProcess : public Process {

	public:
	char    **zz_sl;
	int     zz_ns;
	char    *zz_sn (int);
};

/* ---------------------------------------------------------- */
/* A  special  process  type  to  represent  the dummy Kernel */
/* process acting as the  terminator  of  the  user's  driver */
/* process                                                    */
/* ---------------------------------------------------------- */

class   ZZ_KERNEL : public ZZ_SProcess {

	public:

	virtual char *getTName () {
		return ("KERNEL");
	};

	enum {Start};                   // Only one dummy state

	private:

	// EXposure functions

	void    exPrint0 (char*, int);  // Global request queue
	void    exPrint1 (char*, int);  // Abbreviated global request queue
	void    exPrint2 (char*);       // Global simulation status

	void    exDisplay0 (int);       // Global request queue
	void    exDisplay1 (int);       // Abbreviated global request queue
	void    exDisplay2 ();          // Global simulation status
	void    exDisplay3 ();          // Last event

	public:

	// Exposable
	virtual void zz_expose (int, char *h = NULL, long s = NONE);

	static  int     zz_KERNEL_prcs;

	void setup () {
		zz_typeid = (void*) (&zz_KERNEL_prcs);
		Id = NONE;
		sernum = 0;
		// Only one dummy state
		zz_ns = 1;
		zz_sl = new char* [Start + 1];
		zz_sl [0] = "Start";
	};

	// For exposing special things that don't belong to any
	// specific AI's or other objects

	inline  void    printSts (char *hd = NULL) {
		// Print global simulation status
		printOut (2, hd);
	};

	inline  void    printRqs (char *hd = NULL, long s = NONE) {
		// Request queue
		printOut (0, hd, s);
	};

	inline  void    printARqs (char *hd = NULL, long s = NONE) {
		// Abbreviated request queue
		printOut (1, hd, s);
	};

	void zz_code () {

	    switch (TheState) {

		case Start:
			break;
			// See? I told you -- it does nothing
	    }
	}
};

/* ---------------------------------------------------------- */
/* Inline   functions   removed   from   objects  to  resolve */
/* cross-reference conflicts                                  */
/* ---------------------------------------------------------- */

inline void Port::zz_transmit (Packet *p, int s) {

	startTransfer (p);
	assert (TRate != 0, "Port->transmit: transmission rate undefined");
	Timer->zz_wait ((TIME)TRate * (p->TLength), s);
};

inline void Port::zz_transmit (Packet &p, int s) {

	zz_transmit (&p, s);
};

inline void Link::stdpfmMTR (Packet *p) {
#if     ZZ_ATT
	p++;            // To shut up Cfront
#endif
	NTMessages++;
};

inline void Link::stdpfmMRC (Packet *p) {
#if     ZZ_ATT
	p++;            // To shut up Cfront
#endif
	NRMessages++;
};

inline void Link::stdpfmPTR (Packet *p) {
	NTPackets++;
	NTBits += p->ILength;
};

inline void Link::stdpfmPRC (Packet *p) {
	NRPackets++;
	NRBits += p->ILength;
};

inline void Link::stdpfmPAB (Packet *p) {
	// Empty yet
};

inline int zz_client::getPacket (Packet &p, long mi, long mx, long fr) {

	return (getPacket (&p, mi, mx, fr));
};

inline int zz_client::getPacket (Packet &p, ZZ_MTTYPE tf, long mi, long mx,
	long fr) {

	return (getPacket (&p, tf, mi, mx, fr));
};

/* -------------------------------------------------------- */
/* Releases a packet that has been successfully transmitted */
/* -------------------------------------------------------- */

inline void zz_client::release (Packet &p) {
	release (&p);
};

/* -------------------------- */
/* Traffic version of receive */
/* -------------------------- */

inline void Traffic::receive (Packet *p, Link* l) {

	assert (p->TP == Id, "Traffic->receive: illegal Traffic Id");
	Client->receive (p, l);
};

/* --------------------------------------------------------- */
/* Traffic  member  version  of release (checks against type */
/* consistency)                                              */
/* --------------------------------------------------------- */

inline void Traffic::release (Packet *p) {
	assert (p->TP == Id, "Traffic->release: illegal Traffic Id");
	Client->release (p);
};

/* ------------------ */
/* Non-member receive */
/* ------------------ */

inline  void   receive (Packet &p, Link *l = NULL) {

	Client->receive (&p, l);
};

inline  void   receive (Packet &p, Link &l) {

	Client->receive (&p, &l);
};

/* ------------------ */
/* INSPECT contructor */
/* ------------------ */

inline ZZ_INSPECT::ZZ_INSPECT (Station *s, void *pt, char *nik, int st, int os){

#if     ZZ_OBS

	station = s;
	typeid = pt;
	nickname = nik;
	pstate = st;
	ostate = os;

	if (zz_inlist_tail == NULL) {
		// The first inspect
		zz_inlist_tail = zz_current_observer->inlistHead = this;
	} else {
		zz_inlist_tail -> next = this;
		zz_inlist_tail = this;
	}
	next = NULL;
#endif
};

/* ------------------------------------------------ */
/* Announcing two auxiliary functions for observers */
/* ------------------------------------------------ */

void    zz_jump_rq (int);
void    zz_timeout_rq (TIME, int);

/* ---------- */
/* For create */
/* ---------- */

inline void zz_setths (int i) { TheStation = idToStation (i); };
inline void zz_setths (Station *s) { TheStation = s; };

/* ---------------------------------------------------------- */
/* Random   number  generators  (made  inline  for  increased */
/* efficiency)                                                */
/* ---------------------------------------------------------- */

#if     ZZ_R48

#if     ZZ_ATT
extern  unsigned  short  zz_Seeds [ZZ_N_SEEDS] [ 3 ];
#else
extern  short  zz_Seeds [ZZ_N_SEEDS] [ 3 ];
#endif

extern "C" {
#if     ZZ_ATT != 1
double  erand48 (short*);
long    jrand48 (short*);
#endif
}

#else

extern  long    zz_Seeds [ZZ_N_SEEDS];
extern  double  zz_rndd;

#define         ZZ_RND1         843314861
#define         ZZ_RND2         169366695
#define         ZZ_RND3         453816693

#endif

inline  double  rnd (int seed) {

/* ----------------------------------------------------- */
/* Generates a [0,1) uniformly distributed random number */
/* ----------------------------------------------------- */
#if     ZZ_R48
	return (erand48 (zz_Seeds [seed]));
#else
	register long           cs;

	if ((cs = zz_Seeds [seed] * ZZ_RND1) > ZZ_RND2) {
		cs = cs - MAX_long;
		cs --;
	}
	if ((cs += ZZ_RND3) < 0) {
		cs = cs + MAX_long;
		cs ++;
	}

	return ((double) (zz_Seeds [seed] = cs) / zz_rndd);
#endif
};

inline long    lRndUniform (long a, long b) {

/* ---------------------------------------------------------- */
/* Generates  a  uniformly  distributed random number of type */
/* long                                                       */
/* ---------------------------------------------------------- */

	register long   res;

	res = (long) (a + (b - a + 1) * rnd (SEED_traffic));
	return (res);
};

inline long    lRndUniform (double a, double b) {

/* ---------------------------------------------------------- */
/* Generates  a  uniformly  distributed random number of type */
/* long                                                       */
/* ---------------------------------------------------------- */

	register long   res;

	res = (long) (a + (b - a + 1.0) * rnd (SEED_traffic));
	return (res);
};

inline TIME    tRndUniform (TIME a, TIME b) {

/* ---------------------------------------------------------- */
/* Generates  a  uniformly  distributed random number of type */
/* TIME                                                       */
/* ---------------------------------------------------------- */

	TIME    res;

	res = a + (TIME) (((double)(b - a) + 1.0) * rnd (SEED_traffic));
	return (res > b ? b : res);
};

inline TIME    tRndUniform (double a, double b) {

/* ---------------------------------------------------------- */
/* Generates  a  uniformly  distributed random number of type */
/* TIME                                                       */
/* ---------------------------------------------------------- */

	double  res;

	res = a + ((b - a) + 1.0) * rnd (SEED_traffic);
	if (res > b) res = b;
	return ((TIME) res);
};

inline long lRndPoisson (double mean) {

/* ---------------------------------------------------------- */
/* Exponentially distributed random number of type long based */
/* on double or long mean                                     */
/* ---------------------------------------------------------- */

	return ((long)(-mean * log (1.0 - rnd (SEED_traffic)) + 0.5));
};

inline TIME tRndPoisson (double mean) {

/* ---------------------------------------------------------- */
/* Exponentially distributed random number of type TIME based */
/* on double, long, or TIME-type mean                         */
/* ---------------------------------------------------------- */

	return ((TIME)(-mean * log (1.0 - rnd (SEED_traffic)) + 0.5));
};

inline long toss (long n) {

/* ---------------------------------------------------------- */
/* Tossing  a  multi-sided  coin  (a  random  integer  number */
/* between 0 and n-1)                                         */
/* ---------------------------------------------------------- */

	register double	r;

#if     ZZ_R48
	r = erand48 (zz_Seeds [SEED_toss]);
#else
	register long           cs;

	if ((cs = zz_Seeds [SEED_toss] * ZZ_RND1) > ZZ_RND2) {
		cs = cs - MAX_long;
		cs --;
	}
	if ((cs += ZZ_RND3) < 0) {
		cs = cs + MAX_long;
		cs ++;
	}

	r = ((double) (zz_Seeds [SEED_toss] = cs)) / zz_rndd;
#endif
	return ((long) (n * r));
};

inline int flip () {

/* ---------------------------------------------------------- */
/* Flipping  a  coin,  the  result  is  0  or 1 with the same */
/* probability                                                */
/* ---------------------------------------------------------- */

#if     ZZ_R48
	return ((jrand48 (zz_Seeds [SEED_toss]) & 010000000000) == 0L);
#else
	register long           cs;

	if ((cs = zz_Seeds [SEED_toss] * ZZ_RND1) > ZZ_RND2) {
		cs = cs - MAX_long;
		cs --;
	}
	if ((cs += ZZ_RND3) < 0) {
		cs = cs + MAX_long;
		cs ++;
	}

	zz_Seeds [SEED_toss] = cs;
	return ((cs & 010000000000) == 0L);
#endif
};

INLINE long lRndTolerance (long, long, int);
INLINE long lRndTolerance (double, double, int);

TIME   tRndTolerance   (TIME, TIME, int);

INLINE TIME    tRndTolerance   (double, double, int);

#ifndef ZZ_ATT
#include        "inlines.h"
#endif
