/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ---------------------------------------------------------- */
/* Some inline functions (AT&T Cfront doesn't like them being */
/* inline)                                                    */
/* ---------------------------------------------------------- */

#if     BIG_precision > 1

/* ---------------------------------- */
/* Multiple precision integer package */
/* ---------------------------------- */

/* ----------- */
/* Conversions */
/* ----------- */

INLINE BIG::BIG (long i) {      // Conversion from long integer

		x [0] = i;
		for (int k = 1; k < BIG_precision;) x [k++] = 0L;
};
#if     ZZ_ATT
INLINE BIG::BIG (int i) {       // AT&T Cfront likes to treat them separately

		x [0] = (long) i;
		for (int k = 1; k < BIG_precision;) x [k++] = 0L;
};
#endif

/* ----------------------------------------- */
/* Basic arithmetic relations and operations */
/* ----------------------------------------- */

INLINE  int  operator== (const BIG &a, const BIG &b) {
	for (int k = 0; k < BIG_precision; k++)
		if (a.x [k] != b.x [k]) return (0);
	return (1);
};

INLINE  int  operator!= (const BIG &a, const BIG &b) {
	for (int k = 0; k < BIG_precision; k++)
		if (a.x [k] != b.x [k]) return (1);
	return (0);
};
#endif

INLINE ZZ_REQUEST *ZZ_REQUEST::min_request () {

/* ---------------------------------------------- */
/* Finds the request with the minimum 'when' time */
/* ---------------------------------------------- */

	ZZ_REQUEST      *rm, *rc;
	TIME            t;
	int             rcount;

	for (t = (rm = this) -> when, rc = this -> other, rcount = 1;
		rc != this; rc = rc -> other) {

		if (rc -> when > t) continue;
		if (rc -> when == t) {
			rcount++;
			continue;
		}

		t = (rm = rc) -> when;
		rcount = 1;
	}

	if (rcount == 1) return (rm);

	// More than one request with the minimum time
	// Select one at random

	t = rm -> when;

	for (rcount = (int) toss (rcount); rcount; rcount--)
		for (rm = rm -> other; rm->when != t; rm = rm -> other);

	return (rm);
};

INLINE void ZZ_EVENT::enqueue () {

/* ----------------------------------- */
/* Puts the event into the event queue */
/* ----------------------------------- */

	register ZZ_EVENT *ep;

	zz_npee++;              // Number of pending events

	assert (waketime >= Time,
		"EVENT->enqueue: attempt to change the past (BIG precision too small?)");

	for (ep = zz_eq; ; ep=ep->next) {
		if (ep->waketime >= waketime) {
			ep->prev->next = this;
			prev = ep->prev;
			next = ep;
			ep->prev = this;
			return;
		}
	}
};

INLINE  void    ZZ_EVENT::enqueue (ZZ_EVENT *h) {

/* -------------------------------------------------- */
/* Puts the event into the event queue using "hint" h */
/* -------------------------------------------------- */

	zz_npee++;

	assert (waketime >= Time,
		"EVENT->enqueue(h): attempt to change the past (BIG precision too small?)");

	if (h->waketime > waketime) {
		// Examine prior events
		for (h = h->prev; h->waketime > waketime; h = h->prev);

		h->next->prev = this;
		next = h->next;
		prev = h;
		h->next = this;
		return;
	}

	// Examine past events
	for (; h->waketime < waketime; h = h->next);
	h->prev->next = this;
	prev = h->prev;
	next = h;
	h->prev = this;
};

INLINE  void    ZZ_EVENT::cancel () {

/* --------------------------------------------- */
/* Removes event and its associated request list */
/* --------------------------------------------- */

	ZZ_REQUEST *cp, *cq;

	if (chain != NULL) {
		for (cp = chain; ; ) {
			cq = cp -> other;
			zz_queue_out (cp);
			delete ((void*) cp);
			if (cq == chain) break;
			cp = cq;
		}
	}

	remove ();
};

INLINE  ZZ_EVENT  *ZZ_EVENT::get () {

/* ----------------------------------- */
/* Get next event from the event queue */
/* ----------------------------------- */

	ZZ_EVENT        *e, *eet [ZZ_EETSIZE];
	int             ne;
	TIME            t;

	if (undef (t = (e = zz_eq) -> waketime)) {
		zz_noevents ();
		return (zz_sentinel_event);
	}

	for (eet [ne = 0] = e; t == (e = e->next)->waketime;) {

		if (ne >= ZZ_EETSIZE - 1)
			excptn ("EVENT->get: stack overflow");
		eet [++ne] = e;
	}

	e = ne ? eet [toss (ne+1)] : eet [0];

	assert (t >= Time, "EVENT->get: time going backward");
	e -> remove ();
	return (e);
};

INLINE Port *Station::idToPort (int id) {

	register Port *p;

	for (p = Ports; p != NULL && id--; p = p->nextp);
	Assert (p != NULL, "Station->idToPort: illegal port Id");
	return (p);
};

INLINE Port *idToPort (int id) {

	register Port *p;

	Assert (TheStation != NULL,
		"Station->idToPort: current station undefined");
	for (p = TheStation->Ports; p != NULL && id--; p = p->nextp);
	Assert (p != NULL, "Station->idToPort: illegal port Id");
	return (p);
};

/* ------------------------ */
/* Random number generators */
/* ------------------------ */

INLINE long lRndTolerance (long a, long b, int q) {

/* ---------------------------------------------------------- */
/* Generates  a  number between a and b according to the Beta */
/* (q,q) distribution (long version)                          */
/* ---------------------------------------------------------- */

	double                  y;
	register        double  p;
	register        int     i;

	assert (q > 0, "lRndTolerance: q must be > 0");
	for (p = 1.0, i = 1; i <= q; i++, p *= 1.0 - rnd (SEED_delay));
	y = -log (p);
	for (p = 1.0, i = 1; i <= q; i++, p *= 1.0 - rnd (SEED_delay));
	y = y / (y - log (p));
	return ((long) (a + (b - a + 1) * y));
};

INLINE long lRndTolerance (double a, double b, int q) {

/* ---------------------------------------------------------- */
/* Generates  a  number between a and b according to the Beta */
/* (q,q)  distribution  (long   version   based   on   double */
/* arguments)                                                 */
/* ---------------------------------------------------------- */

	double                  y;
	register        double  p;
	register        int     i;

	assert (q > 0, "lRndTolerance: q must be > 0");
	for (p = 1.0, i = 1; i <= q; i++, p *= 1.0 - rnd (SEED_delay));
	y = -log (p);
	for (p = 1.0, i = 1; i <= q; i++, p *= 1.0 - rnd (SEED_delay));
	y = y / (y - log (p));
	return ((long)(a + (b - a + 1) * y));
};

INLINE TIME tRndTolerance   (double a, double b, int q) {

/* ---------------------------------------------------------- */
/* Generates  a  number between a and b according to the Beta */
/* (q,q) distribution (BIG version based on  double  or  long */
/* arguments)                                                 */
/* ---------------------------------------------------------- */

	double                  y;
	register        double  p;
	register        int     i;

	assert (q > 0, "tRndTolerance: q must be > 0");
	for (p = 1.0, i = 1; i <= q; i++, p *= 1.0 - rnd (SEED_delay));
	y = -log (p);
	for (p = 1.0, i = 1; i <= q; i++, p *= 1.0 - rnd (SEED_delay));
	y = y / (y - log (p));
	return ((TIME)(a + (((b - a) + 1.0) * y)));
};

/* ----------------------- */
/* Miscellaneous functions */
/* ----------------------- */

INLINE TIME Port::bitsToTime (long b) {

	assert (TRate != 0, "Port->bitsToTime: transmission rate undefined");
	return ((TIME)TRate * b);
};
