/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* -------------------------------------------- */
/* The main program: execution starts from here */
/* -------------------------------------------- */

#include        <setjmp.h>
#include        "system.h"

jmp_buf         zz_mloop;                // Return to the main loop
int             zz_setjmp_done = NO;     // Flag == longjmp legal

#if     ZZ_OBS
#if     ZZ_ATT
#define smatch(a,b)     (!strcmp (a, b))
#else
static  inline  int     smatch (register char *s1, register char *s2) {

/* ------------------------- */
/* A local version of strcmp */
/* ------------------------- */

	if (s2 == NULL) return (0);
	while (*s1 == *s2++) if (*s1++ == '\0') return (1);
	return (0);
}
#endif
#endif

#if     ZZ_TOL
void    setTolerance (double t, int q) {

	if ((zz_tolerance = t) == 0.0)
		zz_quality = 0;
	else
		zz_quality = q;
	Assert ((zz_quality >=0) && (zz_quality <= 10) && (t >= 0.0) &&
		(t < 1.0), "setTolerance: illegal clock tolerance parameters");
#else
void    setTolerance (double, int) {
	excptn ("setTolerance illegal: smurph created with '-n'");
#endif
}

void    setEtu (double e) {
	Itu = 1.0 / (Etu = e);
}
void    setItu (double e) {
	Etu = 1.0 / (Itu = e);
}

void    setLimit (long nm, TIME mt, double ct) {

	Assert (nm >= 0, "setLimit: illegal message number limit");
	zz_max_NMessages = (nm == 0) ? MAX_long : nm;
	zz_max_Time = (mt == TIME_0) ? TIME_inf : mt;
	Assert (ct >= 0.0, "setLimit: illegal CPU time limit");
	zz_max_cpuTime = (ct < 0.00001) ? HUGE : ct;
}

void    setLimit (long nm, TIME mt) {
	setLimit (nm, mt, zz_max_cpuTime);
}

void    setLimit (long nm) {
	setLimit (nm, zz_max_Time);
}

void    setFlush () {
	zz_flg_emptyMQueues = YES;
}

void    buildNetwork () {

/* -------------------------- */
/* Forces topology definition */
/* -------------------------- */

	if (!zz_flg_started) {

		// Postprocess the network topolgy definitions by
		// cleaning up temporary data structures and creating
		// port distance vectors.

		System->makeTopology ();

		// Don't do it the next time
		zz_flg_started = YES;
	}
}

#if     ZZ_R48 != 1
static void initrnd () {

/* ------------------------------- */
/* Initializes the divisor for rnd */
/* ------------------------------- */

	double  f, g, h, u, p;

	u = f = (double) MAX_long;
	g = 0.5;

	while (1) {
		h = f + g;
		p = u / h;
		if (p >= 1.0) break;
		zz_rndd = h;
		g = g/2.0;
	}
}
#endif

extern  char    *ProtocolId;

main (int argc, char *argv []) {

	int        ic;
	ZZ_REQUEST *chain;
	ZZ_EVENT   *ev, *eh;

	register     ZZ_REQUEST    *cp, *cq;

#if     ZZ_SMURPHDEBUG == 2
	zz_dbgf = new ostream ("debug", "w");
#endif

#if     ZZ_R48 != 1
	// Initialize the divisor for rnd
	initrnd ();
#endif

	TracingTime = TIME_inf;         // Cannot initialize statically
	zz_max_Time = TIME_inf;

	// Initialize event queue sentinels

	zz_rsent.next = NULL;           // This is the last event
	zz_rsent.prev = &zz_fsent;      // This is the first event
	zz_rsent.chain = NULL;          // Nullify all attributes
	zz_rsent.waketime = TIME_inf;
	zz_rsent.station = NULL;
	zz_rsent.process = NULL;
	zz_rsent.pstate = 0;
	zz_rsent.event_id = 0;
	zz_rsent.ai = NULL;
	zz_rsent.Info01 = zz_rsent.Info02 = NULL;

	zz_fsent.next = &zz_rsent;
	zz_fsent.prev = NULL;
	zz_fsent.chain = NULL;
	zz_fsent.waketime = TIME_0;
	zz_fsent.station = NULL;
	zz_fsent.process = NULL;
	zz_fsent.pstate = 0;
	zz_rsent.event_id = 0;
	zz_fsent.ai = NULL;
	zz_fsent.Info01 = zz_rsent.Info02 = NULL;

	zz_sentinel_event = &zz_rsent;

	zz_init_system (argc, argv);    // Files + options

	// Output the protocol identifier, date, and time

	Ouf << "SMURPH Version " << VERSION << "      ";
	if (ProtocolId != NULL)
		Ouf << ProtocolId << "    ";
	Ouf << tDate () << '\n';

	// Include simulator call arguments
	Ouf << "Call arguments:";
	for (argc--, argv++; argc > 0; argc--, argv++) Ouf << ' ' << *argv;
	Ouf << "\n\n";

	// Create the system process and the system station

	TheProcess = NULL;

	// System = create ZZ_SYSTEM;
	System = new ZZ_SYSTEM;
	System->zz_start ();
	System->setup ();
	TheStation = System;

	// ZZ_Kernel = Kernel = create ZZ_KERNEL;
	ZZ_Kernel = Kernel = new ZZ_KERNEL;
	Kernel->zz_start ();
	Kernel->setup ();
	TheProcess = Kernel;

	// Create and start the user process
	
	zz_c_first_wait = YES;
	ZZRoot ();                      // Provided by the user

	if (ZZ_Main == NULL) excptn ("smurph: Root process not provided");

	// Turn the starting event for Root into a priority event -- to make
	// sure that before the Client is started, the network has been
	// initialized properly

	Assert (zz_pe == NULL,
		"smurph: a priority signal sent upon Root startup");

	for (ev = zz_eq; ev != zz_sentinel_event; ev = ev->next) {
		if (ev->process == ZZ_Main && ev->ai == ZZ_Main &&
			ev->event_id == START) {

			zz_pe = ev;
			zz_pr = ev->chain;
			break;
		}
	}

	Assert (zz_pe != NULL, "smurph: failed to create Root");

	// Process the first event -- the one starting the user's coordinator --
	// 'by hand'. A subtle postprocessing is required after this event has
	// been serviced.

	zz_CE = zz_pe;
	zz_CE -> remove ();
	zz_pe = NULL;

	Time        = zz_CE -> waketime;
	chain       = zz_CE -> chain;
	TheStation  = zz_CE -> station;
	TheProcess  = zz_CE -> process;
	zz_event_id = zz_CE -> event_id;
	Info01      = zz_CE -> Info01;
	Info02      = zz_CE -> Info02;
	TheState    = zz_CE -> pstate;

	Assert ((Time == TIME_0) && (chain != NULL) && (TheStation == System) &&
		(TheProcess == ZZ_Main) && (zz_event_id == START) &&
		(TheState == 0), "smurph: problems starting Root");

	// Deallocate request chain (it should consist of exactly one entry,
	// but who knows ...)
	for (cp = chain; ; cp = cq) {
		cq = cp -> other;
		queue_out (cp);
		delete ((void*) cp);
		if (cq == chain) break;
	}
	zz_c_first_wait = YES;

	// Run the coordinator
	TheProcess->zz_code ();

	delete ((void*) zz_CE);         // Deallocate the event
	zz_CE = NULL;                   // Mark CE as nonexistent

	TheProcess  = Kernel;
	TheStation  = System;
	TheState    = 0;
	zz_event_id = 0;
	zz_start_client ();             // This should do

	// Start counting events. Note that the very first event initializing
	// the coordinator is not counted.
	ic = 0;
	zz_chckint = CHCKINT;

	buildNetwork ();                // Force topology definition

	if (zz_flg_printDefs) {
		// Output network configuration and more garbage
		System->printTop ("Network topology:");
		Client->printDef ("Traffic:");
	}

	zz_adjust_ownership ();         // Move level-0 processes to stations
	zz_init_windows ();             // Initialize the window system

	// Deallocate input file
	if ((zz_ifn != NULL) && ((zz_ifn[0] != '.') || (zz_ifn[1] != '\0'))) {
		Inf.close ();
		delete (zz_ifpp);
		zz_ifpp = NULL;
	}

	zz_setjmp_done = YES;
	if (setjmp (zz_mloop)) {
		ic++; ic--;
	}

	for ( ; ; ic++) {

		// The main loop

		// Check exit conditions

		if (!zz_flg_emptyMQueues) {

			// Continue flag is down, check received messages */
			if (zz_NRMessages >= zz_max_NMessages) {
				zz_exit_code = EXIT_msglimit;
				Kernel->terminate ();
			}

		} else if (zz_flg_emptyMQueues == YES) {

			// Client still active, check generated messages
			if (zz_NGMessages >= zz_max_NMessages) {

			    // Remove Client events from  the  event queue
			    // and continue until message queues are empty

				for (ev = zz_eq; ev != zz_sentinel_event; ) {

					// System events only
					if (ev -> ai != Client || ev -> chain
						!= NULL) {

						ev = ev -> next;
						continue;
					}

					// Release the event
					eh = ev -> next;
					ev -> cancel ();
					ev = eh;
				}

				// Flag -- flushing message queues
				zz_flg_emptyMQueues = NONE;
				// Restart the main loop
				ic --;
				continue;
			}
		} else {

			// Flushing message queues
			if (zz_NRMessages >= zz_NGMessages) {
				zz_flg_emptyMQueues = NO;
				zz_exit_code = EXIT_msglimit;
				Kernel->terminate ();
			}
		}

		// Virtual time limit
		if (Time >= zz_max_Time) {
			zz_exit_code = EXIT_stimelimit;
			Kernel->terminate ();
		}

		if (ic >= zz_chckint) {
			// Do not check execution time limit too often
			if (zz_chckint == 0) {
				// Checkpoint request
				zz_chckint = CHCKINT;
#if     ZZ_SMURPHDEBUG == 2
#else
				if (!DisplayActive && ZZ_Kernel != NULL)
					zz_checkpoint ();
#endif
			} else {
				ic = 0;
				if (cpuTime () >= zz_max_cpuTime) {
					zz_exit_code = EXIT_rtimelimit;
					Kernel->terminate ();
				}
			}
		}

		if (zz_pe != NULL) {
			// Priority event
			zz_pe -> new_top_request (zz_pr);
			zz_CE = zz_pe;
			zz_pe -> remove ();
			zz_pe = NULL;
		} else {
			// Regular event
			zz_CE = zz_eq->get ();
		}

		if (zz_CE == zz_sentinel_event) break;
		Time = zz_CE -> waketime;            // Advance the time
		chain = zz_CE->chain;
		TheStation  = zz_CE -> station;
		TheProcess  = zz_CE -> process;

		zz_event_id = zz_CE -> event_id;
		Info01      = zz_CE -> Info01;
		Info02      = zz_CE -> Info02;
		TheState    = zz_CE -> pstate;
		zz_ai = zz_CE->ai;

		if (TheStation != NULL) {
		    if (zz_ai == Signal) {
			if (chain->what != NULL) {
			    // Deallocate the Signal entry
			    queue_out ((ZZ_SIG*) (chain->what));
			    delete ((ZZ_SIG*)(chain->what));
			    chain->what = NULL;
			}
		    } else if (zz_ai->Class == AIC_port) {
			/* ----------------------------------------------- */
			/* A port AI:  force  priorities  of  simultaneous */
			/* events:                                         */
			/*         **  BMP                                 */
			/*             BOT                                 */
			/*             ACTIVITY                            */
			/*                                                 */
			/*         **  EMP                                 */
			/*             EOT                                 */
			/*             SILENCE                             */
			/* ----------------------------------------------- */

			cp = chain;

			switch (zz_port_event_id = (int) zz_event_id) {

			  case  ACTIVITY:

				for (cp = cp->other ; cp != chain;
					cp = cp->other) {

					if ((cp->ai == zz_ai) &&
					     (cp->when == Time)) {
						if (cp->event_id == BOT) {
						  Info01      = cp -> Info01;
						  // Info02 is the same
						  TheState  = cp -> pstate;
						  zz_event_id = cp -> event_id;
						  goto CHMYP;
						} else if (cp->event_id ==
						  BMP) {
						  Info01      = cp -> Info01;
						  // Info02 is the same
						  TheState  = cp -> pstate;
						  zz_event_id = cp -> event_id;
						  goto CSTOP;
						}
					}
				}
				break;

CHMYP:
			  case  BOT:

				for (cp = cp->other; cp != chain;
					cp = cp->other) {

					if ((cp->ai == zz_ai) &&
					     (cp->when == Time) && 
					    (cp->event_id == BMP)) {
						  Info01      = cp -> Info01;
						  // Info02 is the same
						  TheState  = cp -> pstate;
						  zz_event_id = cp -> event_id;
						  goto CSTOP;
					}
				}
				break;

			  case  SILENCE:

				for (cp = cp->other; cp != chain;
					cp = cp->other) {

					if ((cp->ai == zz_ai) &&
					     (cp->when == Time)) {
						if (cp->event_id == EOT) {
						  Info01      = cp -> Info01;
						  // Info02 is the same
						  TheState  = cp -> pstate;
						  zz_event_id = cp -> event_id;
						  goto CHEMP;
						} else if (cp->event_id ==
						  EMP) {
						  Info01      = cp -> Info01;
						  // Info02 is the same
						  TheState  = cp -> pstate;
						  zz_event_id = cp -> event_id;
						  goto CSTOP;
						}
					}
				}
				break;
CHEMP:
			  case  EOT:

				for (cp = cp->other; cp != chain;
					cp = cp->other) {

					if ((cp->ai == zz_ai) &&
					     (cp->when == Time) && 
					    (cp->event_id == EMP)) {
						  Info01      = cp -> Info01;
						  // Info02 is the same
						  TheState  = cp -> pstate;
						  zz_event_id = cp -> event_id;
						  goto CSTOP;
					}
				}
				break;

			   default: ;   // None
			}
		    }
		}
CSTOP:

#if     ZZ_DBG
		// Check if tracing is on and, if appropriate, print out debug
		// information

		if (def (TracingTime) && (TracingTime <= Time))
			zz_print_debug_info ();
#endif
		// Restart processes waiting for this process entering this
		// state

		if (TheProcess->SWList != NULL) {
		  for (cp = TheProcess->SWList; cp != NULL; cp = cp->next) {
				if (cp->event_id == TheState) {
					if ((ev = cp->event)->waketime > Time ||
						flip ())
						       ev->new_top_request (cp);
				}
		  }
		}

		// Release request chain

		if (chain != NULL) {
		    for (cp = chain; ; cp = cq) {
			cq = cp -> other;
			queue_out (cp);

			if (cp->id < 0) {       // Postprocessing required
			    if (cp->id == RQTYPE_SIG) {
				if (cp->what != NULL)
				    // The Signal returns to the station
					((ZZ_SIG*)(cp->what))->request = NULL;
			    } else {            // A process wait request
				if (((Process*)(cp->ai))->Owner == NULL &&
					((Process*)(cp->ai))->TWList == NULL &&
						((Process*)(cp->ai))->SWList ==
							NULL) {

					// Delete the zombie
assert (((Process*)(cp->ai))->ChList == NULL,
form ("Can't terminate %s -- ownership list not empty",
cp->ai->getOName ()));
					zz_queue_out (cp->ai);
					zz_DREM (cp->ai);
					if (cp->ai->zz_nickname != NULL)
						delete (cp->ai->zz_nickname);
					delete ((void*)(cp->ai));
				}
			    }
			}
			delete ((void*) cp);
			if (cq == chain) break;
		    }
		}
		
		// Run the process
		zz_c_first_wait = YES;
		TheProcess->zz_code ();

#if     ZZ_OBS
		// Check whether should call an observer

		if ((zz_obslist != NULL) && (chain != NULL) &&
			(zz_CE->station != System)) {

			long            smask, pmask, nmask, amask;
			ZZ_INSPECT      *ins, *jns;

			/* -------------------------------------------- */
			/* Note: the following instructions assume that */
			/*       words are (at least) 32-bit long.      */
			/* -------------------------------------------- */

			smask = 1 << ((TheStation->Id) & 037);
			pmask = 1 << ((((int)(TheProcess->zz_typeid)) >> 2) &
				037);
			if (TheProcess->zz_nickname != NULL)
			   nmask = 1 << ((*(TheProcess->zz_nickname)) & 037);
			else
			   nmask = 1;
			amask = 1 << (TheState & 037);

			zz_observer_running = YES;

			for (zz_current_observer = zz_obslist;
				zz_current_observer != NULL;
				zz_current_observer = zz_current_observer ->
				next) {

				if ((zz_current_observer->amask & amask) &&
				    (zz_current_observer->pmask & pmask) &&
				    (zz_current_observer->nmask & nmask) &&
				    (zz_current_observer->smask & smask)  )  {

					// Examine the inspect list

					for (ins =
						zz_current_observer->inlistHead;
							ins != NULL;
							ins = ins->next) {
	// <------

	if (((ins->pstate == ANY) || (ins->pstate == TheState)) &&
	   (((int)(ins->typeid) == ANY) || (ins->typeid ==
	   TheProcess->zz_typeid)) && (((int)(ins->station) == ANY) ||
	   (ins->station == TheStation)) && (((int)(ins->nickname) == ANY) ||
	   smatch(ins->nickname, TheProcess->zz_nickname))) {
		
		TheObserverState = ins->ostate;
		
		// Erase the inspects
		
		for (ins = zz_current_observer->inlistHead; ins != NULL;
			ins=jns) {

			jns = ins -> next;
			delete ((void*) ins);
		}

		zz_current_observer->inlistHead = zz_inlist_tail = NULL;

		// Clear inspect masks

		zz_current_observer->amask = zz_current_observer->pmask =
			zz_current_observer->nmask =
				zz_current_observer->smask = 0;

		// Remove pending timeout event

		if (zz_current_observer->tevent != NULL) {
			zz_current_observer->tevent->cancel ();
			zz_current_observer->tevent = NULL;
		}

		// Call the observer

		do {
		  zz_jump_flag = NO;
		  zz_current_observer->zz_code ();
		  if (DisplayActive) {
		    // Observer stepping check
		    ZZ_SIT   *ws;

		    for (ws = zz_steplist; ws != NULL; ws = ws->next) {
		      if (ws->obs != zz_current_observer) continue;
		      if (ws->station != NULL && ws->station != TheStation)
			continue;
		      if (ws->process != NULL && ws->process != TheProcess)
			continue;
		      if (ws->ai != NULL && ws->ai != zz_CE->ai)
			continue;

		      zz_events_to_display = 0;
		      zz_send_step_phrase = YES;
		      break;
		    }
		  }

		  TheObserverState = zz_state_to_jump;

		} while (zz_jump_flag);

		break;          // Continue with the next observer
	}

	// ------>
					}
				}
			}
		}

		zz_observer_running = NO;
#endif
		if (DisplayActive) {
		  if (zz_steplist_delayed != NULL) {
		    // There are windows waiting to be stepped
		    if (zz_npre >= zz_commit_event || Time >= zz_commit_time) {
		      // Find the windows to be added to steplist
		      ZZ_WSIT  *wsd, *wsq;
		      ZZ_SIT   *ws;

		      for (wsd = zz_steplist_delayed; wsd != NULL; wsd = wsq) {
			wsq = wsd->next;
			if (wsd->cevent <= zz_npre || wsd->ctime <= Time) {
			  // Move it to steplist
			  ws = new ZZ_SIT;
			  ws->win = wsd->win;
			  ws->station = wsd->station;
			  ws->process = wsd->process;
			  ws->ai = wsd->ai;
			  ws->obs = wsd->obs;
			  queue_head (ws, zz_steplist, ZZ_SIT);
			  queue_out (wsd);
			  delete (wsd);
			}
		      }

		      // Calculate new minima

		      for (zz_commit_time = TIME_inf,
			zz_commit_event = MAX_long, wsd = zz_steplist_delayed;
				wsd != NULL; wsd = wsd->next) {

			if (wsd->ctime < zz_commit_time)
			  zz_commit_time = wsd->ctime;
			if (wsd->cevent < zz_commit_event)
			  zz_commit_event = wsd->cevent;
		      }
		    }
		  }

		  // Done with the delayed list, now regular step test

		  if (zz_steplist != NULL) {
		    // There are windows to be stepped: look for match
		    ZZ_SIT   *ws;

		    for (ws = zz_steplist; ws != NULL; ws = ws->next) {
		      if (ws->station != NULL && ws->station != TheStation)
			continue;
		      if (ws->process != NULL && ws->process != TheProcess)
			continue;
		      if (ws->ai != NULL && ws->ai != zz_CE->ai)
			continue;
		      if (ws->obs == NULL) break;
		      // Observers tested elsewhere
		    }
		    if (ws != NULL) {
		      // Step it
		      zz_events_to_display = 0;
		      zz_send_step_phrase = YES;
		    }
		  }

		  if (--zz_events_to_display <= 0) refreshDisplay ();
		}

		// Deallocate the event
		delete ((void*) zz_CE);
		zz_CE = NULL;           // Mark CE as nonexistent
		zz_npre++;              // Number of processed events
	}

	zz_done ();
}

/* ---------------------------------------------------------- */
/* This  function  doesn't  really fit here, but I don't feel */
/* like creating a new file just for it                       */
/* ---------------------------------------------------------- */

TIME    tRndTolerance   (TIME a, TIME b, int q) {

/* ---------------------------------------------------------- */
/* Generates  a  number between a and b according to the Beta */
/* (q,q) distribution (BIG version)                           */
/* ---------------------------------------------------------- */

	double                  y;
	register        double  p;
	register        int     i;
	TIME                    r;

	assert (q > 0, "tRndTolerance: q must be > 0");
	for (p = 1.0, i = 1; i <= q; i++, p *= 1.0 - rnd (SEED_delay));
	y = -log (p);
	for (p = 1.0, i = 1; i <= q; i++, p *= 1.0 - rnd (SEED_delay));
	y = y / (y - log (p));
	r = a + (TIME)(((double)(b - a) + 1.0) * y);
	if (r > b) return (b); else return (r);
};
