/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* --------------- */
/* Process service */
/* --------------- */

#include        <setjmp.h>
#include        "system.h"

/* -------------------- */
/* Borrowed from main.c */
/* -------------------- */
extern  jmp_buf zz_mloop;                // Return to the main loop
extern  int     zz_setjmp_done;          // Flag == longjmp legal

void    Process::zz_start () {

	Process    *savecontext;
	int        cfv;
	ZZ_REQUEST *cwh, *cot;
	ZZ_EVENT   *cwe;
	TIME       cwm;

	Id = sernum++;

	if_from_observer ("Process: observers can't generate processes");

	Owner = TheStation;

	TWList = SWList = NULL;
	ChList = NULL;

	// Save the current context for a second
	savecontext = TheProcess;
	cfv = zz_c_first_wait;
	cwh = zz_c_whead;
	cot = zz_c_other;
	cwe = zz_c_wait_event;
	cwm = zz_c_wait_tmin;

	// Assume the context of the started process (formally nonexistent yet)
	TheProcess = this;
	zz_c_first_wait = YES;

	// Generate the launching event looking as a START request to itself
	zz_wait (START, 0);

	// Restore previous context
	zz_c_first_wait = cfv;
	zz_c_whead = cwh;
	zz_c_other = cot;
	zz_c_wait_event = cwe;
	zz_c_wait_tmin = cwm;
	TheProcess = savecontext;

	Class = AIC_process;

	// Add the object to the owner's list

	if (TheProcess != NULL) {

		// TheProcess == NULL means that Kernel is created
		zz_queue_head (this, TheProcess->ChList, ZZ_Object);

		// Notify the server about a new displayable object.
		zz_DADD (TheProcess, this);
	}
};

char    *Process::zz_sn (int) {

/* -------------------------------------- */
/* Returns the name of the state number i */
/* -------------------------------------- */

	// This is a virtual function
	return ("undefined");
}

char    *ZZ_SProcess::zz_sn (int i) {

/* ---------------------------------------------------------- */
/* Returns  the  name  of  the  state  number i (for a system */
/* process)                                                   */
/* ---------------------------------------------------------- */

	return ((i > zz_ns) ? "undefined" : zz_sl [i]);
}

void    Process::zz_wait (int ev, int pstate) {

/* ------------------------ */
/* The process wait request */
/* ------------------------ */

	TIME    t;

	if_from_observer ("process->wait: called from an observer");

	if (zz_c_first_wait) {

		zz_c_other = NULL;    // No requests so far

		// Create new simulation event
		zz_c_wait_event = new ZZ_EVENT;
		zz_c_wait_event -> station = TheStation;
		zz_c_wait_event -> process = TheProcess;
	}

	assert (ev >= START, "process->wait: illegal event");

	if (ev >= 0) {
		// State wait request
		new ZZ_REQUEST (&SWList, this, ev, pstate, RQTYPE_PRC);
		assert (zz_c_other != NULL,
			"process->wait: internal error -- null request");
		t = TIME_inf;
		if (zz_c_first_wait) {
			zz_c_wait_event -> pstate = pstate;
			zz_c_wait_event -> ai = this;
			zz_c_wait_event -> event_id = ev;
			zz_c_wait_event -> waketime = TIME_inf;
			zz_c_wait_event -> chain = zz_c_other;
			zz_c_wait_event -> Info01 = NULL;
			zz_c_wait_event -> Info02 = NULL;
		}
	} else if (ev == DEATH) {
		// Wait request for termination
		new ZZ_REQUEST (&TWList, this, ev, pstate, RQTYPE_PRC);
		assert (zz_c_other != NULL,
			"process->wait: internal error -- null request");
		t = TIME_inf;
		if (zz_c_first_wait) {
			zz_c_wait_event -> pstate = pstate;
			zz_c_wait_event -> ai = this;
			zz_c_wait_event -> event_id = ev;
			zz_c_wait_event -> waketime = TIME_inf;
			zz_c_wait_event -> chain = zz_c_other;
			zz_c_wait_event -> Info01 = NULL;
			zz_c_wait_event -> Info02 = NULL;
		}
	} else {
		// START request
		new ZZ_REQUEST (this, ev, pstate, RQTYPE_PRC);
		assert (zz_c_other != NULL,
			"process->wait: internal error -- null request");
		t = Time;       // Occurs immediately
		if (zz_c_first_wait || zz_c_wait_tmin > t || flip ()) {
			zz_c_wait_event -> pstate = pstate;
			zz_c_wait_event -> ai = this;
			zz_c_wait_event -> event_id = ev;
			zz_c_wait_tmin =
				zz_c_wait_event -> waketime = Time;
			zz_c_wait_event -> chain = zz_c_other;
			zz_c_wait_event -> Info01 = NULL;
			zz_c_wait_event -> Info02 = NULL;

			if (!zz_c_first_wait)
				zz_c_wait_event -> reschedule ();
		}
	}


	if (zz_c_first_wait) {
		zz_c_whead = zz_c_other;
		zz_c_first_wait = NO;

		if (def (t))
			zz_c_wait_event->enqueue ();
		else
			zz_c_wait_event->store ();
	}

	zz_c_other -> when = t;
	zz_c_other -> event = zz_c_wait_event;
	zz_c_other -> other = zz_c_whead;
}

void    zz_noevents () {

/* -------------------------------------------------------- */
/* Processes the case when the simulator runs out of events */
/* -------------------------------------------------------- */

	zz_exit_code = EXIT_noevents;
	Kernel->terminate ();
}

void terminate (Process *p) {

/* ---------------------------------------------------------- */
/* Stops   and  destroys  the  process.  Wakes  up  processes */
/* awaiting its termination                                   */
/* ---------------------------------------------------------- */

ZZ_EVENT   *ev, *eh;
ZZ_REQUEST *cw, *rq, *rc;

	if (p == Kernel) {
		// Terminates the simulation
		if (ZZ_Kernel == NULL || zz_ofpp == NULL) zz_done ();
		// Avoid it second time
		ZZ_Kernel = NULL;

		// Wake up processes waiting for the Kernel's termination
		for (cw = p->TWList; cw != NULL; cw = cw -> next) {
			if (cw -> event_id == DEATH) {
				cw -> when = Time;

				if ((ev = cw->event) -> waketime > Time ||
					flip ()) {
						cw -> Info01 = (void*)
								   zz_exit_code;
						cw -> Info02 = NULL;
						ev->new_top_request (cw);
				}

			}
		}

		// Remove all events and requests besides the processes waiting
		// for system termination

		for (ev = zz_eq; ev != zz_sentinel_event; ) {

			if (ev->chain == NULL) {
				// Remove without any further checks
				eh = ev -> next;
				ev -> cancel ();
				ev = eh;
				continue;
			}

			for (rq = ev->chain; ;) {
				if (rq->ai == Kernel && rq->event_id == DEATH)
					goto HOLD;
				if ((rq = rq->other) == ev->chain) break;
			}

			// Irrelevant -- deallocate

			eh = ev -> next;
			ev -> cancel ();
			ev = eh;
			continue;

HOLD:                   // Remove all requests except for the Kernel's DEATH

			Assert (rq->when == Time,
		"Kernel->terminate: internal error -- illegal restart time");
			ev->new_top_request (rq);
			for (rq = (rc = ev->chain)->other; rq != ev->chain; ) {
				if (rq->ai == Kernel && rq->event_id == DEATH) {
					rq = (rc = rq)->other;
					continue;
				}
				rc->other = rq->other;
				queue_out (rq);
				delete (rq);
				rq = rc->other;
			}

			ev = ev -> next;
		}

		// Go to process the termination request
		setLimit (MAX_long, TIME_inf, HUGE); // Reset limits to infinity
		if (zz_setjmp_done) longjmp (zz_mloop, NONE);
		// Force unconditional termination
		zz_done ();
		// Never returns
	}

	if (p->ChList != NULL) {

		if (p != ZZ_Main)
			excptn (form ("terminate: process %s has descendants",
				p->getSName ()));
	}

	for (ev = zz_eq; ev != zz_sentinel_event; ) {
		if (ev -> process != p) {
			ev = ev -> next;
			continue;
		}

		eh = ev -> next;
		ev -> cancel ();
		ev = eh;
	}

	// Wake up processes waiting for the termination

	for (cw = p->TWList; cw != NULL; cw = cw -> next) {
		if (cw -> event_id == DEATH) {
			cw -> when = Time;

			if ((ev = cw->event) -> waketime > Time || flip ()) {
					cw -> Info01 = (void*) (p->Ret1);
					cw -> Info02 = (void*) (p->Ret2);
					ev->new_top_request (cw);
			}

		}
	}

	// Deallocate the process: the top-level destructor will remove it
	// from the owner's list

	if (p->TWList == NULL) {
		assert (p->ChList == NULL,
			form ("Can't terminate %s -- ownership list not empty",
				p->getOName ()));
		zz_queue_out (p);
		zz_DREM (p);
		if (p->zz_nickname != NULL) delete (p->zz_nickname);
		delete (p);
	} else
		p->Owner = NULL;        // Flag == zombie process

	// If somebody is waiting for the process, leave it as a zombie
	// until the last waiting process destroys it.
}

sexposure (Process)

	sonpaper {

		sexmode (0)

			exPrint0 (Hdr, (int) SId);    // Request queue

		sexmode (1)

			exPrint1 (Hdr);               // Wait list
	}

	sonscreen {

		sexmode (0)

			exDisplay0 ((int) SId);       // Request queue

		sexmode (1)

			exDisplay1 ();                // Wait list
	}
}

void    Process::exPrint0 (char *hdr, int sid) {

/* ----------------------- */
/* Print the request queue */
/* ----------------------- */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;
	long               l;

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName ();
		if (Owner != NULL)
			Ouf << '/' << Owner->getOName ();
		Ouf << ") All processes waiting for " << getOName ();
		Ouf << ":\n\n";
	}

	if (isStationId (sid))
		Ouf << "           Time ";
	else
		Ouf << "       Time   St";

	Ouf << "    Process/Idn     PState      Event         AI/Idn" <<
		"      State\n";

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (e->chain == NULL || e->station == NULL ||
			isStationId (sid) && sid != e->station->Id) continue;

		for (r = e->chain;;) {

			if (r->ai == this) {

				if (!isStationId (sid))
					ptime (r->when, 11);
				else
					ptime (r->when, 15);

				if (e->waketime < r->when)
					Ouf << ' ';     // Obsolete
				else if (e->chain == r)
					Ouf << '*';     // Uncertain
				else
					Ouf << '?';

				if (!isStationId (sid)) {
					// Identify the station
					Ouf << form ("%4d ",
				zz_trunc (ident (e->station), 3));
				}

				print (e->process->getTName (), 10);
				Ouf << form ("/%03d ",
					zz_trunc (e->process->Id, 3));

				print (zz_eid (r->event_id), 10);
				Ouf << ' ';

				print (e->process->zz_sn (r->pstate), 10);
				Ouf << ' ';

				print (e->ai->getTName (), 10);
				if ((l = e->ai->Id) != NONE) {
					if (e->ai->Class == AIC_port)
						l = l & 0177777;
					Ouf << form ("/%03d ",
						zz_trunc (l, 3));
				} else
					Ouf << "     ";

				print (e->process->zz_sn (e->chain->pstate),10);
				Ouf << '\n';
			}

			if ((r = r->other) == e->chain) break;
		}
	}

	Ouf << "\n(" << getOName ();
	if (Owner != NULL)
		Ouf << '/' << Owner->getOName ();
	Ouf << ") End of list\n\n";
}

void    Process::exPrint1 (char *hdr) {

/* ------------------- */
/* Print the wait list */
/* ------------------- */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;
	long               l;

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName ();
		if (Owner != NULL)
			Ouf << '/' << Owner->getOName ();
		Ouf << ") Request list:\n\n";
	}

	Ouf << "        AI/Idn      Event       State            Time\n";

	for (e = zz_eq; e != zz_sentinel_event; e = e -> next)
		if (e->process == this) break;

	if (e != zz_sentinel_event && (r = e->chain) != NULL) {

		while (1) {

			print (r->ai->getTName (), 10);
			if ((l = r->ai->Id) != NONE) {
				if (r->ai->Class == AIC_port)
					l = l & 0177777;
				Ouf << form ("/%03d ", zz_trunc (l, 3));
			} else
				Ouf << "     ";
			print (r->ai->zz_eid (r->event_id), 10);
			if (r->when > e->waketime)
				Ouf << ' ';
			else if (r == e->chain)
				Ouf << '*';
			else
				Ouf << '?';
			Ouf << ' ';
			print (zz_sn (r->pstate), 10);
			Ouf << ' ';
			ptime (r->when, 15);
			Ouf << '\n';

			if ((r = r->other) == e->chain) break;
		}
	}

	Ouf << "\n(" << getOName ();
	if (Owner != NULL)
		Ouf << '/' << Owner->getOName ();
	Ouf << ") End of list\n\n";
}

void    Process::exDisplay0 (int sid) {

/* ------------------------- */
/* Display the request queue */
/* ------------------------- */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;
	long               l;

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (e->chain == NULL || e->station == NULL ||
			isStationId (sid) && sid != e->station->Id) continue;

		for (r = e->chain;;) {

			if (r->ai == this) {

				dtime (r->when);
				if (e->waketime < r->when)
					display (' ');
				else if (e->chain == r)
					display ('*');
				else
					display ('?');

				if (!isStationId (sid))
				   display (ident (e->station));

				display (e->process->getTName ());
				display (e->process->Id);

				display (zz_eid (r->event_id));

				display (e->process->zz_sn (r->pstate));

				display (e->ai->getTName ());
				if ((l = e->ai->Id) != NONE) {
					if (e->ai->Class == AIC_port)
						l = l & 0177777;
					display (l);
				} else
					display (' ');

				display (e->process->zz_sn (e->chain->pstate));
			}
			if ((r = r->other) == e->chain) break;
		}
	}
}

void    Process::exDisplay1 () {

/* --------------------- */
/* Display the wait list */
/* --------------------- */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;
	long               l;

	for (e = zz_eq; e != zz_sentinel_event; e = e -> next)
		if (e->process == this) break;

	if (e != zz_sentinel_event && (r = e->chain) != NULL) {

		while (1) {

			display (r->ai->getTName ());
			if ((l = r->ai->Id) != NONE) {
				if (r->ai->Class == AIC_port)
					l = l & 0177777;
				display (l);
			} else
				display (' ');

			display (r->ai->zz_eid (r->event_id));
			if (r->when > e->waketime)
				display (' ');
			else if (r == e->chain)
				display ('*');
			else
				display ('?');

			display (zz_sn (r->pstate));
			dtime (r->when);

			if ((r = r->other) == e->chain) break;
		}
	}
}

sexposure (ZZ_KERNEL)

/* ---------------------------------------------------------- */
/* The kernel exposure is used to display things not directly */
/* related to any specific AI                                 */
/* ---------------------------------------------------------- */

	sonpaper {

		sexmode (0)

			exPrint0 (Hdr, (int) SId);  // Full request queue

		sexmode (1)

			exPrint1 (Hdr, (int) SId);  // Abbreviated request queue

		sexmode (2)

			exPrint2 (Hdr);             // Status
	}

	sonscreen {


		sexmode (0)

			exDisplay0 ((int) SId);     // Full request queue

		sexmode (1)

			exDisplay1 ((int) SId);     // Abbreviated request queue

		sexmode (2)

			exDisplay2 ();              // Simulation status

		sexmode (3)

			exDisplay3 ();          // Last event
	}
}

void    ZZ_KERNEL::exPrint2 (char *hdr) {

/* --------------------------- */
/* Print the simulation status */
/* --------------------------- */

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << "(Kernel) Simulation status:\n\n";
	}

	print (zz_processId,  "Smurph process id:              ");
	print (cpuTime ()   , "CPU execution time:             ");
	print (Time         , "Simulated time:                 ");
	print (zz_npre      , "Number of processed events:     ");
	print (zz_npee      , "Number of pending events:       ");
	print (zz_NGMessages, "Number of generated messages:   ");
	print (zz_NRMessages, "Number of received messages:    ");
	print (zz_NQBits    , "Number of queued bits:          ");
	print (MemoryUsed   , "Memory in use:                  ");
	print (Etu * (double)zz_NRBits/((Time == TIME_0) ? TIME_1 : Time)
			     , "Global throughput:              ");

	print ("Last station run:               ", 32);
	if (TheStation == NULL)
		print ("System", 15);
	else
		print (TheStation->getOName (), 15);
	Ouf << '\n';

	print ("Process:                        ", 32);
	if (TheProcess == NULL)
		print ("unknown", 15);
	else
		print (TheProcess->getOName (), 15);
	Ouf << '\n';

	print ("Awakened by:                    ", 32);
	if (zz_CE == NULL)
		print ("unknown AI", 15);
	else
		print (zz_CE->ai->getOName (), 15);
	Ouf << '\n';

	print ("Event:                          ", 32);
	if (zz_CE == NULL)
		print (zz_event_id, 15);
	else
		print (zz_CE->ai->zz_eid (zz_event_id), 15);
	Ouf << '\n';

	print ("Restarted at state:             ", 32);
	if (zz_CE == NULL)
		print (TheState, 15);
	else
		print (TheProcess->zz_sn (TheState), 15);
	Ouf << '\n';

	Ouf << "\n(Kernel) End of list\n\n";
}

void    ZZ_KERNEL::exPrint0 (char *hdr, int sid) {

/* --------------------------------------------- */
/* Print the global request queue (full version) */
/* --------------------------------------------- */

	ZZ_REQUEST                      *r;
	ZZ_EVENT                        *e;
	long                            l, i;

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << "(Kernel) Full event queue";
		if (isStationId (sid))
			Ouf << " for " << idToStation (sid)->getOName ();
		Ouf << ":\n\n";
	}

	if (isStationId (sid))
		Ouf << "     ";
	else
		Ouf << " Stat";

	Ouf << "     Process/Idn               Time           AI/Idn"
		<< "      Event      State\n";

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (isStationId (sid) && ident (e->station) != sid) continue;

		if ((r = e->chain) == NULL) {
			// A system event
			if (zz_flg_nosysdisp) continue;
			if (!isStationId (sid)) {
				print ("Sys", 5);
				Ouf << ' ';
				print (e->process->getTName (), 11);
			} else {
				print (e->process->getTName (), 17);
			}

			Ouf << form ("/%03d ",
					zz_trunc (e->process->Id, 3));

			if (def (e->waketime)) {
				print (e->waketime, 18);
				Ouf << '*';
			} else {
				print ("undefined", 18);
				Ouf << ' ';
			}

			Ouf << ' ';
			print (e->ai->getTName (), 11);
			if ((l = e->ai->Id) != NONE) {
				if (e->ai->Class == AIC_port)
					l = l & 0177777;
				Ouf << form ("/%03d ", zz_trunc (l, 3));
			} else
				Ouf << "     ";
			print (e->ai->zz_eid (e->event_id), 10);
			Ouf << ' ';
			print (e->process->zz_sn (e->pstate), 10);
			Ouf << '\n';

		} else {

			if (e->station == System) continue;
			for (i = 0; ; i++) {

				if (i == 0) {   // The first time around

					if (!isStationId (sid)) {
						if ((l = ident (e->station)) <
							0)
							print ("Sys", 5);
						else
							print (zz_trunc (ident
								(e->station),
									5), 5);
						Ouf << ' ';
						print (e->process->getTName (),
							11);
					} else {
						print (e->process->getTName (),
							17);
					}

					Ouf << form ("/%03d ",
						zz_trunc (e->process->Id, 3));
				} else {

					Ouf << "                      ";
				}

				if (def (r->when)) {
					print (r->when, 18);
					if (r->when > e->waketime)
						Ouf << ' ';
					else if (e->chain == r)
						Ouf << '*';
					else
						Ouf << '?';
				} else {
					print ("undefined", 18);
					Ouf << ' ';
				}

				Ouf << ' ';
				print (r->ai->getTName (), 11);
				if ((l = r->ai->Id) != NONE) {
					if (r->ai->Class == AIC_port)
						l = l & 0177777;
					Ouf << form ("/%03d ", zz_trunc (l, 3));
				} else
					Ouf << "     ";
				print (r->ai->zz_eid (r->event_id), 10);
				Ouf << ' ';
				print (e->process->zz_sn (r->pstate), 10);
				Ouf << '\n';

				if ((r = r->other) == e->chain) break;
			}
		}
	}

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

void    ZZ_KERNEL::exPrint1 (char *hdr, int sid) {

/* ---------------------------------------------------- */
/* Print the global request queue (abbreviated version) */
/* ---------------------------------------------------- */

	ZZ_EVENT                        *e;
	long                            l;

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << "(Kernel) Abbreviated event queue";
		if (isStationId (sid))
			Ouf << " for " << idToStation (sid)->getOName ();
		Ouf << ":\n\n";
	}

	if (isStationId (sid))
		Ouf << "     ";
	else
		Ouf << " Stat";

	Ouf << "     Process/Idn                Time          AI/Idn"
		<< "      Event      State\n";

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (isStationId (sid) && ident (e->station) != sid) continue;
		if ((e->chain == NULL || e->station == System) &&
			zz_flg_nosysdisp) continue;

		if (!isStationId (sid)) {
			if (e->chain == NULL || ident (e->station) < 0)
				print ("Sys", 5);
			else
				print (zz_trunc (ident (e->station), 5), 5);

			Ouf << ' ';
			print (e->process->getTName (), 11);
		} else {
			print (e->process->getTName (), 17);
		}

		Ouf << form ("/%03d ", zz_trunc (e->process->Id, 3));

		ptime (e->waketime, 19);

		Ouf << ' ';
		print (e->ai->getTName (), 11);
		if ((l = e->ai->Id) != NONE) {
			if (e->ai->Class == AIC_port)
				l = l & 0177777;
			Ouf << form ("/%03d ", zz_trunc (l, 3));
		} else
			Ouf << "     ";

		if (e->chain == NULL)
			print (e->process->zz_sn (e->pstate), 10);
		else
			print (e->process->zz_sn (e->chain->pstate), 10);

		Ouf << '\n';
	}

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

void    ZZ_KERNEL::exDisplay2 () {

/* ----------------------------- */
/* Display the simulation status */
/* ----------------------------- */

	display (zz_processId );
	display (cpuTime ()   );
	display (Time         );
	display (zz_npre      );
	display (zz_npee      );
	display (zz_NGMessages);
	display (zz_NRMessages);
	display (zz_NQBits    );
	display (MemoryUsed   );
	display (Etu * (double)zz_NRBits/((Time == TIME_0) ? TIME_1 : Time));
}


void    ZZ_KERNEL::exDisplay0 (int sid) {

/* ----------------------------------------------- */
/* Display the global request queue (full version) */
/* ----------------------------------------------- */

	ZZ_REQUEST                      *r;
	ZZ_EVENT                        *e;
	long                            l, i;

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (isStationId (sid) && ident (e->station) != sid) continue;

		if ((r = e->chain) == NULL) {
			if (zz_flg_nosysdisp) continue;
			// A system event
			if (!isStationId (sid))
				display ("Sys");

			display (e->process->getTName ());
			display (e->process->Id);

			if (def (e->waketime)) {
				display (e->waketime);
				display ('*');
			} else {
				display ("undefined");
				display (' ');
			}

			display (e->ai->getTName ());
			if ((l = e->ai->Id) != NONE) {
				if (e->ai->Class == AIC_port)
					l = l & 0177777;
				display (l);
			} else
				display (' ');

			display (e->ai->zz_eid (e->event_id));
			display (e->process->zz_sn (e->pstate));

		} else {

			if (e->station == System) continue;
			for (i = 0; ; i++) {

				if (i == 0) {   // The first time around

					if (!isStationId (sid)) {
						if ((l = ident (e->station)) <
							0)
							display ("Sys");
						else
						   display (ident (e->station));

					}

					display (e->process-> getTName ());
					display (e->process->Id);

				} else {

					if (!isStationId (sid))
						display (" ");
					display (" ");
					display (" ");
				}

				if (def (r->when)) {
					dtime (r->when);
					if (r->when > e->waketime)
						display (' ');
					else if (e->chain == r)
						display ('*');
					else
						display ('?');
				} else {
					display ("undefined");
					display (' ');
				}

				display (r->ai->getTName ());
				if ((l = r->ai->Id) != NONE) {
					if (r->ai->Class == AIC_port)
						l = l & 0177777;
					display (l);
				} else
					display (' ');

				display (e->ai->zz_eid (r->event_id));
				display (e->process->zz_sn (r->pstate));
				if ((r = r->other) == e->chain) break;
			}
		}
	}
}

void    ZZ_KERNEL::exDisplay1 (int sid) {

/* ------------------------------------------------------ */
/* Display the global request queue (abbreviated version) */
/* ------------------------------------------------------ */

	ZZ_EVENT                        *e;
	long                            l;

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (isStationId (sid) && ident (e->station) != sid) continue;
		if ((e->chain == NULL || e->station == System) &&
			zz_flg_nosysdisp) continue;

		if (!isStationId (sid)) {
			if (e->chain == NULL || ident (e->station) < 0)
				display ("Sys");
			else
				display (ident (e->station));

		}

		display (e->process->getTName ());
		display (e->process->Id);

		dtime (e->waketime);

		display (e->ai->getTName ());
		if ((l = e->ai->Id) != NONE) {
			if (e->ai->Class == AIC_port)
				l = l & 0177777;
			display (l);
		} else
			display (' ');

		if (e->chain == NULL)
			display (e->process->zz_sn (e->pstate));
		else
			display (e->process->zz_sn (e->chain->pstate));
	}
}

void    ZZ_KERNEL::exDisplay3 () {

/* ------------------------------------ */
/* Display information about last event */
/* ------------------------------------ */

	if (TheStation == NULL)
		display ("none");
	else
		display (TheStation->getOName ());

	if (TheProcess == NULL)
		display ("????");
	else
		display (TheProcess->getOName ());

	if (zz_CE == NULL)
		display ("????");
	else
		display (zz_CE->ai->getOName ());

	if (zz_CE == NULL)
		display (zz_event_id);
	else
		display (zz_CE->ai->zz_eid (zz_event_id));

	if (zz_CE == NULL)
		display (TheState);
	else
		display (TheProcess->zz_sn (TheState));
}

void    zz_print_event_info (ostream &os) {

/* -------------------------------------------------- */
/* Prints out information about the last event/status */
/* -------------------------------------------------- */

	os    << ">>> CPU execution time       = " << cpuTime () << '\n';
	os    << ">>> Events processed so far  = " << zz_npre << '\n';
	os    << ">>> Pending events           = " << zz_npee << '\n';
	os    << ">>> Generated messages       = " << zz_NGMessages << '\n';
	os    << ">>> Received messages        = " << zz_NRMessages << '\n';
	os    << ">>> Queued bits              = " << zz_NQBits << '\n';
	os    << ">>> Memory used              = " << MemoryUsed << '\n';
	os    << ">>> Global throughput        = " <<
		((double)zz_NRBits/((Time == TIME_0) ? TIME_1 : Time)) << '\n';
	os    << ">>> Last station run         = ";

	if (TheStation == NULL)
		os << "System";
	else
		os << TheStation->getOName ();
	os    << '\n';

	os    << ">>> Process                  = ";

	if (TheProcess == NULL)
		os << "unknown";
	else
		os << TheProcess->getOName ();

	os    << '\n';

	if (zz_CE != NULL && zz_CE->waketime == Time) {

		os    << ">>> Awakened by              = ";
		os << zz_CE->ai->getOName ();
		os << '\n';

		os    << ">>> Event                    = ";
		os << zz_CE->ai->zz_eid (zz_event_id);
		os    << '\n';

		os    << ">>> Restarted at state       = ";
		os << TheProcess->zz_sn (TheState);
	
		os    << '\n';
	}
}

#if     ZZ_DBG

static  int     header_printed = NO;

void    zz_print_debug_info () {

/* ------------------------------------------------- */
/* Prints out debug information about the last event */
/* ------------------------------------------------- */

	long    l;
	char    *s;

	if (zz_flg_nosysdisp && (zz_CE == NULL ||
		TheStation == System || zz_CE->chain == NULL)) return;
	if (!header_printed) {
		header_printed = YES;
		Ouf <<
"              Time          AI/Idn      Event Station    Process/Idn      State\n";
	}
	print (Time, 18); Ouf << ' ';
	if (zz_CE == NULL)
		print ("unknown    ", 15);
	else {
		print (zz_CE->ai->getTName (), 11);
		if ((l = zz_CE->ai->Id) != NONE) {
			if (zz_CE->ai->Class == AIC_port)
				l = l & 0177777;
			print (form ("/%03d", zz_trunc (l, 3)), 4);
		} else
			Ouf << "    ";
	}

	Ouf << ' ';
	if (zz_CE == NULL)
		print (zz_event_id, 10);
	else
		print (zz_CE->ai->zz_eid (zz_event_id), 10);
	Ouf << ' ';
	if (TheStation == NULL || (l = ident (TheStation)) == NONE)
		print ("System", 7);
	else
		print (zz_trunc (l, 7), 7);
	Ouf << ' ';
	if (TheProcess == NULL) {
		print ("unknown", 10);
		print ("    ");
	} else {
		if ((s = TheProcess->getNName ()) == NULL) {
			print (TheProcess->getTName (), 10);
			print (form ("/%03d", zz_trunc (ident (TheProcess), 3)),
				4);
		} else
			print (s, 14);
	}
	Ouf << ' ';
	if (zz_CE == NULL)
		print (TheState, 10);
	else
		print (TheProcess->zz_sn (TheState), 10);
	Ouf << '\n';

	if (FullTracing) {
		// Link dumps
		int     i;
		char    hdr [256];

		for (i = 0; i < NLinks; i++) {
			strcpy (hdr, form ("Activities in %s:",
				idToLink (i)->getOName ()));

			idToLink (i)->printAct (hdr);
		}
	}
}
#endif
