/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ------------ */
/* The Queue AI */
/* ------------ */

#include        "system.h"

zz_queue::zz_queue () {

/* --------------- */
/* The constructor */
/* --------------- */

	for (int i = 0; i < ZZ_QQSIZE; i++) {
		queuesh [i] = queuest [i] = NULL;
		queuewr [i] = NULL;
	};
	Class = AIC_queue;
	Id = NONE;
};

void    zz_queue::zz_wait (long ev, int pstate) {

/* ---------------------- */
/* The Queue wait request */
/* ---------------------- */

	int    nq;

	if_from_observer ("Queue->wait: called from an observer");

	if (zz_c_first_wait) {

		zz_c_other = NULL;    // No requests so far

		// Create new simulation event
		zz_c_wait_event = new ZZ_EVENT;
		zz_c_wait_event -> station = TheStation;
		zz_c_wait_event -> process = TheProcess;

		// Determine the request Queue

		nq = hashq (TheStation->Id);
		new ZZ_REQUEST (&(queuewr [nq]), this, ev, pstate);

		assert (zz_c_other != NULL,
			"Queue->wait: internal error -- null request");

		zz_c_wait_event -> pstate = pstate;
		zz_c_wait_event -> ai = this;
		zz_c_wait_event -> event_id = ev;
		zz_c_wait_event -> chain = zz_c_other;
		zz_c_wait_event -> Info01 = NULL;
		zz_c_wait_event -> Info02 = NULL;

		zz_c_whead = zz_c_other;

		// Item arrival unknown
		zz_c_wait_event -> waketime =
			zz_c_wait_tmin = zz_c_other -> when = TIME_inf;
		zz_c_other -> event = zz_c_wait_event;
		zz_c_other -> other = zz_c_whead;

		zz_c_first_wait = NO;
		zz_c_wait_event->store ();

	} else {

		// Determine the request Queue

		nq = hashq (TheStation->Id);
		new ZZ_REQUEST (&(queuewr [nq]), this, ev, pstate);

		assert (zz_c_other != NULL,
			"Queue->wait: internal error -- null request");

		zz_c_other -> when = TIME_inf;    // Item arrival time unknown
		zz_c_other -> event = zz_c_wait_event;
		zz_c_other -> other = zz_c_whead;
	}
}

void    *zz_queue::get (int itype) {

/* --------------------------------- */
/* Get an item from Queue type itype */
/* --------------------------------- */

	int        nq;
	ZZ_QITEM   *qq;

	if_from_observer ("Queue->get: called from observer");

	nq = hashq (TheStation->Id);
	for (qq = queuesh [nq]; qq != NULL; qq = qq -> next) {
		if (qq->id == TheStation->Id && (itype == ANY ||
			qq->itype == itype)) {

			// Remove the item from the Queue
			queue_out (qq);
			if (queuesh [nq] == NULL) queuest [nq] = NULL;
			if (queuest [nq] == qq)
				queuest [nq] = qq->prev;
			Info01 = qq->item;
			Info02 = (void*) itype;
			delete (qq);
			return (Info01);
		}
	}
	return (NULL);
}

int     zz_queue::nonempty (int itype) {

/* -------------------------------------------------------------- */
/* Check whether an item of type 'itype' is queued at the station */
/* -------------------------------------------------------------- */

	int        nq;
	ZZ_QITEM   *qq;

	nq = hashq (TheStation->Id);
	for (qq = queuesh [nq]; qq != NULL; qq = qq -> next)
		if (qq->id == TheStation->Id && (itype == ANY ||
			qq->itype == itype)) {

			Info01 = qq->item;
			Info02 = (void*) itype;
			return (YES);
		}
	return (NO);
}

int     zz_queue::put (long station_id, void *item, int itype) {

/* ------------------------------------ */
/* Queue an item at the station's Queue */
/* ------------------------------------ */

	ZZ_QITEM   *qi;
	int        nq, rs;
	ZZ_REQUEST *rq;
	ZZ_EVENT   *e;

	if_from_observer ("Queue->put: called from observer");
	assert (item != NULL, "Queue->put: attempt to queue NULL item");
	assert (isStationId (station_id),
		"Queue->put: illegal station identifier");

	qi = new ZZ_QITEM ((int) station_id, itype, item);

	nq = hashq (station_id);
	if (queuesh [nq] == NULL) {
		assert (queuest [nq] == NULL,
			"Queue->put: internal error - Queue structure screwup");
		qi -> prev = (ZZ_QITEM*) (& (queuesh [nq]));
		queuest [nq] = queuesh [nq] = qi;
	} else {
		qi -> prev = queuest [nq];
		queuest [nq] -> next = qi;
		queuest [nq] = qi;
	}

	// Determine if anybody is waiting for the item

	for (rq = queuewr [nq], rs = PENDING; rq != NULL; rq = rq -> next) {
		if (((rq -> event_id == itype) || (rq -> event_id == ANY)) &&
			(((e = rq -> event) -> station) -> Id == station_id)) {

			rq -> when = Time;
			rs = ACCEPTED;
			if (e -> waketime > Time || flip ()) {
					rq -> Info01 = (void*) item;
					rq -> Info02 = (void*) itype;
					e -> new_top_request (rq);
			}
		}
	}
	return (rs);
}

sexposure (zz_queue)

/* ---------------------------------- */
/* Defines exposures for the Queue AI */
/* ---------------------------------- */

	sonpaper {

		sexmode (0)

			exPrint0 (Hdr, (int) SId);

		sexmode (1)

			exPrint1 (Hdr, (int) SId);

		sexmode (2)

			exPrint2 (Hdr, (int) SId);
	}

	sonscreen {

		sexmode (0)

			exDisplay0 ((int) SId);

		sexmode (1)

			exDisplay1 ((int) SId);

		sexmode (2)

			exDisplay2 ((int) SId);
	}
}

void    zz_queue::exPrint0 (char *hdr, int sid) {

/* --------------------------- */
/* Print the full request list */
/* --------------------------- */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;
	long               l;

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Full AI wait list";
		if (isStationId (sid))
			Ouf << " for " << idToStation (sid)->getOName ();
		Ouf << ":\n\n";
	}

	if (!isStationId (sid))
		Ouf << "       Time   St";
	else
		Ouf << "           Time ";

	Ouf << "    Process/Idn      Queue     QState         AI/Idn" <<
		"      State\n";

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (e->chain == NULL || e->station == NULL) continue;
		if (isStationId (sid) && ident (e->station) != sid) continue;

		for (r = e->chain;;) {

			if (r->ai->Class == AIC_queue) {

				if (!isStationId (sid))
					ptime (r->when, 11);
				else
					ptime (r->when, 15);

				if (e->waketime < r->when)
					Ouf << ' ';     // Obsolete
				else if (e->waketime == r->when)
					Ouf << '?';     // Uncertain
				else
					Ouf << '*';

				if (!isStationId (sid)) {
					// Identify the station
					if (e->station != NULL &&
						ident (e->station) >= 0)
						Ouf << form ("%4d ",
					zz_trunc (ident (e->station), 3));
					else
						Ouf << " Sys ";
				}
				print (e->process->getTName (), 10);
				Ouf << form ("/%03d ",
					zz_trunc (e->process->Id, 3));

				print (zz_eid (r->event_id), 10);
				Ouf << ' ';

				print (e->process->zz_sn (r->pstate), 10);
				Ouf << ' ';

				print (e->ai->getTName (), 10);
				if ((l = e->ai->Id) != NONE) {
					if (e->ai->Class == AIC_port)
						l = l & 0177777;
					Ouf << form ("/%03d ",
						zz_trunc (l, 3));
				} else
					Ouf << "     ";

				print (e->process->zz_sn (e->chain->pstate),10);
				Ouf << '\n';
			}

			if ((r = r->other) == e->chain) break;
		}
	}

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

void    zz_queue::exPrint1 (char *hdr, int sid) {

/* --------------------------------- */
/* Print the abreviated request list */
/* --------------------------------- */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;

	if (hdr != NULL)
		Ouf << hdr << "\n\n";
	else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () <<") Abbreviated AI wait list";
		if (isStationId (sid))
			Ouf << " for " << idToStation (sid)->getOName ();
		Ouf << ":\n\n";
	}

	if (!isStationId (sid))
		Ouf << "                Time     St";
	else
		Ouf << "                      Time ";
	Ouf << "      Process/Ident      Queue        State\n";

	Ouf << "           Time ";
	if (!isStationId (sid)) Ouf << "  St";
	Ouf << "    Process  Id      Queue      State\n";

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (e->chain == NULL || e->station == NULL) continue;
		if (isStationId (sid) && ident (e->station) != sid) continue;

		for (r = e->chain;;) {

			if (r->ai->Class == AIC_queue) {

				if (!isStationId (sid))
					ptime (r->when, 20);
				else
					ptime (r->when, 26);

				if (e->waketime < r->when)
					Ouf << ' ';     // Obsolete
				else if (e->chain == r)
					Ouf << '*';     // Current
				else
					Ouf << '?';

				if (!isStationId (sid)) {
					// Identify the station
					if (e->station != NULL &&
						ident (e->station) >= 0)
						Ouf << form ("%6d ",
					zz_trunc (ident (e->station), 5));
					else
						Ouf << "   Sys ";
				}
				print (e->process->getTName (), 10);
				Ouf << form ("/%05d ",
					zz_trunc (e->process->Id, 5));

				print (Queue->zz_eid (r->event_id), 10);
				Ouf << ' ';

				print (e->process->zz_sn (r->pstate), 12);
				Ouf << '\n';
			}

			if ((r = r->other) == e->chain) break;
		}
	}

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

void    zz_queue::exPrint2 (char *hdr, int sid) {

/* ------------------------ */
/* Print the Queue contents */
/* ------------------------ */

	int             l, i;
	ZZ_QITEM        *q, *qq [ZZ_QQSIZE];
	TIME            t;

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Pending items";
		if (isStationId (sid))
			Ouf << " for " << idToStation (sid)->getOName ();
		Ouf << ":\n\n";
	}

	for (i = 0; i < ZZ_QQSIZE; i++) qq [i] = queuesh [i];

	Ouf << "                Time       Queue";
	if (!isStationId (sid)) Ouf << " Station";
	Ouf << '\n';

	while (1) {

		for (t = TIME_inf, q = NULL, i = 0; i < ZZ_QQSIZE; i++) {
			if (qq [i] == NULL) continue;
			if (qq [i]->RTime < t)
				t = (q = qq [l = i])->RTime;
		}
				
		if (q == NULL) break;           // No more items
		qq [l] = qq [l]->next;
		if (isStationId (sid) && sid != q->id) continue;

		print (q->RTime, 20);
		Ouf << "  ";
		print (Queue->zz_eid (q->itype), 10);
		if (!isStationId (sid)) {
			Ouf << ' ';
			print (q->id, 7);
		}
		Ouf << '\n';
	}

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

void    zz_queue::exDisplay0 (int sid) {

/* ----------------------------- */
/* Display the full request list */
/* ----------------------------- */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;
	long               l;

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (e->chain == NULL || e->station == NULL) continue;
		if (isStationId (sid) && ident (e->station) != sid) continue;

		for (r = e->chain;;) {

			if (r->ai->Class == AIC_queue) {

				dtime (r->when);

				if (e->waketime < r->when)
					display (' ');  // Obsolete
				else if (e->chain == r)
					display ('*');  // Current
				else
					display ('?');

				if (!isStationId (sid)) {
					// Identify the station
					if (e->station != NULL &&
						ident (e->station) >= 0)
						   display (ident (e->station));
					else
						   display ("Sys");
				}
				display (e->process->getTName ());
				display (e->process->Id);
				display (zz_eid (r->event_id));
				display (e->process->zz_sn (r->pstate));
				display (e->ai->getTName ());
				if ((l = e->ai->Id) != NONE) {
					if (e->ai->Class == AIC_port)
						l = l & 0177777;
					display (l);
				} else
					display (' ');

				display (e->process->zz_sn (e->chain->pstate));
			}
			if ((r = r->other) == e->chain) break;
		}
	}
}

void    zz_queue::exDisplay1 (int sid) {

/* ----------------------------------- */
/* Display the abreviated request list */
/* ----------------------------------- */

	ZZ_REQUEST         *r;
	ZZ_EVENT           *e;

	for (e = zz_eq; e != zz_sentinel_event; e = e->next) {

		if (e->chain == NULL || e->station == NULL) continue;
		if (isStationId (sid) && ident (e->station) != sid) continue;

		for (r = e->chain;;) {

			if (r->ai->Class == AIC_queue) {

				dtime (r->when);
				if (e->waketime < r->when)
					display (' ');
				else if (e->chain == r)
					display ('*');
				else
					display ('?');

				if (!isStationId (sid)) {
					// Identify the station
					if (e->station != NULL &&
						ident (e->station) >= 0)
						   display (ident (e->station));
					else
						   display ("Sys");
				}
				display (e->process->getTName ());
				display (e->process->Id);
				display (Queue->zz_eid (r->event_id));
				display (e->process->zz_sn (r->pstate));
			}
			if ((r = r->other) == e->chain) break;
		}
	}
}

void    zz_queue::exDisplay2 (int sid) {

/* -------------------------- */
/* Display the Queue contents */
/* -------------------------- */

	int             l, i;
	ZZ_QITEM        *q, *qq [ZZ_QQSIZE];
	TIME            t;

	for (i = 0; i < ZZ_QQSIZE; i++) qq [i] = queuesh [i];

	while (1) {

		for (t = TIME_inf, q = NULL, i = 0; i < ZZ_QQSIZE; i++) {
			if (qq [i] == NULL) continue;
			if (qq [i]->RTime < t)
				t = (q = qq [l = i])->RTime;
		}
				
		if (q == NULL) break;           // No more items
		qq [l] = qq [l]->next;
		if (isStationId (sid) && sid != q->id) continue;

		display (q->RTime);
		display (Queue->zz_eid (q->itype));
		if (!isStationId (sid))
			display (q->id);
	}
}
