/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* -------------------------------------- */
/* User-invisible macros, constants, etc. */
/* -------------------------------------- */

#include        "global.h"

/* ------------ */
/* Standard AIs */
/* ------------ */

extern  zz_signal       zz_AI_signal;
extern  zz_queue        zz_AI_queue;

/* --------------------------------------- */
/* Global variables (external definitions) */
/* --------------------------------------- */

extern  ZZ_KERNEL       *Kernel,         // The system process
			*ZZ_Kernel;      // Another pointer to it

/* ----------- */
/* Environment */
/* ----------- */

extern  double          Itu, Etu;
extern  int             TheObserverState;       // The observer state
extern  long            MemoryUsed;             // Allocated memory

extern  ZZ_EVENT        *zz_CE;                 // Current event

/* ------------------------------ */
/* Global counters for the Client */
/* ------------------------------ */
extern  long            zz_NQMessages,          // Number of queued messages
			zz_NTMessages,          // Number of transmttd messages
			zz_NRMessages,          // Number of received messages
			zz_NTPackets,           // Number of transmitted packets
			zz_NRPackets,           // Number of received packets
			zz_NGMessages;          // Number of generated messages

extern  BITCOUNT        zz_NQBits,              // Number of queued bits
			zz_NTBits,              // Number of transmitted bits
			zz_NRBits;              // Number of received bits

/* ---------------------------------------------------------- */
/* Temporary list heads for creating station ports and packet */
/* buffers                                                    */
/* ---------------------------------------------------------- */
extern  Port            *zz_ncport;
extern  ZZ_PFItem       *zz_ncpframe;

/* --------- */
/* Debugging */
/* --------- */

extern  int             FullTracing,            // Full trace flag
			EndOfData;              // EOF flag

/* ------------------------------------- */
/* Boundary conditions of the simulation */
/* ------------------------------------- */

extern  long    zz_max_NMessages;               // Default limits and flags
extern  TIME    zz_max_Time;                    // Everything is by default
extern  double  zz_max_cpuTime;                 // unlimited
extern  int     zz_exit_code;                   // Termination code

/* ----------------------- */
/* Program call parameters */
/* ----------------------- */

extern  char    **zz_callArgs;
extern  int     zz_callArgc;                    // Argument count

#if     ZZ_R48

#if     ZZ_ATT
extern  unsigned  short  zz_Seeds [ZZ_N_SEEDS] [ 3 ];
#else
extern  short  zz_Seeds [ZZ_N_SEEDS] [ 3 ];
#endif

#else
extern  long    zz_Seeds [ZZ_N_SEEDS];

		// A divisor for rnd. This is the minimum double number greater 
		// than MAX_long such that (double)MAX_long divided by it is
		// less than 1.0
extern  double  zz_rndd;
#endif
		// Original seeds to be included in error messages; initialized
		// in a standard way (changeable by a call parameter)
extern  long    zz_OrgSeeds [ZZ_N_SEEDS];

		// A flag used to tell whether message queues should be
		// emptied after zz_max_NMessages have been generated
extern  int     zz_flg_emptyMQueues,

		// Tells whether definitions of standard objects should be
		// automatically echoed to the output file
		zz_flg_printDefs,

		// Tells whether the standard Client is to be started. Can be
		// switched off by a program call parameter.
		zz_flg_stdClient,

		// Tells whether information about system events should be
		// suppressed from exposures
		zz_flg_nosysdisp,

		// Tells whether the definition of stations is complete
		zz_flg_started,

		// Tells whether the structure of the ownership three has been
		// modified by zz_adjust_ownership.
		zz_flg_otreeAdjusted;
#if     ZZ_QSL
		// The global limit on the number of messages to be queued at
		// stations awaiting transmission. When this number is reached
		// new generated messages are ignored.
extern  long    zz_QSLimit;
#endif
/* ---------------- */
/* Event processing */
/* ---------------- */

extern  ZZ_EVENT        *zz_c_wait_event;       // Currently built event

extern  long    zz_npre,                        // Number of processed events
		zz_event_id;                    // Current event id

extern  int     zz_c_first_wait,                // Wait for the first time
		zz_port_event_id,               // Last port event id
		zz_chckint;                     // Time limit check interval
						// Also used as checkpoint flag

extern  ZZ_REQUEST *zz_c_whead;                 // Built wait list head

extern  TIME    zz_c_wait_tmin;                 // Earliest request so far

extern  AI      *zz_ai;                         // The current AI pointer

#if     ZZ_OBS

/* ------------------- */
/* Observers internals */
/* ------------------- */

extern  Observer   *zz_obslist;                 // The list of all observers

extern  int        zz_jump_flag,                // Resume in progress
		   zz_state_to_jump;            // The obsrvr state to jump to
#endif

#if     ZZ_TOL
/* -------------------------- */
/* Clock tolerance parameters */
/* -------------------------- */
extern  int    zz_quality;                      // The default clock quality
extern  double zz_tolerance;                    // The default clock tolerance
#endif

/* ---------------------- */
/* Display (visible part) */
/* ---------------------- */

extern  int          DisplayInterval,           // Display interval
		     DisplayOpening,            // Set for a new findow
		     DisplayClosing;            // The window is being closed

/* ------------------------ */
/* Display (invisible part) */
/* ------------------------ */

extern  ZZ_WINDOW       *zz_windows,            // Active windows
			*zz_the_window;         // Current window
extern  ZZ_SIT          *zz_steplist;           // Stepped windows
extern  ZZ_WSIT         *zz_steplist_delayed;   // Waiting to be stepped
extern  ZZ_SUPPLI       *zz_supplist;           // The supplement list (insert)
extern  ZZ_SUPPLD       *zz_suppldst;           // The supplement list (delete)
extern  int             zz_suppcount;           // Supplement count (both)
extern  long            zz_commit_event;        // Starting event number
extern  TIME            zz_commit_time;         // Starting time
extern  volatile long   zz_events_to_display;   // Delay to next display refresh
extern  int             zz_ctimeout,            // Server i/o timeout
			zz_drequest,            // Display requested by '-d'
			zz_send_step_phrase;    // Flag: step condition met

/* ---------------- */
/* Standard streams */
/* ---------------- */

extern  istream *zz_ifpp;                       // Input stream
extern  ostream *zz_ofpp;                       // Output file

extern  char    *zz_ifn,                        // Pointer to input filename
		*zz_ofn;                        // Pointer to output filename

#if     ZZ_SMURPHDEBUG  == 2
extern  ostream *zz_dbgf;
#endif

/* ----------------- */
/* Smurph process id */
/* ----------------- */

extern  int     zz_processId;

/* -------------------------------------------- */
/* Plaintext versions for some hidden constants */
/* -------------------------------------------- */

#define         ACSSIZE                 ZZ_ACSSIZE
#define         EETSIZE                 ZZ_EETSIZE
#define         QQSIZE                  ZZ_QQSIZE

#if     ZZ_ASR
#define         ASSERT                  1
#endif

#if     ZZ_OBS
#define         OBSERVERS               1
#endif

#if     ZZ_AER
#define         AERRORS                 1
#endif

#if     ZZ_TOL
#define         TOLERANCE               1
#endif

#if     ZZ_DBG
#define         DEBUGGING               1
#endif

#define         debug(s)                zz_debug(s)
#define         trace(s)     Ouf << "Time: " << Time << " --> " << s << '\n'

/* ---------------------------------------------------------- */
/* Check  interval:  every  CHCKINT  events smurph will check */
/* against the CPU time limit                                 */
/* ---------------------------------------------------------- */

#define         CHCKINT                 5000

/* --------------------------------------------------- */
/* Scratch non-displayable RVariable (used internally) */
/* --------------------------------------------------- */

class   ZZ_TRVariable : public RVariable {

	friend  class   RVariable;

	public:

	// No additional attributes -- just a separate destructor

#if     ZZ_ATT
	ZZ_TRVariable () { };   // To shut up Cfront
#endif
	~ZZ_TRVariable ();
};

/* ---------------------------------------------------------- */
/* Exposure   macros.  In  the  user  protocol  file  similar */
/* constructs are processed by smpp.                          */
/* ---------------------------------------------------------- */
#define sexposure(otp) void otp::zz_expose (int zz_emode, char *Hdr, long SId){\
	int     zz_em;

#define sonpaper ZZ_ONPP: if ((zz_em = zz_emode) >= ZZ_EMD_0) goto ZZ_ONSC;\
			  if (zz_em == ANY) return; zz_emode = ANY;\
			  switch (zz_em)

#define sonscreen ZZ_ONSC:if ((zz_em = zz_emode) < ZZ_EMD_0) goto ZZ_ONPP;\
			  zz_emode = ANY; zz_em -= ZZ_EMD_0;\
			  switch (zz_em)

#define sexmode(m)  break; case m: zz_dispfound = YES;

/* ------------------------------- */
/* User-invisible global functions */
/* ------------------------------- */
void    zz_init_system (int, char*[]);  // Initialize files and options
void    zz_init_windows ();             // Initialize DDF
void    zz_checkpoint ();               // Checkpoint request
void    zz_start_client ();             // Starts the standard Client
#if     DEBUGGING
void    zz_print_debug_info ();         // One line of debugging information
#endif
void    zz_print_event_info (ostream&); // Prints out last event information
long    zz_trunc (long, int);           // Truncate the number to int digits
void    zz_dterminate ();               // Sends termination phrase to server
void    zz_ptime (TIME&, int);          // Prints out time values

#define ptime(a,b)      zz_ptime (a,b)  // A system alias for the above

/* ------------------------------ */
/* A macro to display time values */
/* ------------------------------ */
#define dtime(a)        do { if (def (a)) display(a); else\
			display ("undefined"); } while (0)

/* -------------------------- */
/* The number of random seeds */
/* -------------------------- */
#define         N_SEEDS         ZZ_N_SEEDS

/* ------------------------------------ */
/* Number of (user-visible) link events */
/* ------------------------------------ */
#define         N_LINK_EVENTS           ZZ_N_LINK_EVENTS

/* ------------------------------ */
/* Protocol-invisible link events */
/* ------------------------------ */
#define         LNK_PURGE       N_LINK_EVENTS
#define         ARC_PURGE       (N_LINK_EVENTS+1)
#define         BOT_HEARD       (N_LINK_EVENTS+2)
#define         EOT_HEARD       (N_LINK_EVENTS+3)

/* ------------- */
/* Client events */
/* ------------- */
#define         ARR_MSG                 -3
#define         ARR_BST                 -4

/* ---------------------------------------------------------- */
/* A  virtual event generated by the process AI -- to start a */
/* process for the first time                                 */
/* ---------------------------------------------------------- */
#define         START                   (-2)

#if     OBSERVERS
/* ------------------------- */
/* Observer AI service event */
/* ------------------------- */
#define         OBS_TSERV               0
#endif

/* ------------------------------------- */
/* Protocol-invisible bus activity types */
/* ------------------------------------- */
#define         TRANSFER_A              5       // Transfer attempt
#define         TRANSFER_J              6       // Followed by a jam

/* ---------------------------------------------------------- */
/* Special  request  'id'  attribute  values  to tell entries */
/* requiring special postprocessing upon deallocation         */
/* ---------------------------------------------------------- */
#define         RQTYPE_SIG              (-2)    // Signal
#define         RQTYPE_PRC              (-3)    // Waiting for process

/* -------------------- */
/* Checkpoint constants */
/* -------------------- */

/* ----------------------------------------- */
/* The index of the stack pointer in jmp_buf */
/* ----------------------------------------- */

#ifdef  vax
#define         CENVSP          2       // For the VAX
#endif

#ifdef  mips
#define         CENVSP          32      // For the MIPS (non-standard buffer)
#endif

#ifdef  mc68000
#if     _JBLEN == 58                    // Sun 3/80
#define         CENVSP          2
#else
#define         CENVSP          14      // Sun 3/.. 2/..
#endif
#endif

#ifdef  sparc
#define         CENVSP          2       // Sun 4
#endif


#define         SIGCHKP         SIGUSR1         // Checkpoint signal
#define         OMASK           0022            // Output file mask
#define         CMAGIC          ((char)(-2))    // The 'magic' character

/* ------------------------------------------------- */
/* Base port number for user accessible socket ports */
/* ------------------------------------------------- */
#define         SOCK_BASE       2000

/* --------------- */
/* Observer macros */
/* --------------- */
#if     ZZ_OBS
#define         if_not_from_observer(t) assert (zz_observer_running, t)
#define         if_from_observer(t)     assert (!zz_observer_running, t)
#else
#define         if_not_from_observer(t) ;
#define         if_from_observer(t)     ;
#endif


/* -------------------------------------------- */
/* Convenient aliases for user-invisible macros */
/* -------------------------------------------- */

/* -------------------- */
/* Operations on queues */
/* -------------------- */
#define         queue_head(item,hd,type) zz_queue_head(item,hd,type)
#define         queue_out(item) zz_queue_out(item)

/* -------------------------------------------------- */
/* A variant of queue_head for queueing AI subclasses */
/* -------------------------------------------------- */
#define         queue_head_AI(item,hd,type)     do {(item)->next = (hd);\
			 (item)->prev = (type*)(((char*)(&(hd)))-sizeof(AI));\
	if ((hd) != NULL) (hd)->prev = (item);\
	(hd) = (item);} while (0)

#define         ierror(a)       zz_ierror(a)
#define         aerror(a)       zz_aerror(a)

/* ----------------------------------------------------------------- */
/* Hash function for identifying QUEUE AI's queues and request lists */
/* ----------------------------------------------------------------- */
#define hashq(s)        ((int)((s)&077))

#if     ZZ_ATT
#define narray(name,type,size)  type *name = new type [size]
#define darray(name)            delete (name)
#define STATIC
#define USESID                  zz_Stroustrup = SId
#define USEEMD                  zz_em = zz_emode; zz_Stroustrup = (long) Hdr
#else
#define narray(name,type,size)  type name [size];
#define darray(name)
#define STATIC  static
#define USESID
#define USEEMD
#endif

#ifdef  sgi
typedef void (*SIGARG) (int, ...);
#else
#ifdef  ZZ_ATT_SUN
#define SIGARG  SIG_PF
#else
typedef void (*SIGARG) ();
#endif
#endif

#ifdef  ZZ_ATT_SUN
#define RWCAST (char*)
#else
#define RWCAST
#endif

