/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */

/* ---------------------------------------------------------- */
/* This  is  a  preprocessor running after GNU cpp and before */
/* the C++ compiler whose purpose  is  to  turn  some  SMURPH */
/* constructs into C++ statements                             */
/* ---------------------------------------------------------- */

#include        "smpp.h"

#if     ZZ_ATT
#define INLINE
#else
#define INLINE  inline
#endif

static INLINE int hash (char *s) {

/* -------------------------------------------- */
/* Calculates the hash code of a symbol/keyword */
/* -------------------------------------------- */

	register int    res;

	res = *s++;
	if (*s == '\0') return (res & (HASHTS-1));
	res += *s++;
	if (*s == '\0') return (res & (HASHTS-1));
	return ((res + *s) & (HASHTS-1));
}

#ifdef DEBUG
int     Nodebug;
void    debug (char *t) {

/* ----------------------- */
/* Print out debug message */
/* ----------------------- */

	char    *s;

	if (Nodebug) return;
	write (STDERR, CFName, strlen (CFName));
	s = form (":%1d: %s\n", LineNumber, t);
	write (STDERR, s, strlen (s));
}
#else
#define debug(t)
#endif

void    ioerror (int fd) {

/* ------------------------- */
/* Process i/o error (fatal) */
/* ------------------------- */

	char    *s;

	s = form ("i/o error on %s", fd == STDIN ? "stdin" : "stdout");
	perror (s);
	exit (2);
}

#if     ZZ_OWNLIB == 1
void    excptn (char *t) {

/* ----------------------- */
/* Memory allocation error */
/* ----------------------- */

	char    *s;

	s = form ("Memory allocation error: %s\n", t);
	write (STDERR, s, strlen (s));
	exit (3);
}
#endif

SymDesc::SymDesc (char *sym, int hsh, int qual, int stts, int mode, int estat) {

/* ----------------------------------------- */
/* Constructor for a symbol dictionary entry */
/* ----------------------------------------- */

	next = SymTab [hsh];
	SymTab [hsh] = this;

	Name = new char [strlen (sym) + 1];
	strcpy (Name, sym);
	Type = qual;
	Status = stts;
	EStat = estat;
	PAnn = NO;
	Mode = mode;
	States = NULL;
}

SymDesc *addSym (char *sym, int qual, int stts, int mode, int estat) {

/* --------------------------------- */
/* Adds a symbol to the symbol table */
/* --------------------------------- */

	SymDesc *cs;
	int     h;

	for (cs = SymTab [h = hash (sym)]; cs != NULL; cs = cs->next) {
	    if (strcmp (cs->Name, sym) == 0) {
	      if (cs->Type != qual) {
		xerror (form ("'%s' conflicitng declarations", sym));
		return (NULL);
	      }
	      if (cs->Status == DEFINED) {
		if (stts == DEFINED) {
		  xerror (form ("'%s' defined more than once", sym));
		  return (NULL);
		}
	      } else {
		cs->Status = stts;
	      }
	      if (cs->Mode != FREE) {
		if (mode != FREE && mode != cs->Mode) {
		  xerror (form ("'%s' mode conflicts with previous declaration",
		    sym));
		  return (NULL);
		}
	      } else {
		cs->Mode = mode;
	      }
	      if (cs->EStat != estat) {
		if (cs->EStat == UNEXPOSABLE) {
		  xerror (form ("'%s' is unexposable", cs->Name));
		  return (NULL);
		}
		if (estat == ANNOUNCED) cs->EStat = ANNOUNCED;
	      }
	      return (cs);
	    }
	}

	return new SymDesc (sym, h, qual, stts, mode, estat);
}

INLINE SymDesc *getSym (char *sym) {

/* ------------------------------------------------- */
/* Gets the symbol description from the symbol table */
/* ------------------------------------------------- */

	SymDesc *cs;
	int     h;

	for (cs = SymTab [h = hash (sym)]; cs != NULL; cs = cs->next)
	  if (strcmp (cs->Name, sym) == 0) return (cs);
	return (NULL);
}

inline  int     isDefined (SymDesc *ob, int qual, int stts) {

/* ---------------------------------------------- */
/* Checks whether the symbol is defined as needed */
/* ---------------------------------------------- */

	return (ob->Type == qual && (ob->Status == DEFINED ||
		stts == ANNOUNCED));
}

KeyDesc::KeyDesc (char *s, KFUNC f, int arr) {

/* ---------------------------------------------------------- */
/* Adds   a  new  keyword  to  the  keyword  table  (used  at */
/* initialization)                                            */
/* ---------------------------------------------------------- */

	int     h;

	PFunct = f;
	Name = new char [strlen (s) + 1];
	Arrow = arr;
	strcpy (Name, s);
	h = hash (s);
	next = KeyTab [h];
	KeyTab [h] = this;
}

INLINE KFUNC getKey (char *s) {

/* ---------------------------------------- */
/* Gets the function for processing the key */
/* ---------------------------------------- */

	register int     h;
	register KeyDesc *ck;

	ck = KeyTab [h = hash (s)];
	for (; ck != NULL; ck = ck->next)
		if ((ArrowInFront == 2) == ck->Arrow && strcmp (ck->Name, s)
			== 0) return (ck->PFunct);
	return (NULL);
}
	
void    readBuf () {

/* ------------------------------------ */
/* Fills the input buffer with new data */
/* ------------------------------------ */

	int     n, m;

	if (Eof) return;                        // Don't read past EOF
	if (InP < OutP) {
		if (InP + 1 == OutP)            // The buffer is full
			return;
		if ((m = read (STDIN, InP, n=OutP-InP-1)) < 0) ioerror (STDIN);
		InP += m;
		if (m == 0) Eof = YES;
		return;
	}
	n = LimitP - InP;
	if (OutP == FirstP) {
		if (--n <= 0)           // The buffer is full
			return;
		if ((m = read (STDIN, InP, n)) < 0) ioerror (STDIN);
		InP += m;
		if (m == 0) Eof = YES;
		return;
	}
	if ((m = read (STDIN, InP, n)) < 0) ioerror (STDIN);
	if (m < n) {
		InP += m;
		if (m == 0) Eof = YES;
		return;
	}
	if (OutP == FirstP+1) {         // No second part
		InP = FirstP;
		return;
	}
	if ((m = read (STDIN, FirstP, n=OutP-FirstP-1)) < 0) ioerror (STDIN);
	InP = FirstP + m;
	return;
}

static INLINE int getC () {

/* ------------------------------------- */
/* Reads a new character from the buffer */
/* ------------------------------------- */

	if (BOLine = (CurrChar == '\n')) LineNumber++;;
	if (OutP == InP) {              // The buffer is empty
		readBuf ();
		if (Eof) return (END);
	}
	CurrChar = *OutP++;
	if (OutP >= LimitP) OutP = FirstP;
	return (CurrChar);
}

static INLINE int peekC () {

/* ----------------------------------------- */
/* Peeks at the next character in the buffer */
/* ----------------------------------------- */

	if (OutP == InP) {              // The buffer is empty
		readBuf ();
		if (Eof) return (END);
	}
	return (*OutP);
}

void    xerror (char *t) {

/* ----------------------- */
/* Print out error message */
/* ----------------------- */

	char    *s;

	ErrorsFound++;
	write (STDERR, CFName, strlen (CFName));
	s = form (":%1d: %s\n", LineNumber, t);
	write (STDERR, s, strlen (s));
	if (ErrorsFound < MAXERRCNT) return;
	// That's it: ignore the rest of the program
	write (STDERR, CFName, strlen (CFName));
	s = form (":%1d: too many errors\n", LineNumber);
	write (STDERR, s, strlen (s));
	while (getC () != END);
	exit (1);
}

int     startLkp () {

/* ------------------------------------ */
/* Starts advance lookup for characters */
/* ------------------------------------ */

	if (OutP == InP) {              // The buffer is empty
		readBuf ();
		if (Eof) return (END);
	}
	EnLkpPtr = OutP;
	return (YES);
}

static INLINE int lkpC () {

/* --------------------------------------- */
/* Lookup the next character in the buffer */
/* --------------------------------------- */

	char    c;

	if (EnLkpPtr == InP) {          // End of buffer reached
		if (Eof) return (END);
		readBuf ();
		if (EnLkpPtr == InP) {
			if (Eof) return (END);
			// Lookup exceeds buffer capacity
			xerror ("construct too long");
			return (ERROR);
		}
	}
	c = *EnLkpPtr++;
	if (EnLkpPtr >= LimitP) EnLkpPtr = FirstP;
	if (c == '\n') SkLineCnt++;
	return (c);
}

void    catchUp () {

/* ------------------------------------------------ */
/* Advance the out pointer to the lookahead pointer */
/* ------------------------------------------------ */

	while (OutP != EnLkpPtr) getC ();
}

void    flushBuf () {

/* ----------------------------------------------- */
/* Writes the output buffer to the standard output */
/* ----------------------------------------------- */

	if (FreeP == StartP) return;    // Nothing to write
	if (write (STDOUT, StartP, FreeP - StartP) < 0) ioerror (STDOUT);
	FreeP = StartP;
}

static inline void putC (char c) {

/* --------------------------------------- */
/* Writes one character to the output file */
/* --------------------------------------- */

	if (FreeP >= EndP) flushBuf ();
	*FreeP++ = c;
}

void    putC (char *s) {

/* ---------------------------------- */
/* Writes a string to the output file */
/* ---------------------------------- */

	while (*s != '\0') putC (*s++);
}

void    finish () {

/* ----------------------- */
/* Completes the execution */
/* ----------------------- */

	SymDesc *or;

	// Check if should append the starter function
	if ((or = getSym ("Root")) != NULL && or->Type == PROCESS &&
	    or->Status == DEFINED) {
		putC ("void ZZRoot () { ZZ_Main = (Process*) new Root; ");
		putC ("((Process*)ZZ_Main)->zz_start (); ");
		putC ("((Process*)ZZ_Main)->setup (); }\n");
	}
		
	flushBuf ();
	exit (ErrorsFound != 0);
}

int     pushDecl (SymDesc *co) {

/* ------------------------------------------ */
/* Push a new object declaration on the stack */
/* ------------------------------------------ */

	if (DeclStackDepth >= MSTACK) {
		xerror ("declarations overnested");
		return (NO);
	}

	DeclStack [DeclStackDepth] . CObject = CObject;
	DeclStack [DeclStackDepth] . CObjectBC = CObjectBC;

	DeclStackDepth ++;

	CObject = co;
	CObjectBC = BraceLevel;
	return (YES);
}

void    processLineCommand () {

/* ---------------------------------------------------------- */
/* Processes the line command of the form # ddd "sssss" [nnn] */
/* ---------------------------------------------------------- */

	int     nln,    // New line number
		bpt;    // File name buffer pointer
	char    c;

	// Copy the '#' character
	putC ('#');
	if (startLkp () == END) {
		xerror ("input file ends prematurely");
		finish ();
	}

	// Line number

	while ((c = lkpC ()) != '\n') if (!isspace (c)) break;
	if (c == '\n' || ! isdigit (c)) goto Copy;
	for (nln = c - '0'; (c = lkpC ()) && isdigit (c); nln = nln*10+(c-'0'));

	// File name

	while (isspace (c) && c != '\n') c = lkpC ();
	if (c != '"') goto CopyAl;

	for (bpt = 0; ;) {
		c = lkpC ();
		if (c == '\n' || c == END || c == ERROR) goto CopyAl;
		if (c == '"') break;
		if (bpt < MFNLEN-1) CFName [bpt++] = c;
	}
	CFName [bpt] = '\0';

	// We don't care about anything else

	LineNumber = nln-1;
	// This number will be incremented properly when we process the
	// end of line

CopyAl:
#if     ZZ_ATT
	if (c == '$') {
		// #pragma
		while (getC () != '$');
		LineNumber--;
		goto Copy;
	}
	putC ("line");
#endif
Copy:   // Copy the rest of the line until EOL

	while ((c = getC ()) != '\n') {
		if (c == END) finish ();
		putC (c);
	}
	putC (c);
}

int     getKeyword (char *kbuf, int fc = 0, int mswd = 0) {

/* ---------------------------------------------------------- */
/* Gets  a  keyword  into  buffer kbuf. Returns the non-blank */
/* delimiter code, ERROR or END.                              */
/* ---------------------------------------------------------- */

	int     i, c;

	// fc is the first character. If it is 0, we assume that we have to
	// start from scratch

	if (fc == 0) {
		// Skip white spaces
		while (1) {
			fc = lkpC ();
			if (!isspace (fc)) break;
		}
	}

	if (fc == END || fc == ERROR) return (fc);

	if (!isalpha(fc) && fc != '_' && fc != '$' && (!mswd || !isdigit(fc))) {
		xerror ("identifier expected");
		return (ERROR);
	}

	kbuf [0] = fc;

	for (i = 1; ; i++) {
		c = lkpC ();
		if (!isalnum (c) && c != '_' && c != '$') break;
		if (i > MAXKWDLEN) {
			xerror ("identifier too long");
			return (ERROR);
		}
		kbuf [i] = c;
	}

	kbuf [i] = '\0';
	while (isspace (c)) c = lkpC ();
	return (c);
}

int     getArg (char *kbuf) {

/* ---------------------------------------------------------- */
/* Gets  an argument (which can be an expression) into buffer */
/* kbuf. Returns the non-blank delimiter code, ERROR or END.  */
/* ---------------------------------------------------------- */

	int     i, c, pc, ws;

	// Skip white spaces
	while (1) {
		c = lkpC ();
		if (!isspace (c)) break;
	}

	if (c == END || c == ERROR) return (c);

	if (c == ',' || c == ')' || c == ';') {
		xerror ("argument missing");
		return (ERROR);
	}

	kbuf [0] = c;
	if (c == '"' || c == '\'')
		ws = c;                 // Within string
	else
		ws = 0;

	if (c == '(')
		pc = 1;                 // Parentheses count
	else
		pc = 0;
	
	for (i = 1; ; ) {
		c = lkpC ();
		if (ws) {
			if (c == '\\') {
				if (i > MAXKWDLEN) {
					xerror ("argument too long");
					return (ERROR);
				}
				kbuf [i++] = c;
				c = lkpC ();
			} else {
				if (c == ws) ws = 0;
			}
		} else {
			if (isspace (c)) continue;
			if (c == ',' || c == ';') {
				if (pc == 0) break;
			} else if (c == ')') {
				if (pc-- == 0) break;
			} else if (c == '(') {
				pc++;
			} else if (c == '\'' || c == '"') {
				ws = c;
			}
		}
			
		if (c == ERROR) return (NO);
		if (c == END) return (c);
		if (i > MAXKWDLEN) {
			xerror ("argument too long");
			return (ERROR);
		}
		kbuf [i++] = c;
	}

	kbuf [i] = '\0';
	while (isspace (c)) c = lkpC ();
	return (c);
}

int     isState (char *st) {

/* --------------------------------------------------------- */
/* Checks if the state 'st' occurs on the process state list */
/* --------------------------------------------------------- */

	char    **sl;

	if ((sl = CodeObject->States) == NULL) {
		xerror (form ("no states declared for '%s'", CodeObject->Name));
		return (NO);
	}

	for ( ; *sl != NULL; sl++)
		if (strcmp (*sl, st) == 0) break;

	if (*sl == NULL) {
	  xerror (form ("'%s' is not a state for '%s'", st, CodeObject->Name));
	  return (NO);
	}

	return (YES);
}

int     processWait (int del) {

/* ----------------------------------------- */
/* Expands the ...->wait (a, b) ... sequence */
/* ----------------------------------------- */

	// wait (a, b) expands into 'zz_wait (a, b)'

	char    arg [MAXKWDLEN+1];
	int     lc;

	if (CodeType != PROCESS) return (NO);
	if (del != '(') return (NO);

	// Get the first argument

	lc = getArg (arg);
	if (lc == END || lc == ERROR) return (NO);
	if (lc != ',') {
		xerror ("event identifier missing");
		return (NO);
	}

	putC ("zz_wait (");
	putC (arg);
	putC (", ");

	// The second argument should be an identifier

	lc = getKeyword (arg);
	if (lc == ERROR) return (NO);
	if (lc == END) {
		xerror ("file ends in the middle of 'wait'");
		return (NO);
	}
	if (lc != ')') {
		xerror ("illegal state identifier");
		return (NO);
	}

	if (!isState (arg)) return (NO);

	putC (arg);
	putC (')');

	catchUp ();     // Catch up with the lookahead pointer
	return (YES);
}

int     processTerminate (int del) {

/* -------------------- */
/* Expands 'terminate;' */
/* -------------------- */

	// terminate; and terminate (); expands into 'terminate (); sleep;'

	if (!CodeType) return (NO);
	if (del != ';') {
		if (del != '(') return (NO);
		for (del = lkpC (); isspace (del); del = lkpC ());
		if (del != ')') return (NO);
		for (del = lkpC (); isspace (del); del = lkpC ());
		if (del != ';') return (NO);
	}

	putC ("do { terminate (); return; } while (0);");
	catchUp ();     // Catch up with the lookahead pointer
	return (YES);
}

int     processTransmit (int del) {

/* --------------------------------------------- */
/* Expands the ...->transmit (a, b) ... sequence */
/* --------------------------------------------- */

	// transmit (a, b) expands into 'zz_transmit (a, b)'

	char    arg [MAXKWDLEN+1];
	int     lc;

	if (CodeType != PROCESS) return (NO);
	if (del != '(') return (NO);

	// Get the first argument

	lc = getArg (arg);
	if (lc == END || lc == ERROR) return (NO);
	if (lc != ',') {
		xerror ("event identifier missing");
		return (NO);
	}

	putC ("zz_transmit (");
	putC (arg);
	putC (", ");

	// The second argument should be an identifier

	lc = getKeyword (arg);
	if (lc == ERROR) return (NO);
	if (lc == END) {
		xerror ("file ends in the middle of 'transmit'");
		return (NO);
	}
	if (lc != ')') {
		xerror ("illegal state identifier");
		return (NO);
	}

	if (!isState (arg)) return (NO);

	putC (arg);
	putC (')');

	catchUp ();     // Catch up with the lookahead pointer
	return (YES);
}

int     processSendJam (int del) {

/* -------------------------------------------- */
/* Expands the ...->sendJam (a, b) ... sequence */
/* -------------------------------------------- */

	// sendjam (a, b)
	//
	// expands into: 'startJam (), zz_AI_timer.zz_wait (a, b)'

	char    arg [MAXKWDLEN+1];
	int     lc;

	if (CodeType != PROCESS) return (NO);
	if (del != '(') return (NO);

	// Get the first argument

	lc = getArg (arg);
	if (lc == END || lc == ERROR) return (NO);
	if (lc != ',') {
		xerror ("event identifier missing");
		return (NO);
	}

	putC ("startJam (), zz_AI_timer.zz_wait (");
	putC (arg);
	putC (", ");

	// The second argument should be an identifier

	lc = getKeyword (arg);
	if (lc == ERROR) return (NO);
	if (lc == END) {
		xerror ("file ends in the middle of 'sendJam'");
		return (NO);
	}
	if (lc != ')') {
		xerror ("illegal state identifier");
		return (NO);
	}

	if (!isState (arg)) return (NO);

	putC (arg);
	putC (')');

	catchUp ();     // Catch up with the lookahead pointer
	return (YES);
}

int     processIdentify (int del) {

/* ---------------------------------- */
/* Expands the `identify t;' sequence */
/* ---------------------------------- */

	// identify (....);
	// identify ...;
	// identify "...";
	//
	// expands into: 'ProtocolId = "...";'

	int     lc, pc;

	if (CodeType) return (NO);

	lc = del;
	if (lc == ERROR) return (NO);
	if (lc == END) {
Endfile:
		xerror ("file ends in the middle of 'identify'");
		return (NO);
	}

	if (Identified) {
		xerror ("'identify' may occur only once in protocol file");
		return (NO);
	}

	Identified = YES;

	putC ("char *ProtocolId = \"");
	if (lc == '(') {
		pc = 1;
	} else {
		pc = 0;
		if (lc == '"') {
			pc = -1;
		} else {
			putC (lc);
		}
	}

	while (1) {
	    switch (lc = lkpC ()) {
		case ';' :
			if (pc) {
				putC (lc);
				continue;
			}
			goto Done;
		case '(' :
			if (pc > 0) {
				pc++;
				continue;
			}
			putC (lc);
			continue;
		case ')' :
			if (pc > 0) {
				if (--pc == 0) {
					lc = lkpC ();
					goto Complete;
				}
				continue;
			}
			putC (lc);
			continue;
		case ' ' :
			if (pc) {
				putC (lc);
				continue;
			}
			goto Complete;
		case '\\' :
			lc = lkpC ();
			if (lc == '"') putC ('\\');
			if (lc == ERROR) return (NO);
			if (lc == END) goto Endfile;
			putC (lc);
			continue;
		case '"' :
			if (pc < 0) {
				lc = lkpC ();
				goto Complete;
			}
			putC ('\\');
			putC ('"');
			continue;
		case ERROR :
			return (NO);
		case END :
			goto Endfile;
		default:
			putC (lc);
			continue;
	    }
	}

Complete:       // Expect only spaces terminated by ';'

	while (isspace (lc)) lc = lkpC ();
	if (lc == ERROR) return (NO);
	if (lc != ';') {
		xerror ("identify syntax error");
		return (NO);
	}

Done:   putC ("\";");
	catchUp ();
	return (YES);
}

int     processState (int del) {

/* ------------------------------------- */
/* Expands the `state nnnnnn :' sequence */
/* ------------------------------------- */

	// state sym :
	//
	// expands into: 'break; case sym :'

	char    arg [MAXKWDLEN+1];
	int     lc;

	if (! CodeType) return (NO);

	// Get the state identifier

	lc = getKeyword (arg, del);
	if (lc == END || lc == ERROR) return (NO);
	if (lc != ':') {
		xerror ("state syntax error");
		return (NO);
	}

	if (!isState (arg)) return (NO);

	if (CodeWasState) {
		putC ("break; ");
	} else {
		if (CodeType == PROCESS)
			putC ("switch (TheState) { ");
		else
			putC ("switch (TheObserverState) { ");
			CodeWasState = YES;
	}
	putC ("case ");
	putC (arg);
	putC (":");

	catchUp ();     // Catch up with the lookahead pointer
	return (YES);
}

int     processTransient (int del) {

/* ------------------------------------------------------- */
/* Expands the controversial `transient nnnnnn :' sequence */
/* ------------------------------------------------------- */

	char    arg [MAXKWDLEN+1];
	int     lc;

	// transient state :
	//
	// expands into: 'case state :'

	if (! CodeType) return (NO);
	if (! CodeWasState) {
	  xerror ("illegal 'transient' before first 'state' in this 'perform'");
		return (NO);
	}

	// Get the state identifier

	lc = getKeyword (arg, del);
	if (lc == END || lc == ERROR) return (NO);
	if (lc != ':') {
		xerror ("state syntax error");
		return (NO);
	}

	if (!isState (arg)) return (NO);

	putC ("case ");
	putC (arg);
	putC (":");

	catchUp ();     // Catch up with the lookahead pointer
	return (YES);
}

int     processSignals (int del) {

/* ----------------------------------- */
/* Expands the `signal {...}' sequence */
/* ----------------------------------- */

	// signals {s0, ..., sn};
	//
	// expands into: 'enum {s0, ..., sn};'

	char    arg [MAXKWDLEN+1];
	int     lc;
	
	lc = del;
	if (lc != '{') return (NO);

	putC ("enum {");

	while (1) {
		lc = getKeyword (arg);
		if (lc == ERROR) return (NO);
		if (lc == END) {
		    xerror ("file ends in the middle of 'signals' declaration");
		    return (NO);
		}
		putC (arg);
		if (lc == ',') {
			putC (',');
		} else if (lc == '}') {
			putC ('}');
			catchUp ();
			return (YES);
		} else {
			xerror ("illegal delimiter in 'signals' declaration");
			return (NO);
		}
	}
}

int     processStates (int del) {

/* ------------------------------- */
/* Expands `states {...}' sequence */
/* ------------------------------- */

	// states {s0, ..., sn};
	//
	// expands into: 'enum {s0, ..., sn};'

	char    arg [MAXKWDLEN+1];
	int     i, lc, ns, sls;
	char    **sl, **tmp;
	
	lc = del;
	if (lc != '{') return (NO);
	if (CObject == NULL || (CObject->Type != PROCESS && CObject->Type !=
	  OBSERVER) || BraceLevel != CObjectBC+1) {
	    xerror ("'states' declaration outside of process/observer scope");
	    return (NO);
	}

	if (CObject->States != NULL) {
		xerror (form ("duplicate 'states' declaration for '%s'",
			CObject->Name));
		return (NO);
	}

	sls = 16;       // Size of the state table
	ns = 0;         // The number of states

	for (sls = 15, ns = 0, sl = new char* [sls+1]; ; ) {
		lc = getKeyword (arg);
		if (lc == ERROR) { delete (sl); return (NO); }
		if (lc == END) {
		    xerror ("file ends in the middle of 'states' declaration");
		    delete (sl);
		    return (NO);
		}

		// Check against double occurrence
		for (i = 0; i < ns; i++) {
		  if (strcmp (sl [i], arg) == 0) {
		    xerror (form ("state '%s' multiply defined", arg));
		    delete (sl);
		    return (NO);
		  }
		}
		
		if (ns >= sls) {
			// Reallocate the state array
			tmp = new char* [ns];
			for (i = 0; i < ns; i++) tmp [i] = sl [i];
			delete (sl);
			sl = new char* [(sls = sls+sls+1) + 1];
			for (i = 0; i < ns; i++) sl [i] = tmp [i];
		}

		sl [ns] = new char [strlen (arg) + 1];
		strcpy (sl [ns], arg);
		ns++;

		if (lc == '}') break;
		if (lc != ',') {
			xerror ("illegal delimiter in 'states' declaration");
			delete (sl);
			return (NO);
		}
	}
	sl [ns] = NULL; // Sentinel
	// Skip to the semicolon which must be the next nonspace character
	del = lkpC ();
	while (isspace (del)) del = lkpC ();
	if (del != ';') {
		xerror ("states declaration not terminated with ';'");
		delete (sl);
		return (NO);
	}

	// Now expand the declaration
	CObject->States = sl;
	if (ns) {
		putC ("enum {");
		for (tmp = sl; ;) {
			putC (*tmp);
			tmp++;
			if (*tmp == NULL) {
				putC ("}; ");
				break;
			} else {
				putC (", ");
			}
		}
	}

	putC ("static const int zz_ns");
	putC (SPref);
#if     ZZ_ATT
	putC ("; ");
#else
	putC (form (" = %1d; ", ns));
#endif
	putC ("static char *zz_sl");
	putC (SPref);
#if     ZZ_ATT
	putC (form ("[%1d];", ns));
#else
	putC (" [] = {");
	for (tmp = sl; ; ) {
		putC ('"');
		putC (*tmp);
		putC ('"');
		if (*(++tmp) == NULL) break;
		putC (", ");
	}
	putC ("}; ");
#endif
	putC ("virtual char *zz_sn (int n) { return ((n > zz_ns");
	putC (SPref);
	putC (") ? \"undefined\" : zz_sl");
	putC (SPref);
	putC (" [n]); }; ");

	catchUp ();
	return (YES);
}

int     processSleep (int del) {

/* ------------- */
/* Expands sleep */
/* ------------- */

	// sleep;
	//
	// expands into 'return;'

	if (! CodeType) return (NO);
	if (del != ';') {
		xerror ("'sleep' syntax error");
		return (NO);
	}

	putC ("return;");
	catchUp ();
	return (YES);
}

int     processProceed (int del) {

/* --------------------------------- */
/* Expands the `proceed s;' sequence */
/* --------------------------------- */

	// proceed state;
	// proceed (state);
	//
	// within a process expands into:
	//    do { zz_AI_timer.zz_proceed (state); return; } while (0);
	//
	// within an observer expands into:
	//    zz_jump_rq (state);

	char    arg [MAXKWDLEN+1];
	int     lc, pc;

	if (! CodeType) return (NO);

	lc = del;

	if (lc == '(') {
		pc = 1;
		while (1) {
			if ((lc = lkpC ()) == '(') pc++;
			if (!isspace (lc)) break;
		}
	} else {
		pc = 0;
	}

	if (lc == ERROR) return (NO);
	if (lc == END) {
		xerror ("file ends in the middle of 'proceed'");
		return (NO);
	}

	lc = getKeyword (arg, lc);

	while (pc) {
		if (lc != ')') {
			xerror ("'proceed' syntax error");
			return (NO);
		}
		pc--;
		while (1) {
			lc = lkpC ();
			if (!isspace (lc)) break;
		}
	}

	if (lc != ';') {
		xerror ("semicolon missing after 'proceed'");
		return (NO);
	}

	if (!isState (arg)) return (NO);

	if (CodeType == PROCESS) {
		putC ("do { zz_AI_timer.zz_proceed (");
		putC (arg);
		putC ("); return; } while (0);");
	} else {
		// The observer version
		putC ("zz_jump_rq (");
		putC (arg);
		putC (");");
	}

	catchUp ();
	return (YES);
}

int     processSkipto (int del) {

/* -------------------------------- */
/* Expands the `skipto s;' sequence */
/* -------------------------------- */

	// skipto state;
	// skipto (state);
	//
	// within a process expands into:
	//     do { zz_AI_timer.zz_skipto (state); return; } while (0);

	char    arg [MAXKWDLEN+1];
	int     lc, pc;

	if (CodeType != PROCESS) return (NO);

	lc = del;

	if (lc == '(') {
		pc = 1;
		while (1) {
			if ((lc = lkpC ()) == '(') pc++;
			if (!isspace (lc)) break;
		}
	} else {
		pc = 0;
	}

	if (lc == ERROR) return (NO);
	if (lc == END) {
		xerror ("file ends in the middle of 'skipto'");
		return (NO);
	}

	lc = getKeyword (arg, lc);

	while (pc) {
		if (lc != ')') {
			xerror ("'skipto' syntax error");
			return (NO);
		}
		pc--;
		while (1) {
			lc = lkpC ();
			if (!isspace (lc)) break;
		}
	}

	if (lc != ';') {
		xerror ("semicolon missing after 'skipto'");
		return (NO);
	}

	if (!isState (arg)) return (NO);

	putC ("do { zz_AI_timer.zz_skipto (");
	putC (arg);
	putC ("); return; } while (0);");

	catchUp ();
	return (YES);
}

int     processEobject (int del) {

/* ------------------------------------ */
/* Expands the EObject type declaration */
/* ------------------------------------ */

	char    arg [MAXKWDLEN+1], par [MAXKWDLEN+1];
	int     lc, mode, fsup;
	SymDesc *ob, *co;

	// eobject A [virtual];
	// expands into 'class A;'
	//
	// eobject A [virtual] : symlist {
	// eobject A [virtual] {
	//
	// expands into:
	//
	// class A : symlist {          // public added to each symbol
	//              public:
	//              virtual char *getTName () {
	//                  return ("A");
	//              };

	lc = del;
	if (lc == END) {
Endfile:
		xerror ("file ends in the middle of 'eobject' declaration");
		return (NO);
	}
	if (lc == ERROR || !isalpha (lc) && lc != '_' && lc != '$') return (NO);

	// This is the EObject type name
	lc = getKeyword (arg, lc);

	// Determine the delimiter

	if (lc == ';') {
		// Just announcing the type
		if (!addSym (arg, EOBJECT)) return (NO);
Announce:
		putC ("class ");
		putC (arg);
		putC (';');
		catchUp ();
		return (YES);
	}

	if (lc != ':' && lc != '{') {
		// Check if followed by 'virtual'
		lc = getKeyword (par, lc);
		if (lc == ERROR) return (NO);
		if (lc == END) goto Endfile;
		if (strcmp (par, "virtual")) {
Synerror:
			xerror ("'eobject' declaration syntax error");
			return (NO);
		}
		if (lc == ';') {
			if (!addSym (arg, EOBJECT, ANNOUNCED, VIRTUAL))
				return (NO);
			goto Announce;
		}
		if (lc != ':' && lc != '{') goto Synerror;
		mode = VIRTUAL;
	} else {
		mode = REAL;
	}

	if (!(co = addSym (arg, EOBJECT, DEFINED, mode))) return (NO);

	putC ("class "); putC (arg);

	fsup = YES;
	if (lc == ':') {
		// Derived from something: go through the list of identifiers
		// and append them to the declaration
		while (1) {
			lc = getKeyword (par);
			if (lc == ERROR) return (NO);
			if (lc == END) goto Endfile;
			if (strcmp (par, arg) == 0) {
				xerror (form ("'%s' refers to itself", par));
				goto Assumereal;
			}
			if ((ob = getSym (par)) == NULL) {
				xerror (form ("'%s' undefined", par));
				goto Assumereal;
			}
			if (!isDefined (ob, EOBJECT, DEFINED)) {
				xerror (form ("'%s' not a defined EObject",
					par));
				goto Assumereal;
			}
			if (fsup) {
				putC (" : ");
				fsup = NO;
				// The first keyword
				if (ob->Mode == VIRTUAL) {
					if (mode == REAL) {
						// Force the REAL status
						putC ("public EObject, ");
					} 
				} else {
					if (mode == VIRTUAL) {
xerror (form ("'%s' declared virtual but '%s' is not virtual", arg, par));
						goto Assumereal;
					}
				}
			} else {
				if (ob->Mode != VIRTUAL) {
				    xerror (form ("'%s' must be virtual", par));
				}
			}
Assumereal:
			putC ("public ");
			putC (par);
			if (lc == ',') {
				putC (',');
			} else if (lc == '{') {
				break;
			} else goto Synerror;
		}
	} else {
		// Derived from EObject or from nothing
		if (mode != VIRTUAL) putC (" : public EObject");
	}

	putC (" { public: ");
	if (pushDecl (co) == NO) return (NO);
	BraceLevel++;
	if (mode != VIRTUAL) {
		putC ("virtual char *getTName () { return (\"");
		putC (arg);
		putC ("\"); }; ");
	}
	catchUp ();
	return (YES);
}

int     processStation (int del) {

/* ------------------------------------ */
/* Expands the Station type declaration */
/* ------------------------------------ */

	char    arg [MAXKWDLEN+1], par [MAXKWDLEN+1];
	int     lc, mode, fsup;
	SymDesc *ob, *co;

	// station A [virtual];
	// expands into 'class A;'
	//
	// station A [virtual] : symlist {
	// station A [virtual] {
	//
	// expands into:
	//
	// class A : symlist {          // public added in front of each item
	//               public:
	//               virtual char *getTName () {
	//                   return ("A");
	//               };

	if (CodeType) {
		xerror ("station declaration within a code method");
		return (NO);
	}
	lc = del;
	if (lc == END) {
Endfile:
		xerror ("file ends in the middle of station declaration");
		return (NO);
	}
	if (lc == ERROR || !isalpha (lc) && lc != '_' && lc != '$') return (NO);

	// This is the new Station type name
	lc = getKeyword (arg, lc);

	// Determine the delimiter

	if (lc == ';') {
		// Just announcing the type
		if (!addSym (arg, STATION)) return (NO);
Announce:
		putC ("class ");
		putC (arg);
		putC (';');
		catchUp ();
		return (YES);
	}

	if (lc != ':' && lc != '{') {
		// Check if followed by 'virtual'
		lc = getKeyword (par, lc);
		if (lc == ERROR) return (NO);
		if (lc == END) goto Endfile;
		if (strcmp (par, "virtual")) {
Synerror:
			xerror ("station declaration syntax error");
			return (NO);
		}
		if (lc == ';') {
			if (!addSym (arg, STATION, ANNOUNCED, VIRTUAL))
				return (NO);
			goto Announce;
		}
		if (lc != ':' && lc != '{') goto Synerror;
		mode = VIRTUAL;
	} else {
		mode = REAL;
	}

	if (!(co = addSym (arg, STATION, DEFINED, mode))) return (NO);

	putC ("class "); putC (arg);

	fsup = YES;
	if (lc == ':') {
		// Derived from something: go through the list of identifiers
		// and append them to the declaration
		while (1) {
			lc = getKeyword (par);
			if (lc == ERROR) return (NO);
			if (lc == END) goto Endfile;
			if (strcmp (par, arg) == 0) {
				xerror (form ("'%s' refers to itself", par));
				goto Assumereal;
			}
			if ((ob = getSym (par)) == NULL) {
				xerror (form ("'%s' undefined", par));
				goto Assumereal;
			}
			if (!isDefined (ob, STATION, DEFINED)) {
				xerror (form ("'%s' not a defined Station",
					par));
				goto Assumereal;
			}
			if (fsup) {
				putC (" : ");
				fsup = NO;
				// The first keyword
				if (ob->Mode == VIRTUAL) {
					if (mode == REAL) {
						// Force the REAL status
						putC ("public Station, ");
					} 
				} else {
					if (mode == VIRTUAL) {
xerror (form ("'%s' declared virtual but '%s' is not virtual", arg, par));
						goto Assumereal;
					}
				}
			} else {
				if (ob->Mode != VIRTUAL) {
				    xerror (form ("'%s' must be virtual", par));
				}
			}
Assumereal:
			putC ("public ");
			putC (par);
			if (lc == ',') {
				putC (',');
			} else if (lc == '{') {
				break;
			} else goto Synerror;
		}
	} else {
		// Derived from Station or from nothing
		if (mode != VIRTUAL) putC (" : public Station");
	}

	putC (" { public: ");
	if (pushDecl (co) == NO) return (NO);
	BraceLevel++;
	if (mode != VIRTUAL) {
		putC ("virtual char *getTName () { ");
		putC ("return (\"");
		putC (arg);
		putC ("\"); }; ");
	}

	catchUp ();
	return (YES);
}

int     processProcess (int del) {

/* ------------------------------------ */
/* Expands the Process type declaration */
/* ------------------------------------ */

	char    arg[MAXKWDLEN+1], par[MAXKWDLEN+1], sar[MAXKWDLEN+1];
	int     lc, mode, fsup;
	SymDesc *ob, *co;

	// process A [virtual];                 -> class A;
	// process A [virtual] [: symlist] [(ST [, PR])] {
	//
	// full expansion:
	//
	// class A : symlist {          // public inserted before each symbol
	//                              // absent symlist -> 'public Process'
	//   public: ST *S;             // Only if ST specified
	//           PR *F;             // Only if PR specified
	//   static  int zz_A_prcs;
	//
	//   A () {
	//      S=(ST*)TheStation;      // Only if ST specified
	//      F=(PR*)TheProcess;      // Only if PR specified
	//      zz_typeid = (void*) (&zz_A_prcs);
	//   };
	//   virtual char *getTName () {
	//      return ("A");
	//   };

	if (CodeType) {
		xerror ("process declaration within a code method");
		return (NO);
	}
	lc = del;
	if (lc == END) {
Endfile:
		xerror ("file ends in the middle of process declaration");
		return (NO);
	}
	if (lc == ERROR || !isalpha (lc) && lc != '_' && lc != '$') return (NO);

	// This is the new process type name
	lc = getKeyword (arg, lc);

	// Determine the delimiter

	if (lc == ';') {
		// Just announcing the type
		if (!addSym (arg, PROCESS)) return (NO);
Announce:
		putC ("class ");
		putC (arg);
		putC (';');
		catchUp ();
		return (YES);
	}

	if (lc != ':' && lc != '{' && lc != '(') {
		// Check if followed by 'virtual'
		lc = getKeyword (par, lc);
		if (lc == ERROR) return (NO);
		if (lc == END) goto Endfile;
		if (strcmp (par, "virtual")) {
Synerror:
			xerror ("process declaration syntax error");
			return (NO);
		}
		if (lc == ';') {
			if (!addSym (arg, PROCESS, ANNOUNCED, VIRTUAL))
				return (NO);
			goto Announce;
		}
		if (lc != ':' && lc != '{' && lc != '(') goto Synerror;
		mode = VIRTUAL;
	} else {
		mode = REAL;
	}

	if (!(co = addSym (arg, PROCESS, DEFINED, mode))) return (NO);

	putC ("class ");
	putC (arg);

	fsup = YES;
	if (lc == ':') {
		// Derived from something: go through the list of identifiers
		// and append them to the declaration
		while (1) {
			lc = getKeyword (par);
			if (lc == ERROR) return (NO);
			if (lc == END) goto Endfile;
			if (strcmp (par, arg) == 0) {
				xerror (form ("'%s' refers to itself", par));
				goto Assumereal;
			}
			if ((ob = getSym (par)) == NULL) {
				xerror (form ("'%s' undefined", par));
				goto Assumereal;
			}
			if (!isDefined (ob, PROCESS, DEFINED)) {
				xerror (form ("'%s' not a defined Process",
					par));
				goto Assumereal;
			}
			if (fsup) {
				putC (" : ");
				fsup = NO;
				// The first keyword
				if (ob->Mode == VIRTUAL) {
					if (mode == REAL) {
						// Force the REAL status
						putC ("public Process, ");
					} 
				} else {
					if (mode == VIRTUAL) {
xerror (form ("'%s' declared virtual but '%s' is not virtual", arg, par));
						goto Assumereal;
					}
				}
			} else {
				if (ob->Mode != VIRTUAL) {
				    xerror (form ("'%s' must be virtual", par));
				}
			}
Assumereal:
			putC ("public ");
			putC (par);
			if (lc == ',') {
				putC (',');
			} else if (lc == '{' || lc == '(') {
				break;
			} else goto Synerror;
		}
	} else {
		// Derived from Process or from nothing
		if (mode != VIRTUAL) putC (" : public Process");
	}

	putC (" { public: ");

	// Flag parent type and station type as undefined
	par [0] = sar [0] = '\0';

	if (lc == '(') {
		if (mode == VIRTUAL) {
xerror ("virtual process cannot specify father process or owning station");
			return (NO);
		}
		// Owning station type goes first
		lc = getKeyword (sar);
		if (lc == ERROR) return (NO);
		if (lc == END) goto Endfile;
		if ((ob = getSym (sar)) == NULL) {
			xerror (form ("'%s' undefined", sar));
		} else if (!isDefined (ob, STATION, ANNOUNCED)) {
			xerror (form ("'%s' not defined as 'Station'", sar));
		} else if (ob->Mode == VIRTUAL) {
			xerror (form ("'%s' is virtual", sar));
		}
		// Continue with this error
		if (lc == ',') {
			// Parent process goes next
			lc = getKeyword (par);
			if (lc == ERROR) return (NO);
			if (lc == END) goto Endfile;
			if ((ob = getSym (par)) == NULL) {
				xerror (form ("'%s' undefined", par));
			} else if (!isDefined (ob, PROCESS, ANNOUNCED)) {
			   xerror (form ("'%s' not defined as 'Process'", par));
			} else if (ob->Mode == VIRTUAL) {
			   xerror (form ("'%s' is virtual", par));
			}
		}
		if (lc != ')') goto Synerror;
		while (1) {
			lc = lkpC ();
			if (!isspace (lc)) break;
		}
		if (lc != '{') goto Synerror;
	}

	if (pushDecl (co) == NO) return (NO);

	BraceLevel++;

	if (sar [0] != '\0') {
		// Owning station defined (impossible for virtual process)
		putC (sar);
		putC (" *S; ");
	}
	if (par [0] != '\0') {
		// Parent process defined (see above)
		putC (par);
		putC (" *F; ");
	}
	if (mode == REAL) {
		putC ("static int zz_");
		putC (arg);
		putC ("_prcs; ");
		putC (arg);
		putC (" () { ");
		if (sar [0] != '\0') {
			// Owning station defined
			putC ("S = (");
			putC (sar);
			putC ("*)TheStation; ");
		}
		if (par [0] != '\0') {
			// Parent process defined
			putC ("F = (");
			putC (par);
			putC ("*)TheProcess; ");
		}
		putC ("zz_typeid = (void*) (&zz_");
		putC (arg);
		putC ("_prcs); }; virtual char *getTName () { return (\"");
		putC (arg);
		putC ("\"); };");
	}

	catchUp ();
	return (YES);
}

int     processObserver (int del) {

/* ------------------------------------- */
/* Expands the Observer type declaration */
/* ------------------------------------- */

	char    arg [MAXKWDLEN+1];
	int     lc;
	SymDesc *co;

	// observer A;                  -> class A;
	// observer A {
	//
	// expands into:
	//
	// class A : public Observer {
	//     public:
	//     virtual char *getTName () {
	//       return ("A");
	//     };

	if (CodeType) {
		xerror ("observer declaration within a code method");
		return (NO);
	}
	lc = del;
	if (lc == END) {
		xerror ("file ends in the middle of observer declaration");
		return (NO);
	}
	if (lc == ERROR || !isalpha (lc) && lc != '_' && lc != '$') return (NO);

	// This is the new observer type name
	lc = getKeyword (arg, lc);

	// Determine the delimiter

	if (lc == ';') {
		// Just announcing the type
		if (!addSym (arg, OBSERVER, ANNOUNCED, REAL)) return (NO);
		putC ("class ");
		putC (arg);
		putC (';');
		catchUp ();
		return (YES);
	}

	if (lc != '{') {
		// No other choice
		xerror ("observer declaration syntax error");
		return (NO);
	}

	if (!(co = addSym (arg, OBSERVER, DEFINED, REAL))) return (NO);

	putC ("class ");
	putC (arg);
	putC (" : public Observer { public: virtual char *getTName () { ");
	putC ("return (\"");
	putC (arg);
	putC ("\"); };");

	if (pushDecl (co) == NO) return (NO);
	BraceLevel++;

	catchUp ();
	return (YES);
}

int     doPerform (int del, SymDesc *obj) {

/* --------------------------------- */
/* Expands the 'perform' declaration */
/* --------------------------------- */

	int     lc;

	// perform;                             // Announcement
	// expands into 'zz_code ();'
	//
	// perform {
	//
	// in a process expands into:
	//      zz_code () {switch (TheState) {
	// in an observer expands into:
	//      zz_code () {switch (TheObserverState) {

	lc = del;
	if (lc == ERROR) return (NO);
	if (lc == END) {
		xerror ("file ends in the middle of 'perform' declaration");
		CodeType = NO;
		CodeObject = NULL;
		return (NO);
	}

	if (lc == ';') {
		// Just announcing
		if (obj != CObject) {
			xerror ("'perform' announced incorrectly");
			return (NO);
		}
		if (obj->Mode != REAL) {
Modeerr:
			xerror ("'perform' illegal in a virtual process");
			return (NO);
		}
		CodeType = NO;
		CodeObject = NULL;
		putC ("zz_code ();");
		obj->PAnn = YES;
		catchUp ();
		return (YES);
	}

	if (lc != '{') {
		CodeType = NO;
		CodeObject = NULL;
		xerror ("'perform' declaration syntax error");
		return (NO);
	}

	BraceLevel++;
	putC ("zz_code () {");
	CodeWasState = NO;

	if (obj->Mode != REAL) goto Modeerr;

	obj->PAnn = YES;

	catchUp ();
	return (YES);
}

int     processTraffic (int del) {

/* ------------------------------------ */
/* Expands the Traffic type declaration */
/* ------------------------------------ */

	char    arg[MAXKWDLEN+1], par[MAXKWDLEN+1], mar[MAXKWDLEN+1];
	int     lc, mode, fsup;
	SymDesc *ob, *co;

	// traffic A [virtual];                 -> class A;
	// traffic A [virtual] [:symlist] [(M [, P])] {
	//
	// full expansion:
	//
	// class A : public symlist {           // public added, abs = Traffic
	//      private:
	//      virtual Message *zz_mkm () { return (Info01 = (void*) new M); };
	//      virtual Packet  *zz_mkp () { Info01 = (void*) new P;
	//              zz_mkpclean ();
	//              return ((Packet*)Info01);
	//      };      // defaults: P = Packet, M = Message
	//      public:
	//      virtual char *getTName () { return ("A"); };

	lc = del;
	if (lc == END) {
Endfile:
		xerror ("file ends in the middle of traffic declaration");
		return (NO);
	}
	if (lc == ERROR || !isalpha (lc) && lc != '_' && lc != '$') return (NO);

	if (CodeType) {
		xerror ("traffic declaration within a code method");
		return (NO);
	}

	// This is the new traffic type name
	lc = getKeyword (arg, lc);

	// Determine the delimiter

	if (lc == ';') {
		// Just announcing the type
		if (!addSym (arg, TRAFFIC)) return (NO);
Announce:
		putC ("class ");
		putC (arg);
		putC (';');
		catchUp ();
		return (YES);
	}

	if (lc != ':' && lc != '{' && lc != '(') {
		// Check if followed by 'virtual'
		lc = getKeyword (par, lc);
		if (lc == ERROR) return (NO);
		if (lc == END) goto Endfile;
		if (strcmp (par, "virtual")) {
Synerror:
			xerror ("traffic declaration syntax error");
			return (NO);
		}
		if (lc == ';') {
			if (!addSym (arg, TRAFFIC, ANNOUNCED, VIRTUAL))
				return (NO);
			goto Announce;
		}
		if (lc != ':' && lc != '{' && lc != '(') goto Synerror;
		mode = VIRTUAL;
	} else {
		mode = REAL;
	}

	if (!(co = addSym (arg, TRAFFIC, DEFINED, mode))) return (NO);

	putC ("class "); putC (arg);

	fsup = YES;
	if (lc == ':') {
		// Derived from something: go through the list of identifiers
		// and append them to the declaration
		while (1) {
			lc = getKeyword (par);
			if (lc == ERROR) return (NO);
			if (lc == END) goto Endfile;
			if (strcmp (par, arg) == 0) {
				xerror (form ("'%s' refers to itself", par));
				goto Assumereal;
			}
			if ((ob = getSym (par)) == NULL) {
				xerror (form ("'%s' undefined", par));
				goto Assumereal;
			}
			if (!isDefined (ob, TRAFFIC, DEFINED)) {
				xerror (form ("'%s' not a defined Traffic",
					par));
				goto Assumereal;
			}
			if (fsup) {
				putC (" : ");
				fsup = NO;
				// The first keyword
				if (ob->Mode == VIRTUAL) {
					if (mode == REAL) {
						// Force the REAL status
						putC ("public Traffic, ");
					} 
				} else {
					if (mode == VIRTUAL) {
xerror (form ("'%s' declared virtual but '%s' is not virtual", arg, par));
						goto Assumereal;
					}
				}
			} else {
				if (ob->Mode != VIRTUAL) {
				    xerror (form ("'%s' must be virtual", par));
				}
			}
Assumereal:
			putC ("public ");
			putC (par);
			if (lc == ',') {
				putC (',');
			} else if (lc == '{' || lc == '(') {
				break;
			} else goto Synerror;
		}
	} else {
		// Derived from Traffic or from nothing
		if (mode != VIRTUAL) putC (" : public Traffic");
	}

	// Flag message type and packet type as undefined
	mar [0] = par [0] = '\0';

	if (lc == '(') {
		if (mode != REAL) {
			xerror (
   "virtual traffic type cannot specify associated message and/or packet type");
			return (NO);
		}
		// Message type goes first
		lc = getKeyword (mar);
		if (lc == ERROR) return (NO);
		if (lc == END) goto Endfile;
		if ((ob = getSym (mar)) == NULL) {
			xerror (form ("'%s' undefined", mar));
		} else if (!isDefined (ob, MESSAGE, ANNOUNCED)) {
			xerror (form ("'%s' not defined as 'Station'", mar));
		} else if (ob->Mode == VIRTUAL) {
			xerror (form ("'%s' is virtual", mar));
		}
		// Continue with this error
		if (lc == ',') {
			// Packet type goes next
			lc = getKeyword (par);
			if (lc == ERROR) return (NO);
			if (lc == END) goto Endfile;
			if ((ob = getSym (par)) == NULL) {
				xerror (form ("'%s' undefined", par));
			} else if (!isDefined (ob, PACKET, ANNOUNCED)) {
			   xerror (form ("'%s' not defined as 'Packet'", par));
			} else if (ob->Mode == VIRTUAL) {
			   xerror (form ("'%s' is virtual", par));
			}
			// Continue with this error
		}
		if (lc != ')') goto Synerror;
		while (1) {
			lc = lkpC ();
			if (!isspace (lc)) break;
		}
		if (lc != '{') goto Synerror;
	}

	if (pushDecl (co) == NO) return (NO);
	BraceLevel++;
	if (mode != VIRTUAL) {
		putC (" { private: virtual Message *zz_mkm () { ");
		putC (" return ((Message*)(Info01 = (void*) new ");
		putC (*mar != '\0' ? mar : "Message");
		putC (")); }; virtual Packet *zz_mkp () {Info01 = (void*)new ");
		putC (*par != '\0' ? par : "Packet");
		putC ("; zz_mkpclean ();return ((Packet*)Info01); }; public: ");
		putC ("virtual char *getTName () { return (\"");
		putC (arg);
		putC ("\"); };");
	} else {
		putC (" { public: ");
	}

	catchUp ();
	return (YES);
}

int     processMessage (int del) {

/* ------------------------------------ */
/* Expands the Message type declaration */
/* ------------------------------------ */

	char    arg [MAXKWDLEN+1], par [MAXKWDLEN+1];
	int     lc, mode, fsup;
	SymDesc *ob, *co;

	// message A [virtual];                 -> class A;
	// message A [virtual] [: symlist] {
	//
	// expands as:
	//
	// class A : symlist {                  // public inserted
	//       public:
	//       virtual int frameSize () { return (sizeof (A)); };

	lc = del;
	if (lc == END) {
Endfile:
		xerror ("file ends in the middle of 'message' declaration");
		return (NO);
	}
	if (lc == ERROR || !isalpha (lc) && lc != '_' && lc != '$') return (NO);

	if (CodeType) {
		xerror ("'message' declaration within a code method");
		return (NO);
	}

	// This is the new Message type name
	lc = getKeyword (arg, lc);

	// Determine the delimiter

	if (lc == ';') {
		// Just announcing the type
		if (!addSym (arg, MESSAGE, ANNOUNCED, FREE, UNEXPOSABLE))
			return (NO);
Announce:
		putC ("class ");
		putC (arg);
		putC (';');
		catchUp ();
		return (YES);
	}

	if (lc != ':' && lc != '{') {
		// Check if followed by 'virtual'
		lc = getKeyword (par, lc);
		if (lc == ERROR) return (NO);
		if (lc == END) goto Endfile;
		if (strcmp (par, "virtual")) {
Synerror:
			xerror ("'message' declaration syntax error");
			return (NO);
		}
		if (lc == ';') {
			if (!addSym (arg, MESSAGE, ANNOUNCED, VIRTUAL,
			    UNEXPOSABLE))
				return (NO);
			goto Announce;
		}
		if (lc != ':' && lc != '{') goto Synerror;
		mode = VIRTUAL;
	} else {
		mode = REAL;
	}

	if (!(co = addSym (arg, MESSAGE, DEFINED, mode, UNEXPOSABLE)))
		return (NO);

	putC ("class ");
	putC (arg);

	fsup = YES;
	if (lc == ':') {
		// Derived from something: go through the list of identifiers
		// and append them to the declaration
		while (1) {
			lc = getKeyword (par);
			if (lc == ERROR) return (NO);
			if (lc == END) goto Endfile;
			if (strcmp (par, arg) == 0) {
				xerror (form ("'%s' refers to itself", par));
				goto Assumereal;
			}
			if ((ob = getSym (par)) == NULL) {
				xerror (form ("'%s' undefined", par));
				goto Assumereal;
			}
			if (!isDefined (ob, MESSAGE, DEFINED)) {
				xerror (form ("'%s' not a defined Message",
					par));
				goto Assumereal;
			}
			if (fsup) {
				putC (" : ");
				fsup = NO;
				// The first keyword
				if (ob->Mode == VIRTUAL) {
					if (mode == REAL) {
						// Force the REAL status
						putC ("public Message, ");
					} 
				} else {
					if (mode == VIRTUAL) {
xerror (form ("'%s' declared virtual but '%s' is not virtual", arg, par));
						goto Assumereal;
					}
				}
			} else {
				if (ob->Mode != VIRTUAL) {
				    xerror (form ("'%s' must be virtual", par));
				}
			}
Assumereal:
			putC ("public ");
			putC (par);
			if (lc == ',') {
				putC (',');
			} else if (lc == '{') {
				break;
			} else goto Synerror;
		}
	} else {
		// Derived from Message or from nothing
		if (mode != VIRTUAL) putC (" : public Message");
	}

	if (pushDecl (co) == NO) return (NO);
	BraceLevel++;
	putC (" { public: ");
	if (mode != VIRTUAL) {
		putC ("virtual int frameSize () { return (sizeof (");
		putC (arg);
		putC (")); };");
	}

	catchUp ();
	return (YES);
}

int     processPacket (int del) {

/* ----------------------------------- */
/* Expands the Packet type declaration */
/* ----------------------------------- */

	char    arg [MAXKWDLEN+1], par [MAXKWDLEN+1];
	int     lc, mode, fsup;
	SymDesc *ob, *co;

	// packet A;                            -> class A;
	// packet A : symlist {
	//
	// expands as:
	//
	// class A : symlist {                  // public inserted
	//       public:
	//       virtual int frameSize () { return (sizeof (A)); };

	lc = del;
	if (lc == END) {
Endfile:
		xerror ("file ends in the middle of 'packet' declaration");
		return (NO);
	}
	if (lc == ERROR || !isalpha (lc) && lc != '_' && lc != '$') return (NO);

	if (CodeType) {
		xerror ("'packet' declaration within a code method");
		return (NO);
	}

	// This is the new Packet type name
	lc = getKeyword (arg, lc);

	// Determine the delimiter

	if (lc == ';') {
		// Just announcing the type
		if (!addSym (arg, PACKET, ANNOUNCED, FREE, UNEXPOSABLE))
			return (NO);
Announce:
		putC ("class ");
		putC (arg);
		putC (';');
		catchUp ();
		return (YES);
	}

	if (lc != ':' && lc != '{') {
		// Check if followed by 'virtual'
		lc = getKeyword (par, lc);
		if (lc == ERROR) return (NO);
		if (lc == END) goto Endfile;
		if (strcmp (par, "virtual")) {
Synerror:
			xerror ("'packet' declaration syntax error");
			return (NO);
		}
		if (lc == ';') {
			if (!addSym (arg, PACKET, ANNOUNCED, VIRTUAL,
			    UNEXPOSABLE))
				return (NO);
			goto Announce;
		}
		if (lc != ':' && lc != '{') goto Synerror;
		mode = VIRTUAL;
	} else {
		mode = REAL;
	}

	if (!(co = addSym (arg, PACKET, DEFINED, mode, UNEXPOSABLE)))
		return (NO);

	putC ("class "); putC (arg);

	fsup = YES;
	if (lc == ':') {
		// Derived from something: go through the list of identifiers
		// and append them to the declaration
		while (1) {
			lc = getKeyword (par);
			if (lc == ERROR) return (NO);
			if (lc == END) goto Endfile;
			if (strcmp (par, arg) == 0) {
				xerror (form ("'%s' refers to itself", par));
				goto Assumereal;
			}
			if ((ob = getSym (par)) == NULL) {
				xerror (form ("'%s' undefined", par));
				goto Assumereal;
			}
			if (!isDefined (ob, PACKET, DEFINED)) {
				xerror (form ("'%s' not a defined Packet",
					par));
				goto Assumereal;
			}
			if (fsup) {
				putC (" : ");
				fsup = NO;
				// The first keyword
				if (ob->Mode == VIRTUAL) {
					if (mode == REAL) {
						// Force the REAL status
						putC ("public Packet, ");
					} 
				} else {
					if (mode == VIRTUAL) {
xerror (form ("'%s' declared virtual but '%s' is not virtual", arg, par));
						goto Assumereal;
					}
				}
			} else {
				if (ob->Mode != VIRTUAL) {
				    xerror (form ("'%s' must be virtual", par));
				}
			}
Assumereal:
			putC ("public ");
			putC (par);
			if (lc == ',') {
				putC (',');
			} else if (lc == '{') {
				break;
			} else goto Synerror;
		}
	} else {
		// Derived from Packet or from nothing
		if (mode != VIRTUAL) putC (" : public Packet");
	}

	if (pushDecl (co) == NO) return (NO);
	BraceLevel++;
	putC (" { public: ");
	if (mode != VIRTUAL) {
		putC ("virtual int frameSize () { return (sizeof (");
		putC (arg);
		putC (")); };");
	}

	catchUp ();
	return (YES);
}

int     isSymbol (char *arg) {

/* -------------------------------------- */
/* Checks is the string is a legal symbol */
/* -------------------------------------- */

	if (!isalpha (*arg) && *arg != '_' && *arg != '$') return (NO);
	arg++;
	while (*arg != '\0') {
		if (!isalnum (*arg) && *arg != '_' && *arg != '$') return (NO);
		arg++;
	}
	return (YES);
}

int     processLink (int del) {

/* --------------------------------- */
/* Expands the Link type declaration */
/* --------------------------------- */

	char    arg [MAXKWDLEN+1], par [MAXKWDLEN+1];
	int     lc, mode, fsup;
	SymDesc *ob, *co;

	// link A [virtual];                    -> class A;
	// link A [virtual] [:symlist] {
	//
	// full expansion:
	//
	// class A : symlist {                  // public added
	//       public:
	//       virtual char *getTName () {
	//           return ("A");
	//       };

	lc = del;
	if (lc == END) {
Endfile:
		xerror ("file ends in the middle of link declaration");
		return (NO);
	}
	if (lc == ERROR || !isalpha (lc) && lc != '_' && lc != '$') return (NO);

	if (CodeType) {
		xerror ("link declaration within a code method");
		return (NO);
	}

	// This is the new Link type name
	lc = getKeyword (arg, lc);

	// Determine the delimiter

	if (lc == ';') {
		// Just announcing the type
		if (!addSym (arg, LINK)) return (NO);
Announce:
		putC ("class ");
		putC (arg);
		putC (';');
		catchUp ();
		return (YES);
	}

	if (lc != ':' && lc != '{') {
		// Check if followed by 'virtual'
		lc = getKeyword (par, lc);
		if (lc == ERROR) return (NO);
		if (lc == END) goto Endfile;
		if (strcmp (par, "virtual")) {
Synerror:
			xerror ("link declaration syntax error");
			return (NO);
		}
		if (lc == ';') {
			if (!addSym (arg, LINK, ANNOUNCED, VIRTUAL))
				return (NO);
			goto Announce;
		}
		if (lc != ':' && lc != '{') goto Synerror;
		mode = VIRTUAL;
	} else {
		mode = REAL;
	}

	if (!(co = addSym (arg, LINK, DEFINED, mode))) return (NO);

	putC ("class "); putC (arg);

	fsup = YES;
	if (lc == ':') {
		// Derived from something: go through the list of identifiers
		// and append them to the declaration
		while (1) {
			lc = getKeyword (par);
			if (lc == ERROR) return (NO);
			if (lc == END) goto Endfile;
			if (strcmp (par, arg) == 0) {
				xerror (form ("'%s' refers to itself", par));
				goto Assumereal;
			}
			if ((ob = getSym (par)) == NULL) {
				xerror (form ("'%s' undefined", par));
				goto Assumereal;
			}
			if (!isDefined (ob, LINK, DEFINED)) {
				xerror (form ("'%s' not a defined Link", par));
				goto Assumereal;
			}
			if (fsup) {
				putC (" : ");
				fsup = NO;
				// The first keyword
				if (ob->Mode == VIRTUAL) {
					if (mode == REAL) {
						// Force the REAL status
						putC ("public Link, ");
					} 
				} else {
					if (mode == VIRTUAL) {
xerror (form ("'%s' declared virtual but '%s' is not virtual", arg, par));
						goto Assumereal;
					}
				}
			} else {
				if (ob->Mode != VIRTUAL) {
				    xerror (form ("'%s' must be virtual", par));
				}
			}
Assumereal:
			putC ("public ");
			putC (par);
			if (lc == ',') {
				putC (',');
			} else if (lc == '{') {
				break;
			} else goto Synerror;
		}
	} else {
		// Derived from Link or from nothing
		if (mode != VIRTUAL) putC (" : public Link");
	}

	if (pushDecl (co) == NO) return (NO);
	BraceLevel++;
	putC (" { public: ");
	if (mode != VIRTUAL) {
		putC ("virtual char *getTName () { ");
		putC ("return (\"");
		putC (arg);
		putC ("\"); }; ");
	}

	catchUp ();
	return (YES);
}

SymDesc *isPrcs (char *s) {

/* ---------------------------------------------------------------- */
/* Checks whether the symbol is a process type id (used in inspect) */
/* ---------------------------------------------------------------- */

	SymDesc  *ob;

	if (!isSymbol (s)) {
		xerror (form ("process type id '%s' is not a symbol", s));
		return (NULL);
	}

	if ((ob = getSym (s)) == NULL) {
		xerror (form ("'%s' undefined", s));
		return (NULL);
	}

	if (ob->Type != PROCESS) {
		xerror (form ("'%s' is not a process type", s));
		return (NULL);
	}
	return (ob);
}

int     processInspect (int del) {

/* ----------------------- */
/* Expands the inpect call */
/* ----------------------- */

	// inspect (A)                  expands into:
	//
	// zz_inspect_rq ((Station*)ANY, (void*)(&(Process::zz_Process_prcs)),
	//   (char*)ANY, ANY, A)
	//
	// inspect (A, B)               expands into:
	//
	// zz_inspect_rq ((Station*)(A), (void*)(&(Process::zz_Process_prcs)),
	//   (char*)ANY, ANY, B)
	//
	// inspect (A, B, C)            expands into:
	//
	// zz_inspect_rq ((Station*)(A), (void*)(&(B::zz_B_prcs)), (char*)ANY,
	//   ANY, C)
	//
	// inspect (A, B, C, D)         expands into:
	//
	// zz_inspect_rq ((Station*)(A), (void*)(&(B::zz_B_prcs)), (char*)ANY,
	//   C, D)
	//
	// inspect (A, B, C, D, E)      expands into:
	//
	// zz_inspect_rq ((Station*)(A), (void*)(&(B::zz_B_prcs)), (char*)(C),
	//   D, E)

	char    arg [5] [MAXKWDLEN+1];  // Up to 5 arguments
	int     lc, ac, ap;
	SymDesc *to;

	if (CodeType != OBSERVER) return (NO);
	lc = del;
	if (lc == ERROR) return (NO);
	if (lc == END) {
Endfile:
		xerror ("file ends in the middle of inspect");
		return (NO);
	}

	if (lc != '(') {
Synerror:
		xerror ("inspect syntax error");
		return (NO);
	}

	for (ac = 0; ac < 5;) {
		// Parse the arguments
		lc = getArg (arg [ac]);
		ac++;
		if (lc == ERROR) return (NO);
		if (lc == END) goto Endfile;
		if (lc == ')') break;
		if (lc != ',') goto Synerror;
	}
	if (lc != ')') {
		xerror ("too many arguments to inspect");
		return (NO);
	}

	putC ("zz_inspect_rq ((Station*)(");

	// The rest off the processing differs depending on the argument count

	switch (ac) {

	    case 1:     // Monitor state only

		if (!isSymbol (arg[0])) {
Msterror:
			xerror ("monitor state id is not a symbol");
			return (NO);
		}
		if (!isState (arg[0])) return (NO);
		putC ("-1), (void*)(&(Process::zz_Process_prcs)), ");
		putC ("(char*) (-1), -1, ");
		putC (arg[0]);
		putC (')');
		break;

	    case 2:     // Station + monitor state

		putC (arg[0]);
		putC ("), (void*)(&(Process::zz_Process_prcs)), ");
		putC ("(char*) (-1), -1, ");
		if (!isSymbol (arg[1])) goto Msterror;
		if (!isState (arg[1])) return (NO);
		putC (arg[1]);
		putC (')');
		break;

	    case 3:     // Station + process type + monitor state

		putC (arg[0]);
		putC ("), (void*)(&(");
		if (strcmp (arg[1], "(-1)") == 0) {
			putC ("Process::zz_Process_prcs");
		} else if (!isPrcs (arg[1])) {
			return (NO);
		} else {
			putC (arg[1]);
			putC ("::zz_");
			putC (arg[1]);
			putC ("_prcs");
		}
		putC (")), (char*) (-1), -1, ");
		if (!isSymbol (arg[2])) goto Msterror;
		if (!isState (arg[2])) return (NO);
		putC (arg[2]);
		putC (')');
		break;

	    case 4:     // Station + process type + process state + mtr state

		putC (arg[0]);
		putC ("), (void*)(&(");
		to = CodeObject;
		ap = (strcmp (arg[1], "(-1)") == 0);
		if (!ap && !(CodeObject = isPrcs (arg[1]))) {
			CodeObject = to;
			return (NO);
		}
		if (ap) {
			putC ("Process::zz_Process_prcs");
		} else {
			putC (arg[1]);
			putC ("::zz_");
			putC (arg[1]);
			putC ("_prcs");
		}
		putC (")), (char*) (-1), ");
		if (strcmp (arg[2], "(-1)")) {
			if (!isSymbol (arg[2])) {
Psterror:
				xerror ("process state is not a symbol");
				CodeObject = to;
				return (NO);
			}
			if (ap) {
Pstanerr:
		xerror ("process type must be specified if state is not 'ANY'");
				CodeObject = to;
				return (NO);
			}
			if (!isState (arg[2])) {
				CodeObject = to;
				return (NO);
			}
			putC (CodeObject->Name);
			putC ("::");
		}
		CodeObject = to;
		putC (arg[2]);
		putC (", ");
		if (!isSymbol (arg[3])) goto Msterror;
		if (!isState (arg[3])) return (NO);
		putC (arg[3]);
		putC (')');
		break;

	    case 5:     // St + prcs type + prcs nknm + prcs state + mtr state

		putC (arg[0]);
		putC ("), (void*)(&(");
		to = CodeObject;
		ap = (strcmp (arg[1], "(-1)") == 0);
		if (!ap && !(CodeObject = isPrcs (arg[1]))) {
			CodeObject = to;
			return (NO);
		}
		if (ap) {
			putC ("Process::zz_Process_prcs");
		} else {
			putC (arg[1]);
			putC ("::zz_");
			putC (arg[1]);
			putC ("_prcs");
		}
		putC (")), ");
		putC (arg[2]);
		putC (", ");
		if (strcmp (arg[3], "(-1)")) {
			if (!isSymbol (arg[3])) goto Psterror;
			if (ap) goto Pstanerr;
			if (!isState (arg[3])) {
				CodeObject = to;
				return (NO);
			}
			putC (CodeObject->Name);
			putC ("::");
		}
		CodeObject = to;
		putC (arg[3]);
		putC (", ");
		if (!isSymbol (arg[4])) goto Msterror;
		if (!isState (arg[4])) return (NO);
		putC (arg[4]);
		putC (')');
		break;
	}

	catchUp ();
	return (YES);
}

int     processTimeout (int del) {

/* ------------------------ */
/* Expands the timeout call */
/* ------------------------ */

	// timeout (a, b);
	//
	// expands into 'zz_timeout_rq (a, b);

	char    arg [MAXKWDLEN+1];
	int     lc;

	if (CodeType != OBSERVER) return (NO);
	lc = del;
	if (lc == ERROR) return (NO);
	if (lc == END) {
Endfile:
		xerror ("file ends in the middle of 'timeout'");
		return (NO);
	}

	if (lc != '(') {
Synerror:
		xerror ("'timeout' syntax error");
		return (NO);
	}

	lc = getArg (arg);
	if (lc == ERROR) return (NO);
	if (lc == END) goto Endfile;
	if (lc != ',') goto Synerror;           // Two arguments expected
	putC ("zz_timeout_rq (");
	putC (arg);
	lc = getKeyword (arg);
	if (lc == ERROR) return (NO);
	if (lc == END) goto Endfile;
	if (lc != ')') goto Synerror;           // And that's it
	if (!isState (arg)) return (NO);
	putC (", ");
	putC (arg);
	putC (')');

	catchUp ();
	return (YES);
}

int     processCreate (int del) {

/* ----------------------- */
/* Expands the create call */
/* ----------------------- */

	// p = create A;                        -> p = create A ();
	// p = create A (plist);
	// p = create A, B;                     -> p = create A, B, ();
	// p = create A, B, (plist);
	//
	// full expansion (no B):
	//
	// ({A *zz_op = new A; zz_op->zz_start (); zz_op->setup (plist);
	//  zz_op);});
	//
	// full expansion (with B -- the nickname):
	//
	// ({ char *zz_cc = (B); A *zz_op = new A; zz_op->zz_start ();
	//  zz_op->zz_nickname = new char [strlen(zz_cc)+1];
	//   strcpy (zz_op->zz_nickname, zz_cc); zz_op->setup (plist); zz_op;});
	//
	// Note: for observers zz_start and ->setup exchange places
	//
	// p = create (expr) A ...
	//
	// sets TheStation to 'expr' or 'idToStation (expr)' before creation
	//

	// In the AT&T version, the following sequence is generated:
	//
	// (zz_setths (expr), zz_bld_A (B), ((A*)(zz_COBJ[zz_clv]))->setup (..),
	//  ((A*)(zz_COBJ[zz_clv--]))
	//
	// for an observer, it is:
	//
	// (zz_setths (expr), zz_bld_A (B), ((A*)(zz_COBJ[zz_clv]))->setup (..),
	//  ((A*)(zz_COBJ[zz_clv]))->zz_start (), ((A*)(zz_COBJ[zz_clv--])))
	// 

	char    tpn [MAXKWDLEN+1], ths [MAXKWDLEN+1], nkn [MAXKWDLEN+1], *sptr;
	int     lc, sls, delim;
	SymDesc *ob;

	lc = del;
	delim = -1;
	if (lc == END) {
Endfile:
		xerror ("file ends in the middle of 'create'");
		return (NO);
	}

	// Get the type name
	if (lc == ERROR) return (NO);
	if (lc == '(') {
		// Request to set 'TheStation'
		lc = getArg (ths);
		if (lc == ERROR) return (NO);
		if (lc == END) goto Endfile;
		if (lc != ')') goto Synerr;
		while (1) {
			lc = lkpC ();
			if (!isspace (lc)) break;
		}
		if (!isalpha (lc) && lc != '_' && lc != '$') goto Synerr;
	} else {
		ths [0] = '\0';
		if (!isalpha (lc) && lc != '_' && lc != '$') return (NO);
	}

	lc = getKeyword (tpn, lc);
	if (lc == ERROR) return (NO);
	if (lc == END) goto Endfile;
	// Check if the type is defined properly
	if ((ob = getSym (tpn)) == NULL) {
		xerror (form ("'%s' undefined", tpn));
		return (NO);
	}
	if (ob->Status != DEFINED) {
		xerror (form ("'%s' announced but undefined", tpn));
		return (NO);
	}
	if (ob->Mode != REAL) {
		xerror (form ("'%s' is virtual", tpn));
		return (NO);
	}
#if     ZZ_ATT
	putC ('(');
	if (ths [0] != '\0') {
		putC ("zz_setths (");
		putC (ths);
		putC ("), ");
	}
	putC ("zz_bld_");
	putC (tpn);
	putC (" (");
	if (lc == ',') {
		// Nickname specified
		if (ob->Type == MESSAGE || ob->Type == PACKET || ob->Type ==
		  SGROUP || ob->Type == CGROUP) {
		    xerror (form (
		       "'%s': objects of this type don't have nicknames", tpn));
		    return (NO);
		}
		lc = getArg (nkn);
		if (lc == ERROR) return (NO);
		if (lc == END) goto Endfile;
		// Output the nickname
		putC (nkn);
		if (lc == ',') {
			// Skip to '(' or ';', whichever comes first
			while (1) {
				lc = lkpC ();
				if (!isspace (lc)) break;
			}
		}
	}
	putC ("), ((");
	putC (tpn);
	putC ("*)(zz_COBJ[zz_clv]))->setup (");
	if (lc == '(') {
		// Setup argument list
		sptr = EnLkpPtr;
		sls = SkLineCnt;
		while (1) {
			// Check if followed immediately by ')'
			lc = lkpC ();
			if (!isspace (lc)) break;
		}
		if (lc == ')') goto Setupdone;
		EnLkpPtr = sptr;        // Backspace
		SkLineCnt = sls;
		while (1) {
			// Process setup arguments
			lc = getArg (nkn);
			if (lc == ERROR) return (NO);
			if (lc == END) goto Endfile;
			putC (nkn);
			if (lc == ')') goto Setupdone;
			if (lc != ',') {
Synerr:
				xerror ("create syntax error");
				return (NO);
			}
			putC (',');
		}
	} else delim = lc;
Setupdone:
	putC ("), ");
	if (ob->Type == OBSERVER) {
		putC ("((");
		putC (tpn);
		putC ("*)(zz_COBJ[zz_clv]))->zz_start (), ");
	}
	putC ('(');
	putC (tpn);
	putC ("*)(zz_COBJ[zz_clv--]))");
	if (delim != -1) putC (delim);
#else
	putC ("({ ");
	putC (tpn);
	putC (" *zz_op; ");
	if (lc == ',') {
		// Nickname specified
		if (ob->Type == MESSAGE || ob->Type == PACKET || ob->Type ==
		  SGROUP || ob->Type == CGROUP) {
		    xerror (form (
		       "'%s': objects of this type don't have nicknames", tpn));
		    return (NO);
		}
		lc = getArg (nkn);
		if (lc == ERROR) return (NO);
		if (lc == END) goto Endfile;
		// Output the sequence with nickname
		putC ("char *zz_cc = (");
		putC (nkn);
		putC ("); ");
		if (ths [0] != '\0') {
			putC ("zz_setths (");
			putC (ths);
			putC ("); ");
		}
		putC ("zz_op = new ");
		putC (tpn);
		putC ("; ");
		if (ob->Type != OBSERVER) putC ("zz_op->zz_start (); ");
		putC ("zz_op->zz_nickname = new char [strlen(zz_cc)+1]; ");
		putC ("strcpy (zz_op->zz_nickname, zz_cc); ");
		putC ("zz_op->setup (");
		if (lc == ',') {
			// Skip to '(' or ';', whichever comes first
			while (1) {
				lc = lkpC ();
				if (!isspace (lc)) break;
			}
		}
	} else {
		// Output the sequence without nickname
		if (ths [0] != '\0') {
			putC ("zz_setths (");
			putC (ths);
			putC ("); ");
		}
		putC ("zz_op = new ");
		putC (tpn);
		putC ("; ");
		if (ob->Type != OBSERVER) putC ("zz_op->zz_start (); ");
		putC ("zz_op->setup (");
	}

	if (lc == '(') {
		// Setup argument list
		sptr = EnLkpPtr;
		sls = SkLineCnt;
		while (1) {
			// Check if followed immediately by ')'
			lc = lkpC ();
			if (!isspace (lc)) break;
		}
		if (lc == ')') goto Setupdone;
		EnLkpPtr = sptr;        // Backspace
		SkLineCnt = sls;
		while (1) {
			// Process setup arguments
			lc = getArg (nkn);
			if (lc == ERROR) return (NO);
			if (lc == END) goto Endfile;
			putC (nkn);
			if (lc == ')') goto Setupdone;
			if (lc != ',') {
Synerr:
				xerror ("create syntax error");
				return (NO);
			}
			putC (',');
		}
	} else delim = lc;
Setupdone:
	putC ("); ");
	if (ob->Type == OBSERVER) putC ("zz_op->zz_start (); ");
	putC ("zz_op; })");
	if (delim != -1) putC (delim);
#endif
	catchUp ();
	return (YES);
}

#if     ZZ_ATT
void    emitBuilder () {

/* ------------------------------------------------------- */
/* Generates the 'builder' function for the current object */
/* ------------------------------------------------------- */

	int     lc, nn;
	char    **cp;

	if (CObject->Mode == VIRTUAL) return;
	if ((lc = startLkp ()) != END) {
		while (1) {
			lc = lkpC ();
			if (!isspace (lc)) break;
		}
		if (lc == ';') putC ("; ");
	}

	if (CObject->Type == PROCESS) {
		// Emit the state table initializer
		if (CObject->States != NULL) {
			putC (" const int ");
			putC (CObject->Name);
			putC ("::zz_ns");
			putC (SPref);
			for (cp = CObject->States, nn = 0; *cp != NULL; cp++)
				nn++;
			putC (form (" = %1d; ", nn));
			putC ("char *");
			putC (CObject->Name);
			putC ("::zz_sl");
			putC (SPref);
			putC (form (" [%1d] = {", nn));
			for (cp = CObject->States; ; ) {
				putC ('"');
				putC (*cp);
				putC ('"');
				if (*(++cp) == NULL) break;
				putC (", ");
			}
			putC ("}; ");
		}
		if (strcmp (CObject->Name, "Root") == 0) goto Done;
	}

	nn = CObject->Type != MESSAGE && CObject->Type != PACKET &&
		CObject->Type != SGROUP && CObject->Type != CGROUP;

	putC ("inline void zz_bld_");
	putC (CObject->Name);
	putC (" (");
	if (nn) putC ("char *nn = 0");
	putC (") { register ");
	putC (CObject->Name);
	putC (" *p; zz_COBJ [++zz_clv] = (void*) (p = new ");
	putC (CObject->Name);
	putC ("); ");
	if (CObject->Type != OBSERVER)
		// For an observer, zz_start must be called 'by hand'
		putC ("p->zz_start (); ");
	if (nn) {
		// Assign the nickname
		putC ("if (nn != 0) { p->zz_nickname = ");
		putC ("new char [strlen (nn) + 1]; ");
		putC ("strcpy (p->zz_nickname, nn); } ");
	}

	putC ("}; ");
Done:
	if (lc != ';' && lc != END && lc != ERROR) putC (lc);
	if (lc != END) catchUp ();
}
#endif

int     processNew (int del) {

/* ---------------------------------------------------------- */
/* Checks  if  'new'  is not followed by the name of a SMURPH */
/* type                                                       */
/* ---------------------------------------------------------- */

	char    arg [MAXKWDLEN+1];
	int     lc;

	lc = del;
	if (lc == END) {
		xerror ("file ends unexpectedly");
		return (NO);
	}
	if (lc == ERROR || !isalpha (lc) && lc != '_' && lc != '$') return (NO);

	// This is the identifier following 'new'
	lc = getKeyword (arg, lc);

	// Determine the delimiter
	if (lc != ';' && lc != '(') return (NO);

	// Not a SMURPH type
	if (getSym (arg) != NULL)
		xerror ("'new' for a SMURPH object type (use 'create')");
	return (NO);
}

int     processTrace (int del) {

/* ---------------------- */
/* Expands the trace call */
/* ---------------------- */

	// trace (plist);                       -> zz_trace (form (plist));

	char    nkn [MAXKWDLEN+1];
	int     lc;

	lc = del;
	if (lc == END) {
Endfile:
		xerror ("file ends in the middle of 'trace'");
		return (NO);
	}

	if (lc != '(') return (NO);
	putC ("(*zz_ofpp) << \"Time: \" << Time << \" --> \" << form (");
	while (1) {
		// Process arguments
		lc = getArg (nkn);
		if (lc == ERROR) return (NO);
		if (lc == END) goto Endfile;
		putC (nkn);
		if (lc == ')') break;
		if (lc != ',') {
			xerror ("trace syntax error");
			return (NO);
		}
		putC (',');
	}
	putC (") << '\\n'");
	catchUp ();
	return (YES);
}

int     doExposure (int del, SymDesc *obj) {

/* ------------------------------- */
/* Expands the exposure definition */
/* ------------------------------- */

	int     lc;

	// exposure;                            Announcement within object
	//
	// expands into:
	//
	// virtual void zz_expose (int, char *h = NULL, long s = NONE);
	//
	// A::exposure {                        Exposure definition
	//
	// expands into:
	//
	// void A::zz_expose (int zz_emode, char *Hdr, long SId) {
	//      int     zz_em;

	lc = del;
	if (lc == END) {
		xerror ("file ends in the middle of 'exposure'");
		return (NO);
	}

	if (lc == ';') {
	  // Just announcing
	  if (obj == NULL || obj != CObject) {
	    xerror ("'exposure' announced incorrectly");
	    return (NO);
	  }
	  if (obj->EStat == UNEXPOSABLE) {
Unexperr:
		xerror (form ("'%s' is unexposable", obj->Name));
		return (NO);
	  }
	  obj->EStat = ANNOUNCED;
	  putC ("virtual zz_expose (int, char *h = 0, long s = -1);");
	  goto Done;
	}
	if (lc != '{') return (NO);

	if (CodeType) {
		xerror ("exposure definition in illegal context");
		return (NO);
	}
	if (WithinExposure) {
		xerror ("exposure definitions cannot be nested");
		return (NO);
	}
	if (obj == NULL) {
		xerror ("'exposure' unrelated to any object");
		return (NO);
	}
	if (obj->EStat == UNEXPOSABLE) goto Unexperr;
	if (obj->EStat != ANNOUNCED && obj != CObject) {
		xerror (form ("'exposure' not announced for type '%s'",
			obj->Name));
		return (NO);
	}
	WithinExposure = BraceLevel++;
	OnpaperDone = OnscreenDone = NO;
	if (obj == CObject) {
		putC ("virtual ");
		putC ("zz_expose (int zz_emode, char *Hdr = 0, long SId =-1)");
		putC (" {int zz_em;");
	} else {
		putC ("zz_expose (int zz_emode, char *Hdr, long SId)");
		putC (" {int zz_em;");
	}
Done:
	catchUp ();
	return (YES);
}

int     processOnpaper (int del) {

/* --------------------------------------------- */
/* Expands the 'onpaper' header of exposure mode */
/* --------------------------------------------- */

	int     lc;

	// onpaper {
	//
	// expands into:
	//
	// ZZ_ONPP: if ((zz_em = zz_emode) >= ZZ_EMD_0) goto ZZ_ONSC;
	//          if (zz_em == ANY) return; zz_emode = ANY;
	//          switch (zz_em) {

	lc = del;
	if (!WithinExposure) return (NO);
	if (lc == ERROR) return (NO);
	if (lc == END) {
		xerror ("file ends in the middle of 'onpaper'");
		return (NO);
	}

	if (lc != '{') {
		xerror ("'onpaper' syntax error");
		return (NO);
	}
	if (OnpaperDone) {
		xerror ("'onpaper' occurs twice ore more within one exposure");
		return (NO);
	}

	OnpaperDone = YES;
	putC ("ZZ_ONPP: if ((zz_em = zz_emode) >= 4096) goto ZZ_ONSC; ");
	putC ("if (zz_em == -1) return; zz_emode = -1; switch (zz_em) {");
	WithinExpMode = BraceLevel++;

	catchUp ();
	return (YES);
}

int     processOnscreen (int del) {

/* ---------------------------------------------- */
/* Expands the 'onscreen' header of exposure mode */
/* ---------------------------------------------- */

	// onscreen {
	//
	// expands into:
	//
	// ZZ_ONSC: if ((zz_em = zz_emode) < ZZ_EMD_0) goto ZZ_ONPP;
	//          zz_emode = ANY; zz_em -= ZZ_EMD_0;
	//          switch (zz_em) {

	int     lc;

	lc = del;
	if (!WithinExposure) return (NO);
	if (lc == ERROR) return (NO);
	if (lc == END) {
		xerror ("file ends in the middle of 'onscreen'");
		return (NO);
	}

	if (lc != '{') {
		xerror ("'onscreen' syntax error");
		return (NO);
	}
	if (OnscreenDone) {
		xerror ("'onscreen' occurs twice ore more within one exposure");
		return (NO);
	}

	OnscreenDone = YES;
	putC ("ZZ_ONSC: if ((zz_em = zz_emode) < 4096) goto ZZ_ONPP; ");
	putC ("zz_emode = -1; zz_em -= 4096; ");
	putC ("switch (zz_em) {");
	WithinExpMode = BraceLevel++;

	catchUp ();
	return (YES);
}

int     processExmode (int del) {

/* ------------------- */
/* Expands 'exmode m:' */
/* ------------------- */

	// exmode m:
	//
	// expands into: 'break; case m: zz_dispfound = YES;'

	int     lc;
	char    m [MAXKWDLEN+1];

	lc = del;
	if (!WithinExposure || !WithinExpMode) return (NO);
	if (lc == ERROR) return (NO);
	if (lc == END) {
		xerror ("file ends in the middle of 'exmode'");
		return (NO);
	}

	// Get the expression
	lc = getKeyword (m, lc, YES);   // May be a number

	if (lc != ':') {
		xerror ("'exmode' syntax error");
		return (NO);
	}
	putC ("break; case ");
	putC (m);
	putC (": zz_dispfound = 1; ");

	catchUp ();
	return (YES);
}

int     doExpose (int del) {

/* -------------------------------------------------- */
/* Expands 'expose;' (must be preceded by 'symbol::') */
/* -------------------------------------------------- */

	// expose;
	//
	// expands into: 'zz_expose (zz_emode, Hdr, SId)'

	int     lc;

	// We know for sure that we are WithinExposure
	lc = del;
	if (lc == ERROR) return (NO);
	if (lc == END) {
		xerror ("file ends in the middle of 'expose'");
		return (NO);
	}

	if (lc != ';') {
		xerror ("'expose' syntax error");
		return (NO);
	}
	putC ("zz_expose (zz_emode, Hdr, SId);");

	catchUp ();
	return (YES);
}

int     processExpose (int del) {

/* -------------------------------------------------------- */
/* Processes illegal exposure -- not preceded by 'symbol::' */
/* -------------------------------------------------------- */

#if     ZZ_ATT
	del--;  // To keep Cfront happy
#endif
	if (WithinExposure)
		xerror ("'expose' must be preceded by 'symbol::'");
	return (NO);
}

int     processPerform (int del) {

/* ----------------------------------------------------- */
/* Preprocesses 'perform' within object type declaration */
/* ----------------------------------------------------- */

	if (CodeType) {
		xerror ("'perform' within another 'perform'");
		return (NO);
	}

	if (CObject == NULL) return (NO);
	if (CObject->Type == PROCESS) {
		CodeType = PROCESS;
	} else if (CObject->Type == OBSERVER) {
		CodeType = OBSERVER;
	} else return (NO);
	CodeTypeBC = BraceLevel;
	CodeObject = CObject;
	putC ("void ");
	return (doPerform (del, CObject));
}

int     processExposure (int del) {

/* ------------------------------------------------------ */
/* Preprocesses 'exposure' declared within an object type */
/* ------------------------------------------------------ */

	putC ("void ");
	return (doExposure (del, CObject));
}

int     processKeyword (int fc) {

/* --------------------------------------------------------- */
/* Try to match the keyword with the ones in your dictionary */
/* --------------------------------------------------------- */

	char    kwd [MAXKWDLEN+1], *septr;
	int     del, skc, status;
	KFUNC   f;

	SkLineCnt = 0;  // Count skipped lines
	if (startLkp () == END) return (NO);
	// This is the keyword we should try to match
	del = getKeyword (kwd, fc);
	// del is the non-blank delimiter

	// Check if it is not a sequence of the form: symbol::perform,
	// symbol::exposure, or symbol::expose.
	// Must be able to tell whether the perform belongs to a process
	// or an observer

	if (!ArrowInFront && del == ':') {
		char    pfm [MAXKWDLEN+1];
		SymDesc *ob;
		// Save the lookahead pointer
		septr = EnLkpPtr;
		skc = SkLineCnt;
		// Get the second delimiter
		del = lkpC ();
		if (del != ':') {
			// Backspace
			del = ':';
			EnLkpPtr = septr;
			SkLineCnt = skc;
			goto Tryother;
		}
		// Now, check if it is followed by 'perform' or 'exposure'
		del = getKeyword (pfm);
		if (strcmp (pfm, "perform")) {
			// Something else
			if (strcmp (pfm, "exposure")) {
				// Not exposure either
				if (!WithinExposure || strcmp (pfm, "expose")) {
					// This was the last chance
					del = ':';
					EnLkpPtr = septr;
					SkLineCnt = skc;
					goto Tryother;
				}
				// Here we have 'expose'
				putC (kwd);
				putC ("::");
				status = doExpose (del);
				goto Done;
			}
			// Here we have exposure
			if ((ob = getSym (kwd)) == NULL) {
			      xerror ("exposure for an undefined object type");
			      return (NO);
			}
			putC ("void ");
			putC (kwd);
			putC ("::");
			status = doExposure (del, ob);
			goto Done;
		}
		if (CodeType) {
			xerror ("'perform' within another 'perform'");
			return (NO);
		}
		if (CObject != NULL) {
		    xerror ("illegal '::perform' declaration inside an object");
		    return (NO);
		}
		// Determine the status of the ownining object
		if ((ob = getSym (kwd)) == NULL) {
			xerror ("'perform' for an undefined process type");
			return (NO);
		}
		if (ob->Type == PROCESS) {
			CodeType = PROCESS;
		} else if (ob->Type == OBSERVER) {
			CodeType = OBSERVER;
		} else {
			xerror ("'perform' neither in process nor in observer");
			return (NO);
		}
		CodeTypeBC = BraceLevel;
		if (ob->Mode != REAL) {
		   xerror ("'perform' for a virtual or announced process type");
		   CodeType = NO;
		   return (NO);
		}
		CodeObject = ob;
		putC ("void ");
		putC (kwd);
		putC ("::");
		status = doPerform (del, ob);
		goto Done;
	}

Tryother:
		
	if (!islower (*kwd)) return (NO);
	if ((f = getKey (kwd)) == NULL) return (NO);

	status = f (del);
Done:
	if (status) {
		// Make sure the output line count agrees
		while (SkLineCnt--) putC ('\n');
	}
	return (status);
}

main (int argc, char *argv []) {

/* ------------- */
/* The main loop */
/* ------------- */

	int     c;      // The current character

#ifdef  DEBUG
	Nodebug = (argc > 1);
#endif
	// Get the 'static' prefix
	if (argc < 2) {
		SPref [0] = '\0';
	} else {
		for (argv++, c = 0; c < 11 && isdigit ((*argv)[c]); c++)
			SPref [c] = (*argv) [c];
		SPref [c] = '\0';
	}

	// Initialize the tables
	for (c = 0; c < HASHTS; c++) {
		SymTab [c] = NULL;
		KeyTab [c] = NULL;
	}

	// Add keywords
	new KeyDesc ("trace", processTrace);
	new KeyDesc ("identify", processIdentify);
	new KeyDesc ("terminate", processTerminate);
	new KeyDesc ("expose", processExpose);
	new KeyDesc ("exposure", processExposure);
	new KeyDesc ("onpaper", processOnpaper);
	new KeyDesc ("onscreen", processOnscreen);
	new KeyDesc ("exmode", processExmode);
	new KeyDesc ("sendJam", processSendJam, YES);
	new KeyDesc ("transient", processTransient);
	new KeyDesc ("new", processNew);
	new KeyDesc ("states", processStates);
	new KeyDesc ("signals", processSignals);
	new KeyDesc ("eobject", processEobject);
	new KeyDesc ("observer", processObserver);
	new KeyDesc ("inspect", processInspect);
	new KeyDesc ("timeout", processTimeout);
	new KeyDesc ("link", processLink);
	new KeyDesc ("traffic", processTraffic);
	new KeyDesc ("message", processMessage);
	new KeyDesc ("packet", processPacket);
	new KeyDesc ("perform", processPerform);
	new KeyDesc ("process", processProcess);
	new KeyDesc ("transmit", processTransmit, YES);
	new KeyDesc ("sleep", processSleep);
	new KeyDesc ("skipto", processSkipto);
	new KeyDesc ("station", processStation);
	new KeyDesc ("proceed", processProceed);
	new KeyDesc ("create", processCreate);
	new KeyDesc ("state", processState);
	new KeyDesc ("wait", processWait, YES);

	addSym ("Station", STATION, DEFINED, REAL);
	addSym ("EObject", EOBJECT, DEFINED, REAL);
	addSym ("Process", PROCESS, DEFINED, REAL);
	addSym ("Observer", OBSERVER, DEFINED, REAL);
	addSym ("Traffic", TRAFFIC, DEFINED, REAL);
	addSym ("Message", MESSAGE, DEFINED, REAL, UNEXPOSABLE);
	addSym ("Packet", PACKET, DEFINED, REAL, UNEXPOSABLE);
	addSym ("Link", LINK, DEFINED, REAL);
	addSym ("BLink", LINK, DEFINED, REAL);
	addSym ("PLink", LINK, DEFINED, REAL);
	addSym ("ULink", LINK, DEFINED, REAL);
	addSym ("CLink", LINK, DEFINED, REAL);
	addSym ("Port", PORT, DEFINED, REAL);
	addSym ("RVariable", RVARIABLE, DEFINED, REAL);
	addSym ("SGroup", SGROUP, DEFINED, REAL, UNEXPOSABLE);
	addSym ("CGroup", CGROUP, DEFINED, REAL, UNEXPOSABLE);

	// Just in case: cccp should do it in the first line
	CFName [0] = '\0';
        putC ("#include \"Smurph.h\"\n");
	while (BraceLevel) {
	  if (WithinString) {
	    if ((c = getC ()) == END) { // End of file
UnclString:
	      xerror (form ("unclosed string opened at line %1d", StrLine));
	      finish ();
	    }
	    putC (c);
	    if (c == '\\') {            // Escape, copy the next char verbatim
	      if ((c = getC ()) == END) goto UnclString;
	      putC (c);
	      continue;
	    }
	    if (c == WithinString)      // The matching character
	      WithinString = NO;
	    continue;
	  }
	  switch (c = getC ()) {

	    case '#' :

	      // Possible line command

	      if (BOLine)
		processLineCommand ();
	      else
		putC (c);
	      continue;

	    case '\'':
	    case '"':

	      // Start of text

	      ArrowInFront = NO;
	      putC (WithinString = c);
	      StrLine = LineNumber;
	      continue;

	    case END:

	      finish ();
	      break;                    // No return from finish - just in case

	    case '}':

	      // Count braces

	      BraceLevel--;
	      ArrowInFront = NO;
	      if (CodeType) {
		if (BraceLevel == CodeTypeBC) {
		  if (CodeWasState) putC ('}');
		  CodeType = NO;
		  CodeObject = NO;
		}
	      }

	      if (CObject != NULL && CObjectBC == BraceLevel) {
		// Pop the stack
		if ((CObject->Type == PROCESS || CObject->Type ==
		  OBSERVER) && !(CObject->PAnn) && CObject->Mode == REAL)
		    putC ("void zz_code (); }");
		else
		    putC ('}');
#if     ZZ_ATT
		emitBuilder ();
#endif
		if (DeclStackDepth) {
		  DeclStackDepth--;
		  CObject = DeclStack [DeclStackDepth] . CObject;
		  CObjectBC = DeclStack [DeclStackDepth] . CObjectBC;
		} else {
		  CObject = NULL;
		}
	      } else if (WithinExposure && BraceLevel == WithinExposure) {
		WithinExposure = NO;
		if (OnpaperDone && !OnscreenDone) {
		  putC ("ZZ_ONSC: return; ");
		} else if (OnscreenDone && !OnpaperDone) {
		  putC ("ZZ_ONPP: return; ");
		}
#if     ZZ_ATT
		// This is garbage -- to avoid Cfront complaints about
		// unused parameters
		if (!OnpaperDone && !OnscreenDone) {
		  putC ("zz_em = zz_emode; ");
		}
		putC ("zz_Stroustrup = (long) Hdr; zz_Stroustrup = SId; ");
#endif
		putC ('}');
	      } else {
		if (WithinExpMode && BraceLevel == WithinExpMode)
		  WithinExpMode = NO;
		putC ('}');               // Put the closing brace
	      }
	      continue;

	    case '{':

	      putC ('{');
	      ArrowInFront = NO;
	      BraceLevel++;
	      continue;

	    case '-':

	      // This may be the first element of a '->'

	      putC ('-');
	      if (!ArrowInFront) ArrowInFront = 1;
	      continue;

	    case '.':

	      putC ('.');
	      if (!ArrowInFront) ArrowInFront = 2;
	      continue;

	    case '>':

	      // Second part of '->' ?

	      putC ('>');
	      if (ArrowInFront == 1)
		ArrowInFront = 2;
	      else
		ArrowInFront = NO;
	      continue;

	    default:

	      if (isspace (c)) {
		putC (c);
		// Don't let it destroy the arrow status
		if (ArrowInFront == 1) ArrowInFront = NO;
		continue;
	      }

	      // Check if it is the beginning of an interesting keyword

	      if (! (isalnum (c) || c == '_' || c == '$')) {
		ArrowInFront = NO;
		putC (c);
		continue;
	      }

	      if (isdigit (c) || !processKeyword (c)) {
		// Skip this keyword
		ArrowInFront = NO;
		putC (c);
		while ((c = peekC ()) == '_' || c == '$' || isalnum (c))
			putC (getC ());
	      }
	      continue;
	  }
	}
}
